############################################################################
# Joshua R. Boverhof, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
#
# Copyright (c) 2001 Zolera Systems.  All rights reserved.
#
# REQUIRES: ZSI-1.2
#
###########################################################################
import BaseHTTPServer
import httplib, cStringIO, types, time

from ZSI.client import *
from pyGlobus import io
from pyGlobus.security import GSSCred

"""
import base64, httplib, cStringIO as StringIO, types, time
from ZSI import _copyright, ParsedSoap, SoapWriter, TC, ZSI_SCHEMA_URI, \
    FaultFromFaultMessage, _child_elements
from ZSI.auth import AUTH
from ZSI.client import Binding

_b64_encode = base64.encodestring

_AuthHeader = '<BasicAuth xmlns="' + ZSI_SCHEMA_URI + '''">
    <Name>%s</Name><Password>%s</Password>
</BasicAuth>'''
"""

################################################################################
# Soap Client
################################################################################
class GSIBinding(Binding):
    '''Object that represents a binding (connection) to a SOAP server.
    Once the binding is created, various ways of sending and
    receiving SOAP messages are available, including a "name overloading"
    style.
    '''

    def __init__(self, **kw):
	'''Initialize.
	Keyword arguments include:
	    host, port -- where server is; default is localhost
	    ssl -- no default
                  =1 use SSL
	          >1 use GSI
	    url -- resource to POST to
	    soapaction -- value of SOAPAction header
	    auth -- (type, name, password) triplet; default is unauth
	    ns -- default namespace
	    nsdict -- namespace entries to add
	    tracefile -- file to dump packet traces
	    cert_file, key_file -- SSL data (q.v.)
	    readerclass -- DOM reader class
	'''
        Binding.__init__(self, **kw)
        
    def accept_delegatation(self, oids, buffers, time_req):
        cred = self.socket.accept_delegation(oids, buffers, time_req)
        return cred
    def Send(self, url, opname, obj, **kw):
	'''Send a message.  If url is None, use the value from the
	constructor (else error). obj is the object (data) to send.
	Data may be described with a requesttypecode keyword, or a
	requestclass keyword; default is the class typecode (if
	there is one), else Any.
	'''
        # Get the TC for the obj.
	if kw.has_key('requesttypecode'):
	    tc = kw['requesttypecode']
	elif kw.has_key('requestclass'):
	    tc = kw['requestclass'].typecode
	elif type(obj) == types.InstanceType:
	    tc = obj.__class__.__dict__.get('typecode')
	    if tc == None: tc = TC.Any(opname, aslist=1)
	else:
	    tc = TC.Any(opname, aslist=1)

	# Determine the SOAP auth element.
	if self.auth_style & AUTH.zsibasic:
	    auth_header = _AuthHeader % (self.auth_user, self.auth_pass)
	else:
	    auth_header = None

	# Serialize the object.
	s = cStringIO.StringIO()
	d = {}
	if self.ns: d[''] = self.ns
	d.update(kw.get('nsdict', self.nsdict) or {})
	sw = SoapWriter(s, nsdict=d, header=auth_header)
	if kw.has_key('_args'):
	    sw.serialize(kw['_args'], tc)
	else:
	    sw.serialize(obj, tc, typed=0)
	sw.close()
	soapdata = s.getvalue()

	# Tracing?
	if self.trace:
	    print >>self.trace, "_" * 33, time.ctime(time.time()), "REQUEST:"
	    print >>self.trace, soapdata

        # Send the request.
	if self.ssl > 1:
	    self.h = io.GSIHTTPConnection(self.host, self.port)
        elif self.ssl == 1:
	    self.h = httplib.HTTPSConnection(self.host, self.port,
			**self.ssl_files)
        else:
            self.h = httplib.HTTPConnection(self.host, self.port)
            
	self.h.connect()
	self.h.putrequest("POST", url or self.url)
	self.h.putheader("Content-length", "%d" % len(soapdata))
	self.h.putheader("Content-type", 'text/xml; charset=utf-8')
	self.h.putheader("SOAPAction", kw.get('soapaction', self.soapaction))
        
        if self.auth_style & AUTH.httpbasic:
	    val = _b64_encode(self.auth_user + ':' + self.auth_pass).strip()
	    self.h.putheader('Authorization', 'Basic ' + val)
	for header,value in self.user_headers:
	    self.h.putheader(header, value)
	self.h.endheaders()
	self.h.send(soapdata)
        
	# Clear prior receive state.
	self.data, self.ps = None, None
        return

################################################################################
# SOAP Server
################################################################################
class GSISoapServer(io.GSITCPSocketServer):
    """A server that processes SOAP messages.

    """
    def __init__(self,addr, requesthandler):
        """Start up the server."""
        io.GSITCPSocketServer.__init__(self, addr, requesthandler)
        self._factory = MessageFactory()
        return
    def set_factory(self, msgFactory):
        self._factory = msgFactory()
    def close(self):
        self.server_close()
    def init_delegation(self, cred, oids, buffers, time_req):
        self.socket.init_delegation(cred, oids, buffers, time_req)
    
