import sys, getopt, unittest, time
from threading import *
from pyGlobus import ioc
from pyGlobus.io import GSITCPSocket, NetIOAttr, AuthData, TCPIOAttr
from pyGlobus.util import Buffer, GlobusException

gPort = None

def auth_callback(arg, handle, identity, context):
    print "inside auth callback"
    print "id is: %s" % identity
    return 1
 
def usage():
    useStr = "test_tcp_client -h host -p port"
    print useStr
    return

class GSITCPSocketCase(unittest.TestCase):
    def setUp(self):
        print "Begin GSITCPSocketServer"
        self._server = testClientServer(client=0)
        self._client = testClientServer(client=1)
        return
    def check_client_server(self):
        time.sleep(1)
        self._server.start()
        time.sleep(1)
        self._client.run(gPort)
        return
    
class testClientServer(Thread):
    def __init__(self, client=None):
        Thread.__init__(self)
        self._client = client
        self.soc = GSITCPSocket()
        self.authData = AuthData()
        self.attr = TCPIOAttr()
        self.attr.set_authentication_mode(ioc.GLOBUS_IO_SECURE_AUTHENTICATION_MODE_GSSAPI)
        self.attr.set_delegation_mode(ioc.GLOBUS_IO_SECURE_DELEGATION_MODE_FULL_PROXY)
        self.attr.set_channel_mode(ioc.GLOBUS_IO_SECURE_CHANNEL_MODE_CLEAR)
        return
    def run(self,port=gPort,host='localhost'):
        if not self._client:
            self.authData.set_callback(auth_callback, None)
            self.attr.set_authorization_mode(\
                ioc.GLOBUS_IO_SECURE_AUTHORIZATION_MODE_CALLBACK, self.authData)
            self.init_Server()
        else:
            self.attr.set_authorization_mode(\
                ioc.GLOBUS_IO_SECURE_AUTHORIZATION_MODE_SELF, self.authData)
            self.init_Client(host, port)
        return
    def init_Server(self):
        port = self.soc.create_listener(self.attr)
        print "created a listener on port %s" % port
        global gPort
        gPort = port
        self.soc.listen()
        print "returned from listen"
        (childSoc, addr) = self.soc.accept(self.attr)
        print "accepted a connection"
        ctx = childSoc.get_security_context()
        print "got the security context"
        tuple = ctx.inquire()
        print "finished inquire_context"
        name = tuple[0].display()
        print "Source name is: %s" % name
        cred = childSoc.get_delegated_credential()
        print "Got the delegated credential"
        size = 46
        buf = Buffer(size)
        print "created a buffer"
        bytesRead = childSoc.read(buf, size, size)
        print "Read %i bytes" % bytesRead
        print "buf is: %s" % buf.as_string(bytesRead)
        print "Writing back buffer"
        str = "Fee, Fie, Foo, Fum"
        childSoc.write(str, len(str))   
        del cred
        del childSoc
        del self.soc
        return 
    def init_Client(self,host,port):
        self.soc.connect(host, port, self.attr)
        print "connected to host %s, on port %i" % (host, port)
        str = "Foobar, fOobar, foObar, fooBar, foobAr, foobaR"
        nBytes = self.soc.write(str, len(str))
        print "Wrote %i bytes out" % nBytes
        size = 18
        buf = Buffer(size)
        print "created a buffer"
        bytesRead = self.soc.read(buf, size, size)
        print "Read %i bytes" % bytesRead
        print "buf is: %s" % buf.as_string(bytesRead)
        return

def makeTestSuite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(GSITCPSocketCase, "check_"))
    return suite

def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "h:p:s")
    except getopt.GetoptError:
        usage()
        sys.exit(-1)
    host = None
    if opts:
        for o, a in opts:
            if o == "-h":
                host = a
            elif o == "-p":
                port = int(a)
            elif o == "-s":
                testCS = testClientServer(client=0)
                testCS.run()
                break
            else:
                usage()
                sys.exit(-1)
        else:
            testCS = testClientServer(client=1)
            testCS.run(host,port)
    else:
        unittest.main(defaultTest="makeTestSuite")
    return
    
if __name__ == '__main__':
    main()

    
