############################################################################
# Keith R. Jackson, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
from pyGlobus.gassCache import GassCache
from pyGlobus.util import GlobusException
import ConfigParser, unittest

#The config file is ../../test/config.py
config = None # ConfigParser object

SECTION = "gassCache"
URL = "url"

#############################################################################
#  Unit test for the gassCache Module.
##############################################################################
class GassCacheTestCase(unittest.TestCase):
    "Test case for the GassCache module"

    def setUp(self):
        global config
        if config == None:
            config = ConfigParser.ConfigParser()
            config.read("config.py")
        print "\nGassCache Test Case begin\n"
        return

    def tearDown(self):
        print "\nGassCacheTestCase finished\n"
        return
    
    def checkGassAddCleanupFileCache(self):
        url = config.get(SECTION, URL)
        try:
            gCache = GassCache()
            gCache.open()
            ret, timestamp, file = gCache.add(url, "foobar", 1)
            gCache.add_done(url, "foobar",timestamp)
            gCache.cleanup_file(url)
            gCache.close()
            del gCache
        except GlobusException, ex:
            print "Exception occured during testing."
            assert 0, ex.args
        return

    def checkGassAddCleanupTagCache(self):
        url = config.get(SECTION, URL)
        try:
            gCache = GassCache()
            gCache.open()
            ret, timestamp, file = gCache.add(url, "foobar", 1)
            gCache.add_done(url, "foobar",timestamp)
            gCache.cleanup_tag(url, "foobar")
            gCache.close()
            del gCache
        except GlobusException, ex:
            print "Exception occured during testing."
            assert 0, ex.args
        return
    
    def checkGassAddDeleteCache(self):
        url = config.get(SECTION, URL)
        try:
            gCache = GassCache()
            gCache.open()
            ret, timestamp, file = gCache.add(url, "foobar", 1)
            gCache.add_done(url, "foobar",timestamp)
            timestamp = gCache.delete_start(url, "foobar")
            gCache.delete(url, "foobar", timestamp, 1)
            gCache.close()
            del gCache
        except GlobusException, ex:
            print "Exception occured during testing."
            assert 0, ex.args

def makeTestSuite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(GassCacheTestCase, "check"))
    return suite

def main():
    unittest.main(defaultTest="makeTestSuite")


if __name__ == "__main__" : main()
    
