############################################################################
# Matt Rodriguez, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################


import ConfigParser, unittest
from pyGlobus.gassTransfer import TransferReqAttr
from pyGlobus.gassTransfer import TransferClient
from pyGlobus.gassTransfer import TransferServer
from pyGlobus.gassTransfer import ListenerAttr
from pyGlobus.gassTransfer import TransListener
from pyGlobus.gassTransfer import TransRequest
from pyGlobus.gassTransfer import TransData
from pyGlobus.util import GlobusException


from threading import *
from pyGlobus.util import Buffer
import sys

"""
Unittest for the TransferReqAttr module.
"""
config_file = "config.py"
config = None # ConfigParser object

class TransferReqAttrTestCase(unittest.TestCase):
    """
    Test case for the TransferReqAttr class.
    """
        
    def chekTransferReqAttr(self):
        global config, config_file
        print "\nTransferReqAttrTestCase begin\n"
        if config == None:
            config = ConfigParser.ConfigParser()
            config.read(config_file)
        try:
            name = TransferReqAttr("http")
            del name
            print "\nTransferReqAttrTestCase completed\n"
        except GlobusException, ex:
            assert 0, ex.args
        return

    


class GassTransferTestCase(unittest.TestCase):         
    """Test cases for the gassTransfer module"""
    def setUp(self):
        global config
        if config == None:
            config = ConfigParser.ConfigParser()
            config.read(config_file)
        scheme = "http"
        length = 1 * 1024
        self._condV = Condition(Lock())
        self._done = 1
        self._reqAttr = TransferReqAttr(scheme)
        self._client = TransferClient()
        self._server = TransferServer(scheme)
        self._url =  self._server.get_base_url()
        self._scheme = scheme
        self._length = length
        self._buffer_recv = Buffer(length)
        self._Trans = TransData()
        self._oldSendCbHandle = None
        self._newSendCbHandle = None
        self._oldRecvCbHandle = None
        self._newRecvCbHandle = None
        self._client_request = None
        self._server_request = None
        return

    def tearDown(self):
        del self._client_request
        del self._server_request
        self._src_file.close()
        self._dest_file.close()
        
    def registerListenCallback(self, cv, request):
        print "in registerListenCallback"
        self.callbackSynchronization(cv)
        
    def registerGetCallback(self,cv,request):
        print "in registerGetCallback"
        self.callbackSynchronization(cv)
        
    def registerPutCallback(self, cv, request):
        print "in registerPutCallback"
        
        
    def registerAcceptCallback(self, cv, request):
        print "in registerAcceptCallback"
        self.callbackSynchronization(cv)
        
    def callbackSynchronization(self, cv):
        cv.acquire()
        self._done = 1
        cv.notify()
        cv.release()

    def receiveBytesCallback(self,cv, request, bufHandle, bufLen,  eof):
        self._dest_file.write(bufHandle)
        self._dest_file.flush()
        if not eof:
            try:
                handle = self._Trans.receive_bytes(self._recv_request,self._buffer_recv,1, self.receiveBytesCallback, cv)
                
                if not self._oldRecvCbHandle == None:
                    self._Trans.free_callback(self._oldRecvCbHandle)
                    
                self._oldRecvCbHandle = self._newRecvCbHandle
                self._newRecvCbHandle = handle
            except Exception, e:
                print e
                traceback.print_exception()
                sys.exit(-1)
        else:
            cv.acquire()
            self._done = 1
            cv.notify()
            cv.release()
        return

    def sendBytesCallback(self,cv, request, bufHandle, bufLen, eof):
        if not eof:
            #The send buffer needs to stay in scope until this callback has been called again
            self._buffer_send = self._src_file.read(self._length)
            
            try:
                
                handle = self._Trans.send_bytes(self._send_request,self._buffer_send,self._buffer_send == "", self.sendBytesCallback, 0)
                if not self._oldSendCbHandle == None:
                    self._Trans.free_callback(self._oldSendCbHandle)

                self._oldSendCbHandle = self._newSendCbHandle
                self._newSendCbHandle = handle
            except Exception, e:
                print e
                traceback.print_exception()
                sys.exit(-1)
        return
    
    def checkGassGet(self):
       global config, config_file
       try:
           print "testing GASS Get with nonblocking calls "
           if config == None:
               config = ConfigParser.ConfigParser()
               config.read(config_file)
           
           src = config.get("gassTransfer", "src")
           #src = '/tmp/bigFile_remove'
           dest = config.get("gassTransfer", "destGet")
           print "source file is: ", src
           print "destination file is: ", dest
           self._dest_file = open(dest, "w")
           self._src_file = open(src,"r") 
           print "url needed for test_client"
           url = self._url + src
           print url
           self._condV.acquire()
           self._done = 0
           self._server.register_listen(self.registerListenCallback,self._condV)
       
    
           client_request = self._client.register_get(self._reqAttr,url, self.registerGetCallback,self._condV)

           while self._done == 0:
               print "waiting for register_listen callback"
               self._condV.wait()
           self._condV.release()

           self._condV.acquire()
           self._done = 0
           server_request = self._server.register_accept(self._reqAttr, self.registerAcceptCallback, self._condV)

           while self._done == 0:
               print "waiting for callback register_accept callback"
               self._condV.wait()
           self._condV.release()
    
           self._condV.acquire()
           self._done = 0
           server_request = self._server.authorize(server_request, 0)    

           while self._done == 0:
               print "waiting for callback register_get callback"
               self._condV.wait()
           self._condV.release()

           self._buffer_send = self._src_file.read(self._length)
           self._send_request = server_request
           self._newCbHandle = self._Trans.send_bytes(self._send_request, self._buffer_send, self._buffer_send == "", self.sendBytesCallback,0)
           
           buffer_receive = Buffer(self._length)
           self._condV.acquire()
           self._done = 0
           self._recv_request = client_request
           self._Trans.receive_bytes(self._recv_request,buffer_receive,1, self.receiveBytesCallback, self._condV)
           while self._done == 0:
               print "waiting for callback receive_bytes callback"
               self._condV.wait()
           self._condV.release()
           
       except GlobusException, ex:
           print ex
       print "DONE"

    def checkGassPut(self):
        print "testing GASS PUT with non blocking calls"
        try:
            
            global config, config_file
            if config == None:
                config = ConfigParser.ConfigParser()
                config.read(config_file)
            src = config.get("gassTransfer", "src")
            dest =  config.get("gassTransfer", "destPut")
            self._dest_file = open(dest, "w")
            self._src_file = open(src,"r")
            print "Putting file: ", src
            print "url needed for test_client"
            url = self._url + dest
            print url
                 
            self._condV.acquire()
            self._done = 0
            self._server.register_listen(self.registerListenCallback,self._condV)
            client_request = self._client.register_put(self._reqAttr, url, self.registerPutCallback, self._condV)
            
            while self._done == 0:
                print "waiting for callback register_listen"
                self._condV.wait()
            self._condV.release()
            
            self._done = 0
            self._condV.acquire()
            server_request = self._server.register_accept(self._reqAttr,self.registerAcceptCallback, self._condV)
           
            while self._done == 0:
                print "waiting for callback register_put"
                self._condV.wait()
            self._condV.release()
            
            self._server_request = self._server.authorize(server_request)  
            self._buffer_send = self._src_file.read(self._length)
            self._send_request = client_request
            
            self._Trans.send_bytes(self._send_request,self._buffer_send, self._buffer_send == "", self.sendBytesCallback,0)  

            self._buffer_recv = Buffer(self._length)
            self._condV.acquire()
            self._done = 0
            self._recv_request = server_request
            self._Trans.receive_bytes(self._recv_request, self._buffer_recv,1 , self.receiveBytesCallback, self._condV)

            while self._done == 0:
                print "waiting for callback receive_bytes"
                self._condV.wait()
            self._condV.release()
            self._dest_file.close()
        except GlobusException, ex:
            print ex
    print "DONE"
    
    def checkGassAppend(self):
        print "testing GASS APPEND with non blocking calls"
        try:    
            global config, config_file
            if config == None:
                config = ConfigParser.ConfigParser()
                config.read(config_file)
            src = config.get("gassTransfer", "src")
            dest =  config.get("gassTransfer", "destAppend")
            self._dest_file = open(dest, "a")
            self._src_file = open(src,"r")
            print "Putting file: ", src
            print "url needed for test_client"
            url = self._url + dest
            print url
        
            self._condV.acquire()
            self._done = 0
            self._server.register_listen(self.registerListenCallback,self._condV)
            client_request = self._client.register_append(self._reqAttr, url, self.registerPutCallback, self._condV)
            
            while self._done == 0:
                print "waiting for callback register_listen"
                self._condV.wait()
            self._condV.release()
            
            self._done = 0
            self._condV.acquire()
            server_request = self._server.register_accept(self._reqAttr,self.registerAcceptCallback, self._condV)
           
            while self._done == 0:
                print "waiting for callback register_put"
                self._condV.wait()
            self._condV.release()
          
            self._server_request = self._server.authorize(server_request)  
            self._buffer_send = self._src_file.read(self._length)
            self._send_request = client_request
            self._Trans.send_bytes(self._send_request,self._buffer_send, self._buffer_send == "", self.sendBytesCallback,0)   
            self._buffer_recv = Buffer(self._length)
            self._condV.acquire()
            self._done = 0
            self._recv_request = server_request
            self._Trans.receive_bytes(self._recv_request, self._buffer_recv,1 , self.receiveBytesCallback, self._condV)

            while self._done == 0:
                print "waiting for callback receive_bytes"
                self._condV.wait()
            self._condV.release()
            self._dest_file.close()
        except GlobusException, ex:
            print ex
    print "DONE"

class GassTransListenerTestCase(unittest.TestCase):

    def setUp(self):
        self._listener = TransListener()
        self._done = 0
        
    def listenerCB(self,cv, listener):
        print "in listenerCB"
        cv.acquire()
        self._done = 1
        cv.notify()
        cv.release()
        
    def checkListenerUrl(self):
        print "Checking Listener Url"
        url = self._listener.get_base_url()
        print url
        print "DONE"
        
    def checkCloseListener(self):
        print "checkCloseListener"
        print "Listener is open at url: ", self._listener.get_base_url()
        cv = Condition(Lock())
        cv.acquire()
        self._listener.close_listener(self.listenerCB, cv)
        while self._done == 0:
            print "Waiting for listenerCB"
            cv.wait()
        cv.release()
        print "Closed listener"
        print "DONE"

class GassTransListenerAttrTestCase(unittest.TestCase):

    def setUp(self):
        self._listenerAttr = ListenerAttr()

    def checkListenerAttr(self):
        del self._listenerAttr
        
def makeTestSuite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TransferReqAttrTestCase, "check"))
    suite.addTest(unittest.makeSuite(GassTransferTestCase, "check"))
    suite.addTest(unittest.makeSuite(GassTransListenerTestCase, "check"))
    suite.addTest(unittest.makeSuite(GassTransListenerAttrTestCase, "check"))
    return suite

def main():
    unittest.main(defaultTest="makeTestSuite")


if __name__ == "__main__" : main()
    
                  

