###########################################################################
# Joshua R. Boverhof, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
#
###########################################################################
import sys, unittest, ConfigParser, time, getopt, socket
from threading import *
from pyGMA.message import Message, MessageFactory, \
     GMAMessageFactory, EventMessage, EventFilter
from pyGMA.event import Event, EventField, EventContainer
from pyGlobus import utilc
from pyGlobus.io import GSITCPSocketException, AuthData, GSITCPSocketServer
from pyGlobus.util import GlobusException
from pyGlobus.ZSISOAP import GSISoapServer, GSIBinding, WsdlSoapRequestHandler
from ZSI import TC
from pyGMA.core import SoapRequestHandler

g_port = 8090

###########################################################################
# RequestHandler
###########################################################################
class RFTRequestHandler(SoapRequestHandler):
    """Perform processing of SOAP messages for the producer.

    Transform the XML document into method arguments,
    and the return arguments back into XML.

    """
    def __init__(self, *l):
        SoapRequestHandler.__init__(self, *l)
        return
    def completeTransferReply(self,msg):
        e = Event('handle')
        rpy = completeTransferReply(e)
        return rpy
    def listTransferReply(self,msg):
        '''Very simple verification that things are OK.'''
        p = Event("ProducerEntry")
        e = [Event("Ping"),Event("Iperf")]
        e[0].add(EventField("RTT",value=1.23,units="ms"))
        e[1].add(EventField("BW",value=83.5,units="Mb/s"))
        for i in (0,1):
            c = EventContainer("Endpoints")
            c.add(EventField("Source",value="src.gma.net"))
            c.add(EventField("Destination",value="dst.gma.net"))
            e[i].add(c)
        p.addAll(e)
        # easily modified..
        p["Ping"]["RTT"] = 400
        #p.flatten()
        print "3--"
        rpy = listTransferReply(p)
        return rpy
class WsdlHandler(SoapRequestHandler):
      def __init__(self, *l):
          SoapRequestHandler.__init__(self, *l)
          return
      def handle(self, request):
          return self._safeHandleRequest(self._handle, request)
      def _handle(self,request):
          #See if the required name header is present
          if not request.headers.has_key('name'): return None
          name = request.headers['name']
          repo = Core.GetRepository()
          try:
              model = repo.getModel()
              #Get all of the statements from the model that define WSDL
              #documents with the requested name
              stmts = model.complete(None,"http://schema.4suite.org#wsdl",name)
              if not stmts:
                  return self._createErrorResponse(request, 404, uri=name)
              #Fetch the document
              uri = stmts[0].subject
              body = repo.fetchDocument(uri).getContent()
          finally:
              repo.txRollback()
          headers = {'Content-type' : 'text/xml; charset=iso-8859-1',
                     'Content-length' : len(body),
                     }
          return CommonHandler.Response(200, headers, body)
          
def Register(properties):
    #Register are new handler to accept "WSDL" requests
    handler = WsdlHandler()
    return [(handler.handle, 'WSDL')]



###########################################################################
#  MessageFactory
###########################################################################
class RFTMessageFactory(GMAMessageFactory):
    """Create RFT messages from DOM objects or XML strings.

    """
    def __init__(self):
        """Set up factory."""
        MessageFactory.__init__(self)
        return
    def listTransferRequest(self,root):
        """A message containing an event, which does not solicit a reply."""
        e = self._grabEvent(root)
        print "1--"
        return listTransferReply(event=e)
    def completeTransferRequest(self,root):
        """A message containing an event, which does not solicit a reply."""
        e = self._grabEvent(root)
        return completeTransferReply(event=e)

###########################################################################
#  Messages, TC struct
###########################################################################
class qRequest:
      def __init__(self, *l):
          self.handle = l[0]
          try:
              self.v2 = l[1]
              self.v3 = l[2]
              self.v4 = l[3]
          except IndexError, args:
              pass
          return
class listTransferReply(Message):
    """Request for data from a consumer to a producer.
    
    """
    def __init__(self,event,filter=None):
        """Create query request with an Event object.
        """
        Message.__init__(self)
        self.event = EventMessage(event)
        if filter is not None:
            self.filter = EventFilter(filter)
        else:
            self.filter = None
        print "2--"
        return
    def build(self,doc=None):
        """Build a query request.
        """
        Message.build(self)
        self.addRoot()
        self.addMessages((self.event,self.filter))
        return
class completeTransferReply(Message):
    """Request for data from a consumer to a producer.
    
    """
    def __init__(self,event,filter=None):
        """Create query request with an Event object.
        """
        Message.__init__(self)
        self.event = EventMessage(event)
        if filter is not None:
            self.filter = EventFilter(filter)
        else:
            self.filter = None
        return
    def build(self,doc=None):
        """Build a query request.
        """
        Message.build(self)
        self.addRoot()
        self.addMessages((self.event,self.filter))
        return
################################################################################
# TestCases
################################################################################
class RFTSoapServer(GSITCPSocketServer):
    def __init__(self,addr, requesthandler):
        """Start up the server."""
        GSITCPSocketServer.__init__(self, addr, requesthandler)
        self._factory = RFTMessageFactory()
        return
    def set_factory(self, msgFactory):
        self._factory = msgFactory()
        return
    def close(self):
        self.server_close()
        return
        
class testServer(Thread):
    def setup(self, port):
        self.handle = RFTSoapServer(('localhost',g_port), RFTRequestHandler)
        return self.handle
    def run(self):
        while 1:
            try:
                self.handle.handle_request()
            except GSITCPSocketException, args:
                print args
                break
        return
    def server_close(self):
        self.handle.server_close()
        return
class testClient:
    def run(self):
        kw = {'ssl':2, 'namespace':'urn:ReliableFM',
              'host':'localhost', 'port':g_port, 'tracefile':sys.stdout}
	try:
            self.client = GSIBinding(**kw)
            handle = None
	except GlobusException, ex:
            print "caught an exception ", ex.args
            sys.exit(-1)
        return
    def doCommand(self, cmd, **kv):
        tcs = []
        for k,v in kv.items():
            if type(v) == int:
                tcs += [TC.Integer(k)]
            if type(v) == str:
                tcs += [TC.String(k)]
        qRequest.typecode=TC.Struct(qRequest,
                                    tcs,
                                    cmd,
                                    inline=1)
        try:
            result_list = self.client.RPC(None,
                                          cmd,
                                          apply(qRequest, tuple(kv.values())),
                                          TC.Any(aslist=1))
            result = result_list[0]
        except Exception, ex:
            print 'reply=', self.client.reply_code
            print 'reply_msg=', self.client.reply_msg
            print 'headers=', self.client.reply_headers
            print 'data=', self.client.data
            raise
	return
class testZSICase(unittest.TestCase):
    def setUp(self):
        print "Begin GSITCPSocketServer"
        self._server = testServer()
        self._client = testClient()
        self._server.setup(g_port)
        time.sleep(1)
        self._server.start()
        time.sleep(1)
        self._client.run()
    def check_int(self):
        vals = {'handle':333}
        self._client.doCommand('listTransferRequest', **vals)
        time.sleep(2)
    def check_str(self):
        vals = {'handle':'333'}
        self._client.doCommand('listTransferRequest', **vals)
        time.sleep(2)
    def check_val(self):
        vals = {'handle':333}
        self._client.doCommand('listTransferRequest', **vals)
        time.sleep(2)
    def tearDown(self):
        self._server.server_close()
        print "done with GSITCPSocketServer"

def makeTestSuite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(testZSICase, "check_"))
    return suite

def main():
   try:
        opt, arg = getopt.getopt(sys.argv[1:], 'hs')
   except getopt.error:
       print "Bad command-line arg [-h][-s]"
       sys.exit(-1)
   if opt:
        for o,a in opt:
            if o == '-h':
                #soap_proxy = socket.gethostbyaddr(a)
                client = testClient()
            elif o == '-s':
                server = testServer()
            else:
                print 'what is opt: ', o
                sys.exit(-1)
        try:
            vals = {'handle':333}
            client.run()
            client.doCommand('listTransferRequest', **vals)
        except Exception, ex:
            print ex
            server.setup(g_port)
            time.sleep(1)
            server.start()
   else:
        unittest.main(defaultTest="makeTestSuite")

if __name__ == "__main__" : main()
