/*############################################################################
# Keith R. Jackson, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################*/
%module ftpClientc
%{
#include "globus_common.h"
#include "globus_ftp_client.h"
#include "globus_ftp_client_debug_plugin.h"
#include "globus_ftp_client_restart_marker_plugin.h"
#include "globus_ftp_client_perf_plugin.h"
#include "globus_ftp_client_throughput_plugin.h"
/*
#if defined(GLOBUS_BUILD_WITH_NETLOGGER)
#include "globus_ftp_client_throughput_nl_plugin.h"
#endif
*/
%}

%include ./src/pyGlobus-typemaps.i

%types(globus_off_t*, globus_abstime_t*);

%typemap(python, in) FILE * {
    if (!PyFile_Check($input)) {
        PyErr_SetString(PyExc_TypeError, "expected PyFile");
        return NULL;
    }
    $1=PyFile_AsFile($input);
}

%apply char **OUTPUT { char **usr, char** passwd, char **account,
                       char **subject, char** markerString };

typedef int globus_bool_t;
typedef unsigned char globus_byte_t;

%{



typedef struct {
    PyObject* callback;
    PyObject* arg;
} callbackStruct;

typedef struct {
     PyObject* cb1;
     PyObject* cb2;
     PyObject* cb3;
     PyObject* arg;
 } callbackStruct2;

typedef struct {
     PyObject* cb1;
     PyObject* cb2;
     PyObject* cb3;
     PyObject* cb4;
     PyObject* arg;
 } callbackStruct3;

typedef enum {
    INSTANTIATE,
    DESTROY,
    GET,
    PUT,
    TRANSFER,
    ABORT,
    CONNECT,
    AUTH,
    READ,
    WRITE,
    DATA,
    COMMAND,
    RESPONSE,
    FAULT,
    PLUGIN_SPECIFIC
} pythonPluginList;

static PyThreadState *thread_state = NULL;


static void PythonCompleteCallback(void* user_data, 
                                   globus_ftp_client_handle_t* handle,
                                   globus_object_t* error) {
    
    PyObject *func, *arglist, *result, *arg, *ptr;
    char* err;
    callbackStruct* cbStruct = (callbackStruct*)user_data;
    PyEval_AcquireThread(thread_state);
    //printf("IN CALLBACK\n");
    func = cbStruct->callback;
    arg = cbStruct->arg;
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_client_handle_t, 0);
    
    if (error != GLOBUS_SUCCESS) {
        err = globus_object_printable_to_string(error);
    }
    else {
        err = "NONE";
    }
    
    arglist = Py_BuildValue("(OOs)", arg, ptr, err);
    //printf("About to run callback\n");
    result = PyEval_CallObject(func, arglist);
    //printf("Ran callback\n");
    Py_DECREF(arglist);
    Py_XDECREF(result);
    if (error != GLOBUS_SUCCESS)
        globus_free(err);
    //printf("Releasing interp thread_state\n");
    PyEval_ReleaseThread(thread_state);
    //printf("Leaving CAllback\n");
    return;
}


static void PythonDataCallback(void* user_data,
                               globus_ftp_client_handle_t* handle,
                               globus_object_t* error,
                               globus_byte_t* buffer,
                               globus_size_t length,
                               globus_off_t offset,
                               globus_bool_t eof) {
    PyObject *func, *arglist, *result, *arg, *buf, *oSet, *ptr, *ptr2;
    char* err;
    callbackStruct* cbStruct;
    PyEval_AcquireThread(thread_state);
    cbStruct = (callbackStruct*)user_data;
    func = cbStruct->callback;
    arg = cbStruct->arg;
    if (error != GLOBUS_SUCCESS) {
        err = globus_object_printable_to_string(error);
    }
    else {
        err = "NONE";
    }
    if (sizeof(globus_off_t) == 8) {
        /* XXX, This assumes globus_off_t is a long long.  This code will
           probably have to be ifdef'd by platform to account for 
           different types for size_t */
        oSet = PyLong_FromLongLong(offset);
    }
    else {
        oSet = PyInt_FromLong(offset);
    }
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_client_handle_t, 0);
    ptr2 = SWIG_NewPointerObj((void*)buffer, SWIGTYPE_p_globus_byte_t, 0);
    buf = PyBuffer_FromReadWriteMemory((void*)buffer, length * sizeof(globus_byte_t));
    arglist = Py_BuildValue("(OOOOiOis)", arg, ptr, buf, ptr2, length, 
                            oSet, (int)eof, err);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    if (error != GLOBUS_SUCCESS)
        globus_free(err);
    PyEval_ReleaseThread(thread_state);
    return;
}

// ------------------------------------------------------------------------------------
//  RestartMarker Callbacks
//
//
// ------------------------------------------------------------------------------------
static void PythonRestartMarkerPluginCompleteCallback(void* user_data, 
                                                      globus_ftp_client_handle_t* handle,
                                                      globus_object_t* error,
                                                      const char* url) {
    PyObject *func, *arglist, *result, *arg, *ptr;
    char* err;
    callbackStruct2* cbStruct = (callbackStruct2*)user_data;
    PyEval_AcquireThread(thread_state);
    func = cbStruct->cb3;
    arg = cbStruct->arg;
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_client_handle_t, 0);
    if (error != GLOBUS_SUCCESS) {
        err = globus_object_printable_to_string(error);
    }
    else {
        err = "NONE";
    }
    arglist = Py_BuildValue("(OOss)", arg, ptr, err, url);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    if (error != GLOBUS_SUCCESS)
        globus_free(err);
    PyEval_ReleaseThread(thread_state);
    return;
}

globus_bool_t PythonRestartMarkerPluginBeginCallback(void* user_data, 
                                                     globus_ftp_client_handle_t *handle,
                                                     const char* srcUrl,
                                                     const char* destUrl,
                                                     globus_ftp_client_restart_marker_t* marker) {
    PyObject *func, *arglist, *result, *arg, *ptr, *markerPtr;
    globus_bool_t ret;
    callbackStruct2* cbStruct = (callbackStruct2*)user_data;
    PyEval_AcquireThread(thread_state);    
    func = cbStruct->cb1;
    arg = cbStruct->arg;
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_client_handle_t, 0);
    markerPtr = SWIG_NewPointerObj((void*)marker, 
                                   SWIGTYPE_p_globus_ftp_client_restart_marker_t, 0);
    arglist = Py_BuildValue("(OOssO)", arg, ptr, srcUrl, destUrl, markerPtr);
    result = PyEval_CallObject(func, arglist);
    if (PyInt_AsLong(result) == GLOBUS_SUCCESS) {
        ret = GLOBUS_SUCCESS;
    }
    else {
        ret = GLOBUS_FALSE;
    }
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return ret;
}

void PythonRestartMarkerPluginMarkerCallback(void* user_data, 
                                            globus_ftp_client_handle_t *handle,
                                            globus_ftp_client_restart_marker_t* marker) {
    PyObject *func, *arglist, *result, *arg, *ptr, *markerPtr;
    callbackStruct2* cbStruct = (callbackStruct2*)user_data;
    PyEval_AcquireThread(thread_state);
    func = cbStruct->cb2;
    arg = cbStruct->arg;
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_client_handle_t, 0);
    markerPtr = SWIG_NewPointerObj((void*)marker, 
                                   SWIGTYPE_p_globus_ftp_client_restart_marker_t, 0);
    arglist = Py_BuildValue("(OOO)", arg, ptr, markerPtr);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}
// ------------------------------------------------------------------------------------
//  Performance Callbacks
//
//
// ------------------------------------------------------------------------------------
static void 
PythonPerfMarkerPluginCompleteCallback(void* user_data, 
						   globus_ftp_client_handle_t* handle,
						   globus_bool_t success) {
    PyObject *func, *arglist, *result, *arg, *ptr;
    char* err;
    callbackStruct2* cbStruct = (callbackStruct2*)user_data;
    PyEval_AcquireThread(thread_state);
    func = cbStruct->cb3;
    arg = cbStruct->arg;
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_client_handle_t, 0);
    if (success) {
        err = "NONE";
    }
    else {
	err = "Transfer Interrupted";
    }
        arglist = Py_BuildValue("(OOs)", arg, ptr, err);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    if (success != GLOBUS_SUCCESS)
        globus_free(err);
    PyEval_ReleaseThread(thread_state);
    return;
}

static void 
PythonPerfMarkerPluginBeginCallback(void* user_data, 
					 globus_ftp_client_handle_t *handle,
					 const char* srcUrl,
					 const char* destUrl,
					 globus_bool_t restart) {
    PyObject *func, *arglist, *result, *arg, *ptr;
    globus_bool_t ret;
    callbackStruct2* cbStruct = (callbackStruct2*)user_data;
    PyEval_AcquireThread(thread_state);
    func = cbStruct->cb1;
    arg = cbStruct->arg;
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_client_handle_t, 0);
    arglist = Py_BuildValue("(OOssi)", arg, ptr, srcUrl, destUrl, (int)restart);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}

static void 
PythonPerfMarkerPluginMarkerCallback(void* user_data, 
					  globus_ftp_client_handle_t *handle,
					  long time_stamp_int,
					  char time_stamp_length,
					  int strip_ndx,
					  int num_stripes,
					  globus_off_t nbytes) {
    PyObject *func, *arglist, *result, *arg, *ptr, *tBytesPtr;
    callbackStruct2* cbStruct = (callbackStruct2*)user_data;
    PyEval_AcquireThread(thread_state);
    func = cbStruct->cb2;
    arg = cbStruct->arg;
    if (sizeof(globus_off_t) == 8) {
        /* XXX, This assumes globus_off_t is a long long.  This code will
           probably have to be ifdef'd by platform to account for 
           different types for size_t */
	tBytesPtr = PyLong_FromLongLong(nbytes);
    }
    else {
        tBytesPtr = PyInt_FromLong(nbytes);
    }
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_client_handle_t, 0);
    arglist = Py_BuildValue("(OOlciiO)", arg, ptr, time_stamp_int, time_stamp_length, 
    			    strip_ndx, num_stripes, tBytesPtr);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}

// ------------------------------------------------------------------------------------
//  Throughput Callbacks
//
//
// ------------------------------------------------------------------------------------
void PythonThroughputPluginBeginCallback(void *user_data,
					 globus_ftp_client_handle_t *handle,
					 const char *source_url, 
					 const char *dest_url) {
    PyObject *func, *arglist, *result, *arg, *ptr;
    globus_bool_t ret;
    callbackStruct3* cbStruct = (callbackStruct3*)user_data;
    PyEval_AcquireThread(thread_state);
    func = cbStruct->cb1;
    arg = cbStruct->arg;
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_client_handle_t, 0);
    arglist = Py_BuildValue("(OOss)", arg, ptr, source_url, dest_url);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}
void PythonThroughputPluginStripeCallback(void *user_data, 
					  globus_ftp_client_handle_t *handle,
					  int stripe_ndx,
					  globus_off_t nbytes,
					  float instantaneous_throughput,
					  float avg_throughput) {
    PyObject *func, *arglist, *result, *arg, *ptr, *tBytesPtr;
    callbackStruct3* cbStruct = (callbackStruct3*)user_data;
    PyEval_AcquireThread(thread_state);
    func = cbStruct->cb2;
    arg = cbStruct->arg;
    if (sizeof(globus_off_t) == 8) {
        /* XXX, This assumes globus_off_t is a long long.  This code will
           probably have to be ifdef'd by platform to account for 
           different types for size_t */
	tBytesPtr = PyLong_FromLongLong(nbytes);
    }
    else {
        tBytesPtr = PyInt_FromLong(nbytes);
    }
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_client_handle_t, 0);
    arglist = Py_BuildValue("(OOiOff)", arg, ptr, stripe_ndx, tBytesPtr, 
			    instantaneous_throughput, avg_throughput);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}
void PythonThroughputPluginTotalCallback(void *user_data,
					 globus_ftp_client_handle_t *handle,
					 globus_off_t nbytes,
					 float instantaneous_throughput,
					 float avg_throughput) {
    PyObject *func, *arglist, *result, *arg, *ptr, *tBytesPtr;
    callbackStruct3* cbStruct = (callbackStruct3*)user_data;
    PyEval_AcquireThread(thread_state);
    func = cbStruct->cb3;
    arg = cbStruct->arg;
    if (sizeof(globus_off_t) == 8) {
        /* XXX, This assumes globus_off_t is a long long.  This code will
           probably have to be ifdef'd by platform to account for 
           different types for size_t */
	tBytesPtr = PyLong_FromLongLong(nbytes);
    }
    else {
        tBytesPtr = PyInt_FromLong(nbytes);
    }
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_client_handle_t, 0);
    arglist = Py_BuildValue("(OOOff)", arg, ptr, tBytesPtr,
			    instantaneous_throughput, avg_throughput);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}
void PythonThroughputPluginCompleteCallback(void *user_data,
					    globus_ftp_client_handle_t *handle,
					    globus_bool_t success) {
    PyObject *func, *arglist, *result, *arg, *ptr;
    char* err;
    callbackStruct3* cbStruct = (callbackStruct3*)user_data;
    PyEval_AcquireThread(thread_state);
    func = cbStruct->cb4;
    arg = cbStruct->arg;
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_client_handle_t, 0);
    if (success) {
        err = "NONE";
    }
    else {
        err = "Transfer Interrupted";
    }
    arglist = Py_BuildValue("(OOs)", arg, ptr, err);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    if (success != GLOBUS_SUCCESS)
        globus_free(err);
    PyEval_ReleaseThread(thread_state);
    return;
}

/*
static globus_ftp_client_plugin_t*
PythonPluginInstantiate(globus_ftp_client_plugin_t* self, 
                 globus_ftp_client_handle_t* handle) {
    char ptemp[128];
    char ptemp1[128];
    PyObject *func, *list, *result, *arglist;
    PyEval_AcquireThread(thread_state);
    list = self->plugin_specific;
    func = PyList_GetItem(list, INSTANTIATE);
    if (func == NULL)
        return NULL;
    SWIG_NewPointerObj(ptemp, (void*)self, "_globus_ftp_client_plugin_t_p");
    SWIG_NewPointerObj(ptemp1, (void*)handle, "_globus_ftp_client_handle_t_p");
    arglist = Py_BuildValue("(ss)", ptemp, ptemp1);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    if (result == NULL)
        return NULL;
    PyEval_ReleaseThread(thread_state);
    return self;
}

static void
PythonPluginDestroy(globus_ftp_client_plugin_t* self, 
                    globus_ftp_client_handle_t* handle) {
    char ptemp[128];
    char ptemp1[128];
    PyObject *func, *list, *result, *arglist;
    PyEval_AcquireThread(thread_state);
    list = self->plugin_specific;
    func = PyList_GetItem(list, DESTROY);
    SWIG_NewPointerObj(ptemp, (void*)self, "_globus_ftp_client_plugin_t_p");
    SWIG_NewPointerObj(ptemp1, (void*)handle, "_globus_ftp_client_handle_t_p");
    arglist = Py_BuildValue("(ss)", ptemp, ptemp1);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}

static void
PythonPluginGet(globus_ftp_client_plugin_t* self, 
                globus_ftp_client_handle_t* handle, const char* url,
                const globus_ftp_client_attr_t * attr, globus_bool_t restart) {
    char ptemp[128];
    char ptemp1[128];
    char ptemp2[128];
    PyObject *func, *list, *result, *arglist;
    PyEval_AcquireThread(thread_state);
    list = self->plugin_specific;
    func = PyList_GetItem(list, GET);
    SWIG_NewPointerObj(ptemp, (void*)self, "_globus_ftp_client_plugin_t_p");
    SWIG_NewPointerObj(ptemp1, (void*)handle, "_globus_ftp_client_handle_t_p");
    SWIG_NewPointerObj(ptemp2, (void*)attr, "_globus_ftp_client_attr_t_p");
    arglist = Py_BuildValue("(ssssi)", ptemp, ptemp1, url, ptemp2, (int)restart);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}

static void
PythonPluginPut(globus_ftp_client_plugin_t* self, 
                globus_ftp_client_handle_t* handle, const char* url,
                const globus_ftp_client_attr_t * attr, globus_bool_t restart) {
    char ptemp[128];
    char ptemp1[128];
    char ptemp2[128];
    PyObject *func, *list, *result, *arglist;
    PyEval_AcquireThread(thread_state);
    list = self->plugin_specific;
    func = PyList_GetItem(list, PUT);
    SWIG_NewPointerObj(ptemp, (void*)self, "_globus_ftp_client_plugin_t_p");
    SWIG_NewPointerObj(ptemp1, (void*)handle, "_globus_ftp_client_handle_t_p");
    SWIG_NewPointerObj(ptemp2, (void*)attr, "_globus_ftp_client_attr_t_p");
    arglist = Py_BuildValue("(ssssi)", ptemp, ptemp1, url, ptemp2, (int)restart);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}

static void
PythonPluginTransfer(globus_ftp_client_plugin_t* self, 
                     globus_ftp_client_handle_t* handle, const char* srcUrl,
                     const globus_ftp_client_attr_t * srcAttr, 
                     const char* dstUrl, 
                     const globus_ftp_client_attr_t* dstAttr, 
                     globus_bool_t restart) {
    char ptemp[128];
    char ptemp1[128];
    char ptemp2[128];
    char ptemp3[128];
    PyObject *func, *list, *result, *arglist;
    PyEval_AcquireThread(thread_state);
    list = self->plugin_specific;
    func = PyList_GetItem(list, TRANSFER);
    SWIG_NewPointerObj(ptemp, (void*)self, "_globus_ftp_client_plugin_t_p");
    SWIG_NewPointerObj(ptemp1, (void*)handle, "_globus_ftp_client_handle_t_p");
    SWIG_NewPointerObj(ptemp2, (void*)srcAttr, "_globus_ftp_client_attr_t_p");
    SWIG_NewPointerObj(ptemp3, (void*)dstAttr, "_globus_ftp_client_attr_t_p");
    arglist = Py_BuildValue("(ssssssi)", ptemp, ptemp1, srcUrl, ptemp2, 
                            dstUrl, ptemp3, (int)restart);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}

static void
PythonPluginAbort(globus_ftp_client_plugin_t* self, 
                  globus_ftp_client_handle_t* handle) {
    char ptemp[128];
    char ptemp1[128];
    PyObject *func, *list, *result, *arglist;
    PyEval_AcquireThread(thread_state);
    list = self->plugin_specific;
    func = PyList_GetItem(list, ABORT);
    SWIG_NewPointerObj(ptemp, (void*)self, "_globus_ftp_client_plugin_t_p");
    SWIG_NewPointerObj(ptemp1, (void*)handle, "_globus_ftp_client_handle_t_p");
    arglist = Py_BuildValue("(ss)", ptemp, ptemp1);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}

static void
PythonPluginConnect(globus_ftp_client_plugin_t* self, 
                    globus_ftp_client_handle_t* handle,
                    const globus_url_t* url) {
    char ptemp[128];
    char ptemp1[128];
    char ptemp2[128];
    PyObject *func, *list, *result, *arglist;
    PyEval_AcquireThread(thread_state);
    list = self->plugin_specific;
    func = PyList_GetItem(list, CONNECT);
    SWIG_NewPointerObj(ptemp, (void*)self, "_globus_ftp_client_plugin_t_p");
    SWIG_NewPointerObj(ptemp1, (void*)handle, "_globus_ftp_client_handle_t_p");
    SWIG_NewPointerObj(ptemp2, (void*)url, "_globus_url_t_p");
    arglist = Py_BuildValue("(sss)", ptemp, ptemp1, ptemp2);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}

static void
PythonPluginAuth(globus_ftp_client_plugin_t* self, 
                 globus_ftp_client_handle_t* handle,
                 const globus_url_t* url,
                 const globus_ftp_control_auth_info_t* authInfo) {
    char ptemp[128];
    char ptemp1[128];
    char ptemp2[128];
    char ptemp3[128];
    PyObject *func, *list, *result, *arglist;
    PyEval_AcquireThread(thread_state);
    list = self->plugin_specific;
    func = PyList_GetItem(list, AUTH);
    SWIG_NewPointerObj(ptemp, (void*)self, "_globus_ftp_client_plugin_t_p");
    SWIG_NewPointerObj(ptemp1, (void*)handle, "_globus_ftp_client_handle_t_p");
    SWIG_NewPointerObj(ptemp2, (void*)url, "_globus_url_t_p");
    SWIG_NewPointerObj(ptemp3, (void*)authInfo, "_globus_ftp_control_auth_info_t_p");
    arglist = Py_BuildValue("(ssss)", ptemp, ptemp1, ptemp2, ptemp3);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}

static void
PythonPluginRead(globus_ftp_client_plugin_t* self, 
                 globus_ftp_client_handle_t* handle,
                 const globus_byte_t* buffer,
                 globus_size_t bufferLen) {
    char ptemp[128];
    char ptemp1[128];
    char ptemp2[128];
    PyObject *func, *list, *result, *arglist, *length;
    PyEval_AcquireThread(thread_state);
    list = self->plugin_specific;
    func = PyList_GetItem(list, READ);
    SWIG_NewPointerObj(ptemp, (void*)self, "_globus_ftp_client_plugin_t_p");
    SWIG_NewPointerObj(ptemp1, (void*)handle, "_globus_ftp_client_handle_t_p");
    SWIG_NewPointerObj(ptemp2, (void*)buffer, "_globus_byte_t_p");
    if (sizeof(globus_size_t) == 8) {
        /* XXX, This assumes size_t is a long long.  This code will
           probably have to be ifdef'd by platform to account for 
           different types for size_t */
/*
        length = PyLong_FromLongLong(bufferLen);
    } else {
        length = PyInt_FromLong(bufferLen);
    }
    arglist = Py_BuildValue("(sssO)", ptemp, ptemp1, ptemp2, length);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}

static void
PythonPluginWrite(globus_ftp_client_plugin_t* self, 
                  globus_ftp_client_handle_t* handle,
                  const globus_byte_t* buffer,
                  globus_size_t bufferLen,
                  globus_off_t offset, globus_bool_t eof) {
    char ptemp[128];
    char ptemp1[128];
    char ptemp2[128];
    PyObject *func, *list, *result, *arglist, *length, *oset;
    PyEval_AcquireThread(thread_state);
    list = self->plugin_specific;
    func = PyList_GetItem(list, WRITE);
    SWIG_NewPointerObj(ptemp, (void*)self, "_globus_ftp_client_plugin_t_p");
    SWIG_NewPointerObj(ptemp1, (void*)handle, "_globus_ftp_client_handle_t_p");
    SWIG_NewPointerObj(ptemp2, (void*)buffer, "_globus_byte_t_p");
    if (sizeof(globus_size_t) == 8) {
        /* XXX, This assumes size_t is a long long.  This code will
           probably have to be ifdef'd by platform to account for 
           different types for size_t */
/*  
      length = PyLong_FromLongLong(bufferLen);
    } else {
        length = PyInt_FromLong(bufferLen);
    }
    if (sizeof(globus_off_t) == 8) {
        /* XXX, This assumes size_t is a long long.  This code will
           probably have to be ifdef'd by platform to account for 
           different types for size_t */
/*
        oset = PyLong_FromLongLong(offset);
    } else {
        oset = PyInt_FromLong(offset);
    }
    arglist = Py_BuildValue("(sssOOi)", ptemp, ptemp1, ptemp2, 
                            length, oset, (int)eof);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}

static void
PythonPluginData(globus_ftp_client_plugin_t* self, 
                 globus_ftp_client_handle_t* handle,
                 globus_object_t* error, 
                 const globus_byte_t* buffer,
                 globus_size_t bufferLen,
                 globus_off_t offset, globus_bool_t eof) {
    char ptemp[128];
    char ptemp1[128];
    char ptemp2[128];
    char* tmpstr;
    PyObject *func, *list, *result, *arglist, *length, *oset;
    PyEval_AcquireThread(thread_state);
    list = self->plugin_specific;
    func = PyList_GetItem(list, DATA);
    SWIG_NewPointerObj(ptemp, (void*)self, "_globus_ftp_client_plugin_t_p");
    SWIG_NewPointerObj(ptemp1, (void*)handle, "_globus_ftp_client_handle_t_p");
    SWIG_NewPointerObj(ptemp2, (void*)buffer, "_globus_byte_t_p");
    if (sizeof(globus_size_t) == 8) {
        /* XXX, This assumes size_t is a long long.  This code will
           probably have to be ifdef'd by platform to account for 
           different types for size_t */
/*
        length = PyLong_FromLongLong(bufferLen);
    } else {
        length = PyInt_FromLong(bufferLen);
    }
    if (sizeof(globus_off_t) == 8) {
        /* XXX, This assumes size_t is a long long.  This code will
           probably have to be ifdef'd by platform to account for 
           different types for size_t */
/*
        oset = PyLong_FromLongLong(offset);
    } else {
        oset = PyInt_FromLong(offset);
    }
    if (error != GLOBUS_SUCCESS) 
        tmpstr = globus_object_printable_to_string(globus_error_get(error));
    else
        tmpstr = "";
    arglist = Py_BuildValue("(ssssOOi)", ptemp, ptemp1, tmpstr, ptemp2, 
                            length, oset, (int)eof);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}

static void
PythonPluginCommand(globus_ftp_client_plugin_t* self, 
                    globus_ftp_client_handle_t* handle,
                    const globus_url_t* url, const char* command) {
    char ptemp[128];
    char ptemp1[128];
    char ptemp2[128];
    PyObject *func, *list, *result, *arglist;
    PyEval_AcquireThread(thread_state);
    list = self->plugin_specific;
    func = PyList_GetItem(list, COMMAND);
    SWIG_NewPointerObj(ptemp, (void*)self, "_globus_ftp_client_plugin_t_p");
    SWIG_NewPointerObj(ptemp1, (void*)handle, "_globus_ftp_client_handle_t_p");
    SWIG_NewPointerObj(ptemp2, (void*)url, "_globus_url_t_p");
    arglist = Py_BuildValue("(ssss)", ptemp, ptemp1, ptemp2, command);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}

static void
PythonPluginResponse(globus_ftp_client_plugin_t* self, 
                     globus_ftp_client_handle_t* handle,
                     const globus_url_t* url, globus_object_t* error,
                     const globus_ftp_control_response_t* response) {
    char ptemp[128];
    char ptemp1[128];
    char ptemp2[128];
    char ptemp3[128];
    char* tmpstr;
    PyObject *func, *list, *result, *arglist;
    PyEval_AcquireThread(thread_state);
    list = self->plugin_specific;
    func = PyList_GetItem(list, RESPONSE);
    SWIG_NewPointerObj(ptemp, (void*)self, "_globus_ftp_client_plugin_t_p");
    SWIG_NewPointerObj(ptemp1, (void*)handle, "_globus_ftp_client_handle_t_p");
    SWIG_NewPointerObj(ptemp2, (void*)url, "_globus_url_t_p");
    if (error != GLOBUS_SUCCESS) 
        tmpstr = globus_object_printable_to_string(globus_error_get(error));
    else
        tmpstr = "";
    SWIG_NewPointerObj(ptemp3, (void*)response, "_globus_ftp_control_response_t_p");
    arglist = Py_BuildValue("(sssss)", ptemp, ptemp1, ptemp2, tmpstr, ptemp3);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}

static void
PythonPluginFault(globus_ftp_client_plugin_t* self, 
                  globus_ftp_client_handle_t* handle,
                  const globus_url_t* url, globus_object_t* error) {
    char ptemp[128];
    char ptemp1[128];
    char ptemp2[128];
    char* tmpstr;
    PyObject *func, *list, *result, *arglist;
    PyEval_AcquireThread(thread_state);
    list = self->plugin_specific;
    func = PyList_GetItem(list, FAULT);
    SWIG_NewPointerObj(ptemp, (void*)self, "_globus_ftp_client_plugin_t_p");
    SWIG_NewPointerObj(ptemp1, (void*)handle, "_globus_ftp_client_handle_t_p");
    SWIG_NewPointerObj(ptemp2, (void*)url, "_globus_url_t_p");
    if (error != GLOBUS_SUCCESS) 
        tmpstr = globus_object_printable_to_string(globus_error_get(error));
    else
        tmpstr = "";
    arglist = Py_BuildValue("(ssss)", ptemp, ptemp1, ptemp2, tmpstr);
    result = PyEval_CallObject(func, arglist);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return;
}
*/
%}

%inline %{
globus_module_descriptor_t*
get_module(void) {
    return GLOBUS_FTP_CLIENT_MODULE;
}

globus_module_descriptor_t*
get_debug_plugin_module(void) {
    return GLOBUS_FTP_CLIENT_DEBUG_PLUGIN_MODULE;
}

globus_module_descriptor_t*
get_restart_marker_plugin_module(void) {
    return GLOBUS_FTP_CLIENT_RESTART_MARKER_PLUGIN_MODULE;
}
globus_module_descriptor_t*
get_performance_marker_plugin_module(void) {
    return GLOBUS_FTP_CLIENT_PERF_PLUGIN_MODULE;
}
globus_module_descriptor_t*
get_throughput_plugin_module(void) {
    return GLOBUS_FTP_CLIENT_THROUGHPUT_PLUGIN_MODULE;
}
/*
globus_module_descriptor_t*
get_netlogger_plugin_module(void) {
    return GLOBUS_FTP_CLIENT_RESTART_MARKER_PLUGIN_MODULE;
}
*/
PyObject*
handleattr_init() {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_ftp_client_handleattr_t* attr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_init: PyTuple_New()");
        return NULL;
    }
    attr = (globus_ftp_client_handleattr_t*)globus_malloc(sizeof(globus_ftp_client_handleattr_t));
    ret = globus_ftp_client_handleattr_init(attr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(attr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)attr, 
				 SWIGTYPE_p_globus_ftp_client_handleattr_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
handleattr_copy(globus_ftp_client_handleattr_t* srcAttr) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_ftp_client_handleattr_t* attr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_copy: PyTuple_New()");
        return NULL;
    }
    attr = (globus_ftp_client_handleattr_t*)globus_malloc(sizeof(globus_ftp_client_handleattr_t));
    ret = globus_ftp_client_handleattr_copy(attr, srcAttr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(attr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)attr, 
				 SWIGTYPE_p_globus_ftp_client_handleattr_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
handleattr_destroy(globus_ftp_client_handleattr_t* attr) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_destroy: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_handleattr_destroy(attr);
    globus_free(attr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
handleattr_add_cached_url(globus_ftp_client_handleattr_t* attr,
                          const char* url) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_add_cached_url: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_handleattr_add_cached_url(attr, url);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
handleattr_remove_cached_url(globus_ftp_client_handleattr_t* attr,
                          const char* url) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_remove_cached_url: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_handleattr_remove_cached_url(attr, url);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
handleattr_set_cache_all(globus_ftp_client_handleattr_t* attr,
                         globus_bool_t cacheAll) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_set_cache_all: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_handleattr_set_cache_all(attr, cacheAll);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
handleattr_get_cache_all(globus_ftp_client_handleattr_t* attr,
                         globus_bool_t* boolRet) {
    globus_result_t ret;
    PyObject* list;
    char* tmpstr;
    list = PyList_New(0);
    ret = globus_ftp_client_handleattr_get_cache_all(attr, boolRet);
    if (ret != GLOBUS_SUCCESS) {
        PyList_Append(list, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyList_Append(list, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyList_Append(list, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyList_Append(list, Py_None);
    }
    return list;
}
/*
PyObject*
handleattr_set_netlogger(globus_ftp_client_handleattr_t* attr,
                         globus_netlogger_handle_t* nl_handle) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_set_netlogger: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_handleattr_set_netlogger(attr, nl_handle);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
}
*/
PyObject*
handleattr_add_plugin(globus_ftp_client_handleattr_t* handleattr,
                  globus_ftp_client_plugin_t* plugin) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_add_plugin: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_handleattr_add_plugin(handleattr, plugin);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
handleattr_remove_plugin(globus_ftp_client_handleattr_t* handleattr,
                     globus_ftp_client_plugin_t* plugin) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_remove_plugin: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_handleattr_remove_plugin(handleattr, plugin);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    globus_free(handleattr);
    return tuple;
}

PyObject*
operationattr_init() {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_ftp_client_operationattr_t* attr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_init: PyTuple_New()");
        return NULL;
    }
    attr = (globus_ftp_client_operationattr_t*)globus_malloc(sizeof(globus_ftp_client_operationattr_t));
    ret = globus_ftp_client_operationattr_init(attr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(attr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)attr, 
				 SWIGTYPE_p_globus_ftp_client_operationattr_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
operationattr_destroy(globus_ftp_client_operationattr_t* attr) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_destroy: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_operationattr_destroy(attr);
    globus_free(attr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
operationattr_copy(const globus_ftp_client_operationattr_t* srcAttr) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_ftp_client_operationattr_t* attr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_copy: PyTuple_New()");
        return NULL;
    }
    attr = (globus_ftp_client_operationattr_t*)globus_malloc(sizeof(globus_ftp_client_operationattr_t));
    ret = globus_ftp_client_operationattr_copy(attr, srcAttr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(attr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)attr, 
				 SWIGTYPE_p_globus_ftp_client_operationattr_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
operationattr_set_append(globus_ftp_client_operationattr_t* attr, globus_bool_t append) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_set_append: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_operationattr_set_append(attr, append);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
operationattr_get_append(globus_ftp_client_operationattr_t* attr, globus_bool_t* boolRet) {
    globus_result_t ret;
    PyObject* list;
    char* tmpstr;
    list = PyList_New(0);
    ret = globus_ftp_client_operationattr_get_append(attr, boolRet);
    if (ret != GLOBUS_SUCCESS) {
        PyList_Append(list, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyList_Append(list, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyList_Append(list, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyList_Append(list, Py_None);
    }
    return list;
}

PyObject*
operationattr_set_resume_third_party_transfer(globus_ftp_client_operationattr_t* attr, 
                                              globus_bool_t resume) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, 
                       "operationattr_set_resume_third_party_transfer: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_operationattr_set_resume_third_party_transfer(attr, resume);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
operationattr_get_resume_third_party_transfer(globus_ftp_client_operationattr_t* attr, 
                                                          globus_bool_t* boolRet) {
    globus_result_t ret;
    PyObject* list;
    char* tmpstr;
    list = PyList_New(0);
    ret = globus_ftp_client_operationattr_get_resume_third_party_transfer(attr, boolRet);
    if (ret != GLOBUS_SUCCESS) {
        PyList_Append(list, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyList_Append(list, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyList_Append(list, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyList_Append(list, Py_None);
    }
    return list;
}

PyObject*
operationattr_set_authorization(globus_ftp_client_operationattr_t* attr, 
                  gss_cred_id_t* credential, PyObject* usr,
                  const char* passwd, const char* account, 
                  const char* subject) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    char* user;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_set_authorization: PyTuple_New()");
        return NULL;
    }
    if (usr == Py_None) {
        user = GLOBUS_NULL;
        passwd = GLOBUS_NULL;
        subject = GLOBUS_NULL;
    }
    else {
        user = PyString_AsString(usr);
    }
    ret = globus_ftp_client_operationattr_set_authorization(attr, *credential,
                                                user, passwd, account, subject);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
operationattr_get_authorization(globus_ftp_client_operationattr_t* attr, 
                                gss_cred_id_t* credential, char** usr,
                                char** passwd, char** account, char** subject) {
    globus_result_t ret;
    PyObject* list;
    char* tmpstr;
    list = PyList_New(0);
    ret = globus_ftp_client_operationattr_get_authorization(attr, credential, usr, passwd,
                                                        account, subject);
    if (ret != GLOBUS_SUCCESS) {
        PyList_Append(list, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyList_Append(list, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyList_Append(list, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyList_Append(list, Py_None);
    }
    return list;
}

PyObject*
operationattr_set_dcau(globus_ftp_client_operationattr_t* attr,
                  const globus_ftp_control_dcau_t* dcau) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_set_dcau: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_operationattr_set_dcau(attr, dcau);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
operationattr_get_dcau(globus_ftp_client_operationattr_t* attr) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_ftp_control_dcau_t* dcau;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_get_dcau: PyTuple_New()");
        return NULL;
    }
    dcau = (globus_ftp_control_dcau_t*)globus_malloc(sizeof(
                                                         globus_ftp_control_dcau_t));
    ret = globus_ftp_client_operationattr_get_dcau(attr, dcau);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(dcau);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)dcau, 
				 SWIGTYPE_p_globus_ftp_control_dcau_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
operationattr_set_data_protection(globus_ftp_client_operationattr_t* attr, 
                                  char protection) {
    globus_result_t ret;
    globus_ftp_control_protection_t prt;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_set_data_protection: PyTuple_New()");
        return NULL;
    }
    if (protection == 'C')
        prt = GLOBUS_FTP_CONTROL_PROTECTION_CLEAR;
    else if (protection == 'S')
        prt = GLOBUS_FTP_CONTROL_PROTECTION_SAFE;
    else if (protection == 'E')
        prt = GLOBUS_FTP_CONTROL_PROTECTION_CONFIDENTIAL;
    else if (protection == 'P')
        prt = GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE;
    ret = globus_ftp_client_operationattr_set_data_protection(attr, prt);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
operationattr_get_data_protection(globus_ftp_client_operationattr_t* attr) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    char prt[2];
    globus_ftp_control_protection_t protection;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "get_mode: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_operationattr_get_data_protection(attr, &protection);
    if (protection == GLOBUS_FTP_CONTROL_MODE_NONE)
        prt[0] = 'C';
    else if (protection == GLOBUS_FTP_CONTROL_MODE_STREAM)
        prt[0] = 'S';
    else if (protection == GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK)
        prt[0] = 'E';
    else if (protection == GLOBUS_FTP_CONTROL_MODE_COMPRESSED)
        prt[0] = 'P';
    prt[1] = '\0';
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", prt));
        return tuple;
    }
}

PyObject*
operationattr_set_control_protection(globus_ftp_client_operationattr_t* attr, 
                                  char protection) {
    globus_result_t ret;
    globus_ftp_control_protection_t prt;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_set_control_protection: PyTuple_New()");
        return NULL;
    }
    if (protection == 'C')
        prt = GLOBUS_FTP_CONTROL_PROTECTION_CLEAR;
    else if (protection == 'S')
        prt = GLOBUS_FTP_CONTROL_PROTECTION_SAFE;
    else if (protection == 'E')
        prt = GLOBUS_FTP_CONTROL_PROTECTION_CONFIDENTIAL;
    else if (protection == 'P')
        prt = GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE;
    ret = globus_ftp_client_operationattr_set_control_protection(attr, prt);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
operationattr_get_control_protection(globus_ftp_client_operationattr_t* attr) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    char prt[2];
    globus_ftp_control_protection_t protection;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "get_mode: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_operationattr_get_control_protection(attr, &protection);
    if (protection == GLOBUS_FTP_CONTROL_MODE_NONE)
        prt[0] = 'C';
    else if (protection == GLOBUS_FTP_CONTROL_MODE_STREAM)
        prt[0] = 'S';
    else if (protection == GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK)
        prt[0] = 'E';
    else if (protection == GLOBUS_FTP_CONTROL_MODE_COMPRESSED)
        prt[0] = 'P';
    prt[1] = '\0';
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", prt));
        return tuple;
    }
}

PyObject*
operationattr_set_layout(globus_ftp_client_operationattr_t* attr,
                  const globus_ftp_control_layout_t* layout) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_set_layout: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_operationattr_set_layout(attr, layout);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
operationattr_get_layout(globus_ftp_client_operationattr_t* attr) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_ftp_control_layout_t* layout;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_get_layout: PyTuple_New()");
        return NULL;
    }
    layout = (globus_ftp_control_layout_t*)globus_malloc(sizeof(
                                                         globus_ftp_control_layout_t));
    ret = globus_ftp_client_operationattr_get_layout(attr, layout);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(layout);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)layout, 
				 SWIGTYPE_p_globus_ftp_control_layout_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
operationattr_set_mode(globus_ftp_client_operationattr_t* attr, char mode) {
    globus_result_t ret;
    globus_ftp_control_mode_t md;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_set_mode: PyTuple_New()");
        return NULL;
    }
    if (mode == '0')
        md = GLOBUS_FTP_CONTROL_MODE_NONE;
    else if (mode == 'S')
        md = GLOBUS_FTP_CONTROL_MODE_STREAM;
    else if (mode == 'B')
        md = GLOBUS_FTP_CONTROL_MODE_BLOCK;
    else if (mode == 'E')
        md = GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK;
    else if (mode == 'C')
        md = GLOBUS_FTP_CONTROL_MODE_COMPRESSED;
    ret = globus_ftp_client_operationattr_set_mode(attr, md);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
operationattr_get_mode(globus_ftp_client_operationattr_t* attr) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    char mode[2];
    globus_ftp_control_mode_t md;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "get_mode: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_operationattr_get_mode(attr, &md);
    if (md == GLOBUS_FTP_CONTROL_MODE_NONE)
        mode[0] = '0';
    else if (md == GLOBUS_FTP_CONTROL_MODE_STREAM)
        mode[0] = 'S';
    else if (md == GLOBUS_FTP_CONTROL_MODE_BLOCK)
        mode[0] = 'B';
    else if (md == GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK)
        mode[0] = 'E';
    else if (md == GLOBUS_FTP_CONTROL_MODE_COMPRESSED)
        mode[0] = 'C';
    else 
        mode[0] = 'Z';
    mode[1] = '\0';
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", mode));
        return tuple;
    }
}

PyObject*
operationattr_set_parallelism(globus_ftp_client_operationattr_t* attr,
                     const globus_ftp_control_parallelism_t* par) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handle_destroy: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_operationattr_set_parallelism(attr, par);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
operationattr_get_parallelism(globus_ftp_client_operationattr_t* attr) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_ftp_control_parallelism_t* par;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "attr_init: PyTuple_New()");
        return NULL;
    }
    par = (globus_ftp_control_parallelism_t*)globus_malloc(sizeof(
                                                        globus_ftp_control_parallelism_t));
    ret = globus_ftp_client_operationattr_get_parallelism(attr, par);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)par, 
				 SWIGTYPE_p_globus_ftp_control_parallelism_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
operationattr_set_tcp_buffer(globus_ftp_client_operationattr_t* attr,
                    const globus_ftp_control_tcpbuffer_t* buffer) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_set_tcp_buffer: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_operationattr_set_tcp_buffer(attr, buffer);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
operationattr_get_tcp_buffer(globus_ftp_client_operationattr_t* attr) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_ftp_control_tcpbuffer_t* buffer;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_get_tcp_buffer: PyTuple_New()");
        return NULL;
    }
    buffer = (globus_ftp_control_tcpbuffer_t*)globus_malloc(sizeof(
                                                      globus_ftp_control_tcpbuffer_t));
    ret = globus_ftp_client_operationattr_get_tcp_buffer(attr, buffer);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)buffer, 
				 SWIGTYPE_p_globus_ftp_control_tcpbuffer_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
operationattr_set_type(globus_ftp_client_operationattr_t* attr, char type) {
    globus_result_t ret;
    char* tmpstr;
    PyObject* tuple;
    globus_ftp_control_type_t tp;
    if (type == '0')
        tp = GLOBUS_FTP_CONTROL_TYPE_NONE;
    else if (type == 'A')
        tp = GLOBUS_FTP_CONTROL_TYPE_ASCII;
    else if (type == 'E')
        tp = GLOBUS_FTP_CONTROL_TYPE_EBCDIC;
    else if (type == 'I')
        tp = GLOBUS_FTP_CONTROL_TYPE_IMAGE;
    else if (type == 'L')
        tp = GLOBUS_FTP_CONTROL_TYPE_LOCAL;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_get_type: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_operationattr_set_type(attr, tp);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
operationattr_get_type(globus_ftp_client_operationattr_t* attr) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    char type[2];
    globus_ftp_control_type_t tp;
    ret = globus_ftp_client_operationattr_get_type(attr, &tp);
    if (tp == GLOBUS_FTP_CONTROL_TYPE_NONE)
        type[0] = '0';
    else if (tp == GLOBUS_FTP_CONTROL_TYPE_ASCII)
        type[0] = 'A';
    else if (tp == GLOBUS_FTP_CONTROL_TYPE_EBCDIC)
        type[0] = 'E';
    else if (tp == GLOBUS_FTP_CONTROL_TYPE_IMAGE)
        type[0] = 'I';
    else if (tp == GLOBUS_FTP_CONTROL_TYPE_LOCAL)
        type[0] = 'L';
    else 
        type[0] = 'Z';
    type[1] = '\0';
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "operationattr_get_type: PyTuple_New()");
        return NULL;
    }
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", type));
    }
    return tuple;
}

PyObject*
restart_marker_init() {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_ftp_client_restart_marker_t* marker;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "restart_marker_init: PyTuple_New()");
        return NULL;
    }
    marker = (globus_ftp_client_restart_marker_t*)globus_malloc(sizeof(
                                               globus_ftp_client_restart_marker_t));
    ret = globus_ftp_client_restart_marker_init(marker);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(marker);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)marker, 
				 SWIGTYPE_p_globus_ftp_client_restart_marker_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
restart_marker_destroy(globus_ftp_client_restart_marker_t* marker) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "restart_marker_destroy: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_restart_marker_destroy(marker);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    globus_free(marker);
    return tuple;
}

PyObject*
restart_marker_copy(globus_ftp_client_restart_marker_t* srcMarker) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_ftp_client_restart_marker_t* marker;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "restart_marker_copy: PyTuple_New()");
        return NULL;
    }
    marker = (globus_ftp_client_restart_marker_t*)globus_malloc(sizeof(
                                       globus_ftp_client_restart_marker_t));
    ret = globus_ftp_client_restart_marker_copy(marker, srcMarker);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(marker);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)marker, 
				 SWIGTYPE_p_globus_ftp_client_restart_marker_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
restart_marker_insert_range(globus_ftp_client_restart_marker_t* marker,
                            PyObject* pOffset, PyObject* pEndOffset) {
    globus_result_t ret;
    char* tmpstr;
    PyObject* tuple;
    globus_off_t offset, endOffset;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "restart_marker_insert_range: PyTuple_New()");
        return NULL;
    }
    if (sizeof(globus_off_t) == 8) {
        /* XXX. this assumes off_t is a long long.  It will
           probably be necessary to ifdef this code for various
           platforms.  This works on 64bit Solaris 7, but I don't
           know about others. */
        if (PyLong_Check(pOffset)) {
            offset = (globus_off_t)PyLong_AsLongLong(pOffset);
        }
        else {
            offset = (globus_off_t)PyInt_AsLong(pOffset);
        }
        if (PyLong_Check(pEndOffset)) {
            endOffset = (globus_off_t)PyLong_AsLongLong(pEndOffset);
        }
        else {
            endOffset = (globus_off_t)PyInt_AsLong(pEndOffset);
        }
    }
    else {
        offset = (globus_off_t)PyInt_AsLong(pOffset);
        endOffset = (globus_off_t)PyInt_AsLong(pEndOffset);
    }
    ret = globus_ftp_client_restart_marker_insert_range(marker, offset, endOffset);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;    
}

PyObject*
restart_marker_set_offset(globus_ftp_client_restart_marker_t* marker,
                          PyObject* pOffset) {
    globus_result_t ret;
    char* tmpstr;
    PyObject* tuple;
    globus_off_t offset;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "restart_marker_set_offset: PyTuple_New()");
        return NULL;
    }
    if ((sizeof(globus_off_t) == 8) && (PyLong_Check(pOffset))) {
        /* XXX. this assumes off_t is a long long.  It will
           probably be necessary to ifdef this code for various
           platforms.  This works on 64bit Solaris 7, but I don't
           know about others. */
        offset = (globus_off_t)PyLong_AsLongLong(pOffset);
    }
    else {
        offset = (globus_off_t)PyInt_AsLong(pOffset);
    }
    ret = globus_ftp_client_restart_marker_set_offset(marker, offset);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;    
}

PyObject*
restart_marker_set_ascii_offset(globus_ftp_client_restart_marker_t* marker,
                                PyObject* pOffset, PyObject* pAsciiOffset) {
    globus_result_t ret;
    char* tmpstr;
    PyObject* tuple;
    globus_off_t offset, asciiOffset;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "restart_marker_set_ascii_offset: PyTuple_New()");
        return NULL;
    }
    if (sizeof(globus_off_t) == 8) {
        /* XXX. this assumes off_t is a long long.  It will
           probably be necessary to ifdef this code for various
           platforms.  This works on 64bit Solaris 7, but I don't
           know about others. */
        if (PyLong_Check(pOffset)) {
            offset = (globus_off_t)PyLong_AsLongLong(pOffset);
        }
        else {
            offset = (globus_off_t)PyInt_AsLong(pOffset);
        }
        if (PyLong_Check(pAsciiOffset)) {
            asciiOffset = (globus_off_t)PyLong_AsLongLong(pAsciiOffset);
        }
        else {
            asciiOffset = (globus_off_t)PyInt_AsLong(pAsciiOffset);
        }
    }
    else {
        offset = (globus_off_t)PyInt_AsLong(pOffset);
        asciiOffset = (globus_off_t)PyInt_AsLong(pAsciiOffset);
    }
    ret = globus_ftp_client_restart_marker_set_ascii_offset(marker,
                                                            offset,
                                                            asciiOffset);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
restart_marker_get_total(globus_ftp_client_restart_marker_t* marker) {
    globus_result_t ret;
    char* tmpstr;
    PyObject* tuple;
    PyObject* oSet;
    globus_off_t offset;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "restart_marker_get_total: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_restart_marker_get_total(marker, &offset);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        if (sizeof(globus_off_t) == 8) {
            /* XXX, This assumes globus_off_t is a long long.  This code will
               probably have to be ifdef'd by platform to account for 
               different types for size_t */
            oSet = PyLong_FromLongLong(offset);
        }
        else {
            oSet = PyInt_FromLong(offset);
        }
        PyTuple_SET_ITEM(tuple, 1, oSet);
    }
    return tuple;
}

PyObject*
restart_marker_to_string(globus_ftp_client_restart_marker_t* marker,
                         char** markerString) {
    globus_result_t ret;
    char* tmpstr;
    char* str;
    PyObject* list;
    list = PyList_New(0);
    ret = globus_ftp_client_restart_marker_to_string(marker, markerString);
    printf("marker string in C is: %s\n", *markerString);
    if (ret != GLOBUS_SUCCESS) {
        PyList_Append(list, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyList_Append(list, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyList_Append(list, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyList_Append(list, Py_None);
    }
    return list;
}

PyObject*
restart_marker_from_string(globus_ftp_client_restart_marker_t* marker,
                           const char* markerString) {
    globus_result_t ret;
    char* tmpstr;
    PyObject* tuple;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "restart_marker_from_string: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_restart_marker_from_string(marker, markerString);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
handle_init(globus_ftp_client_handleattr_t* attr) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_ftp_client_handle_t* handle;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handle_init: PyTuple_New()");
        return NULL;
    }
    handle = (globus_ftp_client_handle_t*)globus_malloc(sizeof(globus_ftp_client_handle_t));
    ret = globus_ftp_client_handle_init(handle, attr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(handle);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)handle, 
				 SWIGTYPE_p_globus_ftp_client_handle_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
handle_destroy(globus_ftp_client_handle_t* handle) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handle_destroy: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_handle_destroy(handle);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    globus_free(handle);
    return tuple;
}

PyObject*
handle_add_plugin(globus_ftp_client_handle_t* handle,
                  globus_ftp_client_plugin_t* plugin) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handle_add_plugin: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_handle_add_plugin(handle, plugin);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
handle_remove_plugin(globus_ftp_client_handle_t* handle,
                     globus_ftp_client_plugin_t* plugin) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handle_remove_plugin: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_handle_remove_plugin(handle, plugin);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    globus_free(handle);
    return tuple;
}

PyObject*
handle_cache_url_state(globus_ftp_client_handle_t* handle,
                       const char* url) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handle_cache_url_state: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_handle_cache_url_state(handle, url);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    globus_free(handle);
    return tuple;
}

PyObject*
handle_flush_url_state(globus_ftp_client_handle_t* handle,
                       const char* url) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handle_flush_url_state: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_handle_flush_url_state(handle, url);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    globus_free(handle);
    return tuple;
}

PyObject*
handle_set_user_pointer(globus_ftp_client_handle_t* handle,
			PyObject* obj) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handle_set_user_pointer: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_handle_set_user_pointer(handle, (void*)obj);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        Py_INCREF(obj);  /*How does this ref get freed?*/
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
handle_get_user_pointer(const globus_ftp_client_handle_t* handle) {
    globus_result_t ret;
    PyObject* tuple;
    PyObject* userData;
    void** ptr;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handle_get_user_pointer: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_handle_get_user_pointer(handle, ptr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("o",(PyObject*)*ptr));
    }
    return tuple;
}

PyObject*
py_delete(globus_ftp_client_handle_t* handle, const char* url,
       globus_ftp_client_operationattr_t* attr,
       PyObject* pyfunc, PyObject* arg) {
    
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
	
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_delete: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
   
    ret = globus_ftp_client_delete(handle, url, attr, 
                                PythonCompleteCallback, (void*)cbStruct);
    
   

    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, 
				 SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
py_mkdir(globus_ftp_client_handle_t* handle, const char* url,
         globus_ftp_client_operationattr_t* attr,
         PyObject* pyfunc, PyObject* arg) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_mkdir: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    ret = globus_ftp_client_mkdir(handle, url, attr, 
                                  PythonCompleteCallback, (void*)cbStruct);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
py_rmdir(globus_ftp_client_handle_t* handle, const char* url,
         globus_ftp_client_operationattr_t* attr,
         PyObject* pyfunc, PyObject* arg) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_mkdir: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    ret = globus_ftp_client_rmdir(handle, url, attr, 
                                  PythonCompleteCallback, (void*)cbStruct);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
py_list(globus_ftp_client_handle_t* handle, const char* url,
        globus_ftp_client_operationattr_t* attr,
        PyObject* pyfunc, PyObject* arg) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_list: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    ret = globus_ftp_client_list(handle, url, attr, 
                                 PythonCompleteCallback, (void*)cbStruct);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
verbose_list(globus_ftp_client_handle_t* handle, const char* url,
              globus_ftp_client_operationattr_t* attr,
              PyObject* pyfunc, PyObject* arg) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "verbose_list: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    ret = globus_ftp_client_verbose_list(handle, url, attr, 
                                         PythonCompleteCallback, (void*)cbStruct);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}


PyObject*
size(globus_ftp_client_handle_t* handle, const char* url,
              globus_ftp_client_operationattr_t* attr,PyObject* pyfunc, PyObject* arg) {
    
    globus_result_t ret;
    PyObject *tuple, * ptr, *ptr2;
    char *tmpStr;
    callbackStruct* cbStruct;
    globus_off_t *fileSize;
    if (!(tuple=PyTuple_New(3)))
    {
        PyErr_SetString(PyExc_RuntimeError, "size: PyTuple_New()");
        return NULL;
    }
    
    fileSize = (globus_off_t*)globus_malloc(sizeof(globus_off_t));
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
	
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    

    ret = globus_ftp_client_size(handle, url, attr, fileSize, 
                                         PythonCompleteCallback, (void*)cbStruct);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", -1));
        tmpStr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 2, Py_BuildValue("s", tmpStr));
        globus_libc_free(tmpStr);
        globus_libc_free(cbStruct);
    }
    else 
    {
	
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
      	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	ptr2 = SWIG_NewPointerObj((void*)fileSize, SWIGTYPE_p_globus_off_t, 0);
	Py_XINCREF(ptr2);
	PyTuple_SET_ITEM(tuple, 1, ptr2);
	PyTuple_SET_ITEM(tuple, 2, ptr);
    }
    return tuple;
}


	
PyObject *
exists(globus_ftp_client_handle_t* handle, const char* url,
              globus_ftp_client_operationattr_t* attr,PyObject* pyfunc, PyObject* arg) {
    
    globus_result_t ret;
    PyObject *tuple, * ptr;
    char *tmpStr;
    callbackStruct* cbStruct;
    
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "size: PyTuple_New()");
        return NULL;
    }
    
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
	
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    

    ret = globus_ftp_client_exists(handle, url, attr,  
                                         PythonCompleteCallback, (void*)cbStruct);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpStr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpStr));
        globus_libc_free(tmpStr);
        globus_libc_free(cbStruct);
    }
    else 
    {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
      	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
modification_time(globus_ftp_client_handle_t* handle, const char* url,
              globus_ftp_client_operationattr_t* attr,
              PyObject* pyfunc, PyObject* arg) {
    
    globus_result_t ret;
    PyObject *tuple, * ptr, *ptr2;
    char *tmpStr;
    callbackStruct* cbStruct;
    globus_abstime_t * modTime;
    time_t *t;

    if (!(tuple=PyTuple_New(3)))
    {
        PyErr_SetString(PyExc_RuntimeError, "size: PyTuple_New()");
        return NULL;
    }
     cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
     modTime = (globus_abstime_t *)globus_malloc(sizeof(globus_abstime_t));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    ret = globus_ftp_client_modification_time(handle, url, attr, modTime,
                                         PythonCompleteCallback, (void*)cbStruct);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpStr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpStr));
        PyTuple_SET_ITEM(tuple, 2, Py_BuildValue("i", -1));
	globus_libc_free(tmpStr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
	ptr2 = SWIG_NewPointerObj((void*)modTime, SWIGTYPE_p_globus_abstime_t, 0);
	Py_XINCREF(ptr2);
	PyTuple_SET_ITEM(tuple, 2, ptr2);
    }
    return tuple;
}

PyObject*
move(globus_ftp_client_handle_t* handle, const char* srcUrl,
     const char* destUrl, globus_ftp_client_operationattr_t* attr,
     PyObject* pyfunc, PyObject* arg) {
    callbackStruct *cbStruct;
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "verbose_list: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    ret = globus_ftp_client_move(handle, srcUrl, destUrl, attr, 
                                         PythonCompleteCallback, (void*)cbStruct);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
get(globus_ftp_client_handle_t* handle, const char* url,
           globus_ftp_client_operationattr_t* attr, 
           globus_ftp_client_restart_marker_t* marker,
           PyObject* pyfunc, PyObject* arg) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_get: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_client_get(handle, url, attr, marker, 
                                PythonCompleteCallback, (void*)cbStruct);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
partial_get(globus_ftp_client_handle_t* handle, const char* url,
            globus_ftp_client_operationattr_t* attr, 
            globus_ftp_client_restart_marker_t* marker,
            PyObject* pyOffset, PyObject* pyEndOffset,
	    PyObject* pyfunc, PyObject* arg) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_off_t offset, endOffset;
    callbackStruct* cbStruct;
   
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_get: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    if (sizeof(globus_off_t) == 8) {
        /* XXX. this assumes off_t is a long long.  It will
           probably be necessary to ifdef this code for various
           platforms.  This works on 64bit Solaris 7, but I don't
           know about others. */
        if (PyLong_Check(pyOffset)) {
            offset = (globus_off_t)PyLong_AsLongLong(pyOffset);
        }
        else {
            offset = (globus_off_t)PyInt_AsLong(pyOffset);
        }
        if (PyLong_Check(pyEndOffset)) {
            endOffset = (globus_off_t)PyLong_AsLongLong(pyEndOffset);
        }
        else {
            endOffset = (globus_off_t)PyInt_AsLong(pyEndOffset);
        }
    }
    else {
        offset = (globus_off_t)PyInt_AsLong(pyOffset);
        endOffset = (globus_off_t)PyInt_AsLong(pyEndOffset);
    }
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_client_partial_get(handle, url, attr, marker, 
					offset, endOffset,
					PythonCompleteCallback, (void*)cbStruct);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
extended_get(globus_ftp_client_handle_t* handle, const char* url,
             globus_ftp_client_operationattr_t* attr, 
             globus_ftp_client_restart_marker_t* marker,
             const char* eret_alg_str,
             PyObject* pyfunc, PyObject* arg) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_extended_get: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_client_extended_get(handle, url, attr, marker, 
                                         eret_alg_str, 
                                         PythonCompleteCallback, (void*)cbStruct);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
put(globus_ftp_client_handle_t* handle, const char* url,
           globus_ftp_client_operationattr_t* attr, 
           globus_ftp_client_restart_marker_t* marker,
           PyObject* pyfunc, PyObject* arg) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_put: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_client_put(handle, url, attr, marker, 
                                PythonCompleteCallback, (void*)cbStruct);
    
    Py_END_ALLOW_THREADS
	
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
partial_put(globus_ftp_client_handle_t* handle, const char* url,
	    globus_ftp_client_operationattr_t* attr,
	    globus_ftp_client_restart_marker_t* marker,
	    PyObject* pyOffset, PyObject* pyEndOffset,
	    PyObject* pyfunc, PyObject* arg) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_off_t offset, endOffset;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_partial_put: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    if (sizeof(globus_off_t) == 8) {
        /* XXX. this assumes off_t is a long long.  It will
           probably be necessary to ifdef this code for various
           platforms.  This works on 64bit Solaris 7, but I don't
           know about others. */
        if (PyLong_Check(pyOffset)) {
            offset = (globus_off_t)PyLong_AsLongLong(pyOffset);
        }
        else {
            offset = (globus_off_t)PyInt_AsLong(pyOffset);
        }
        if (PyLong_Check(pyEndOffset)) {
            endOffset = (globus_off_t)PyLong_AsLongLong(pyEndOffset);
        }
        else {
            endOffset = (globus_off_t)PyInt_AsLong(pyEndOffset);
        }
    }
    else {
        offset = (globus_off_t)PyInt_AsLong(pyOffset);
        endOffset = (globus_off_t)PyInt_AsLong(pyEndOffset);
    }
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_client_partial_put(handle, url, attr, marker,
					offset, endOffset,
					PythonCompleteCallback, (void*)cbStruct);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
extended_put(globus_ftp_client_handle_t* handle, const char* url,
             globus_ftp_client_operationattr_t* attr, 
             globus_ftp_client_restart_marker_t* marker,
             const char* esto_alg_str,
             PyObject* pyfunc, PyObject* arg) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_extended_put: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_client_extended_put(handle, url, attr, marker, esto_alg_str, 
                                PythonCompleteCallback, (void*)cbStruct);
    
    Py_END_ALLOW_THREADS
	
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
third_party_transfer(globus_ftp_client_handle_t* handle, const char* srcUrl,
                     globus_ftp_client_operationattr_t* srcAttr, const char* dstUrl,
                     globus_ftp_client_operationattr_t* dstAttr, 
                     globus_ftp_client_restart_marker_t* marker,
                     PyObject* pyfunc, PyObject* arg) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_get: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_client_third_party_transfer(handle, srcUrl, srcAttr, 
                                                 dstUrl, dstAttr, marker, 
                                                 PythonCompleteCallback, 
                                                 (void*)cbStruct);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
partial_third_party_transfer(globus_ftp_client_handle_t* handle, 
			     const char* srcUrl, 
			     globus_ftp_client_operationattr_t* srcAttr, 
			     const char* dstUrl,
			     globus_ftp_client_operationattr_t* dstAttr, 
			     globus_ftp_client_restart_marker_t* marker,
			     PyObject* pyOffset, PyObject* pyEndOffset,
			     PyObject* pyfunc, PyObject* arg) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_off_t offset, endOffset;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_third_party: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    if (sizeof(globus_off_t) == 8) {
        /* XXX. this assumes off_t is a long long.  It will
           probably be necessary to ifdef this code for various
           platforms.  This works on 64bit Solaris 7, but I don't
           know about others. */
        if (PyLong_Check(pyOffset)) {
            offset = (globus_off_t)PyLong_AsLongLong(pyOffset);
        }
        else {
            offset = (globus_off_t)PyInt_AsLong(pyOffset);
        }
        if (PyLong_Check(pyEndOffset)) {
            endOffset = (globus_off_t)PyLong_AsLongLong(pyEndOffset);
        }
        else {
            endOffset = (globus_off_t)PyInt_AsLong(pyEndOffset);
        }
    }
    else {
        offset = (globus_off_t)PyInt_AsLong(pyOffset);
        endOffset = (globus_off_t)PyInt_AsLong(pyEndOffset);
    }
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_client_partial_third_party_transfer(handle, srcUrl, srcAttr, 
							 dstUrl, dstAttr, marker,
							 offset, endOffset,
							 PythonCompleteCallback, 
							 (void*)cbStruct);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
extended_third_party_transfer(globus_ftp_client_handle_t* handle, 
                              const char* srcUrl,
                              globus_ftp_client_operationattr_t* srcAttr, 
                              const char* eret_alg_str,
                              const char* dstUrl,
                              globus_ftp_client_operationattr_t* dstAttr,
                              const char* esto_alg_str,
                              globus_ftp_client_restart_marker_t* marker,
                              PyObject* pyfunc, PyObject* arg) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_extended_third_party: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_client_extended_third_party_transfer(handle, srcUrl, 
                                                          srcAttr, eret_alg_str, 
                                                          dstUrl, dstAttr, 
                                                          esto_alg_str, marker, 
                                                          PythonCompleteCallback, 
                                                          (void*)cbStruct);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
py_abort(globus_ftp_client_handle_t* handle) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "abort: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_abort(handle);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
} 
	
PyObject*
register_read(globus_ftp_client_handle_t* handle, globus_byte_t* buffer,
              int size, PyObject* pyfunc, PyObject* arg) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_get: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    ret = globus_ftp_client_register_read(handle, buffer, size, 
                                          PythonDataCallback, cbStruct);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
register_write(globus_ftp_client_handle_t* handle, PyObject* buffer,
               int size, PyObject* offSet, globus_bool_t eof, 
               PyObject* pyfunc, PyObject* arg) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_off_t offset;
    globus_byte_t* buf;
    int sizeNew;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_get: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    if ((sizeof(globus_off_t) == 8) && (PyLong_Check(offSet))) {
        /* XXX. this assumes off_t is a long long.  It will
           probably be necessary to ifdef this code for various
           platforms.  This works on 64bit Solaris 7, but I don't
           know about others. */
        offset = (globus_off_t)PyLong_AsLongLong(offSet);
    }
    else {
        offset = (globus_off_t)PyInt_AsLong(offSet);
    }
    /* buf = (globus_byte_t*)PyString_AsString(buffer);*/
    PyString_AsStringAndSize(buffer, (char**)&buf, &sizeNew);
    ret = globus_ftp_client_register_write(handle, buf, sizeNew, offset, eof, 
                                          PythonDataCallback, cbStruct);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

void
free_callback(callbackStruct* cbStruct) {
    Py_XDECREF(cbStruct->callback);
    Py_XDECREF(cbStruct->arg);
    globus_free(cbStruct);
}

void
free_callback2(callbackStruct2* cbStruct) {
    Py_XDECREF(cbStruct->cb1);
    Py_XDECREF(cbStruct->cb2);
    Py_XDECREF(cbStruct->cb3);
    Py_XDECREF(cbStruct->arg);
    globus_free(cbStruct);
}

void
free_callback3(callbackStruct3* cbStruct) {
    Py_XDECREF(cbStruct->cb1);
    Py_XDECREF(cbStruct->cb2);
    Py_XDECREF(cbStruct->cb3);
    Py_XDECREF(cbStruct->cb4);
    Py_XDECREF(cbStruct->arg);
    globus_free(cbStruct);
}

PyObject*
debug_plugin_init(FILE* stream, const char* text) {
    globus_ftp_client_plugin_t* plugin;
    globus_result_t ret;
    char* tmpstr;
    PyObject *tuple, *ptr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "debug_plugin_init: PyTuple_New()");
        return NULL;
    }
    plugin = (globus_ftp_client_plugin_t*)globus_malloc(sizeof(globus_ftp_client_plugin_t));
    ret = globus_ftp_client_debug_plugin_init(plugin, stream, text);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(plugin);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)plugin, 
				 SWIGTYPE_p_globus_ftp_client_plugin_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
debug_plugin_destroy(globus_ftp_client_plugin_t* plugin) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "debug_plugin_destroy: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_debug_plugin_destroy(plugin);
    globus_free(plugin);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
restart_marker_plugin_init(PyObject* beginCB, PyObject* markerCB, 
                           PyObject* completeCB, PyObject* arg) {
    globus_ftp_client_plugin_t* plugin;
    globus_result_t ret;
    char* tmpstr;
    PyObject *tuple, *ptr, *ptr2;
    callbackStruct2* cbStruct;
    if (!(tuple=PyTuple_New(3)))
    {
        PyErr_SetString(PyExc_RuntimeError, "restart_marker_plugin_init: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct2*)globus_malloc(sizeof(callbackStruct2));
    Py_XINCREF(beginCB);
    Py_XINCREF(markerCB);
    Py_XINCREF(completeCB);
    Py_XINCREF(arg);
    cbStruct->cb1 = beginCB;
    cbStruct->cb2 = markerCB;
    cbStruct->cb3 = completeCB;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    plugin = (globus_ftp_client_plugin_t*)globus_malloc(sizeof(globus_ftp_client_plugin_t));
    ret = globus_ftp_client_restart_marker_plugin_init(plugin, PythonRestartMarkerPluginBeginCallback, 
                                                       PythonRestartMarkerPluginMarkerCallback,
                                                       PythonRestartMarkerPluginCompleteCallback, (void*)cbStruct);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 2, Py_None);
        globus_libc_free(tmpstr);
        globus_libc_free(plugin);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)plugin, 
				 SWIGTYPE_p_globus_ftp_client_plugin_t, 0);
         ptr2 = SWIG_NewPointerObj((void*)cbStruct, 
				 SWIGTYPE_p_callbackStruct2, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
        PyTuple_SET_ITEM(tuple, 2, ptr2);
    }
    return tuple;
}

PyObject*
restart_marker_plugin_destroy(globus_ftp_client_plugin_t* plugin) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "restart_marker_plugin_destroy: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_restart_marker_plugin_destroy(plugin);
    globus_free(plugin);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
performance_marker_plugin_init(PyObject* beginCB, PyObject* markerCB, 
			       PyObject* completeCB, PyObject* arg) {
    globus_ftp_client_plugin_t* plugin;
    globus_result_t ret;
    char* tmpstr;
    PyObject *tuple, *ptr, *ptr2;
    callbackStruct2* cbStruct;
    if (!(tuple=PyTuple_New(3)))
    {
        PyErr_SetString(PyExc_RuntimeError, "restart_marker_plugin_init: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct2*)globus_malloc(sizeof(callbackStruct2));
    Py_XINCREF(beginCB);
    Py_XINCREF(markerCB);
    Py_XINCREF(completeCB);
    Py_XINCREF(arg);
    cbStruct->cb1 = beginCB;
    cbStruct->cb2 = markerCB;
    cbStruct->cb3 = completeCB;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    plugin = (globus_ftp_client_plugin_t*)globus_malloc(sizeof(globus_ftp_client_plugin_t));
    ret = globus_ftp_client_perf_plugin_init(plugin, PythonPerfMarkerPluginBeginCallback, 
					     PythonPerfMarkerPluginMarkerCallback,
					     PythonPerfMarkerPluginCompleteCallback, (void*)cbStruct);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 2, Py_None);
        globus_libc_free(tmpstr);
        globus_libc_free(plugin);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)plugin, 
				 SWIGTYPE_p_globus_ftp_client_plugin_t, 0);
	ptr2 = SWIG_NewPointerObj((void*)cbStruct, 
				  SWIGTYPE_p_callbackStruct2, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
        PyTuple_SET_ITEM(tuple, 2, ptr2);
    }
    return tuple;
}

PyObject*
performance_marker_plugin_destroy(globus_ftp_client_plugin_t* plugin) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "performance_marker_plugin_destroy: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_perf_plugin_destroy(plugin);
    globus_free(plugin);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}
PyObject*
throughput_plugin_init(PyObject* beginCB, PyObject* stripeCB, PyObject* totalCB,
		       PyObject* completeCB, PyObject* arg) {
    globus_ftp_client_plugin_t* plugin;
    globus_result_t ret;
    char* tmpstr;
    PyObject *tuple, *ptr, *ptr2;
    callbackStruct3* cbStruct;

    if (!(tuple=PyTuple_New(3)))
    {
        PyErr_SetString(PyExc_RuntimeError, "throughput_plugin_init: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct3*)globus_malloc(sizeof(callbackStruct3));
    Py_XINCREF(beginCB);
    Py_XINCREF(stripeCB);
    Py_XINCREF(totalCB);
    Py_XINCREF(completeCB);
    Py_XINCREF(arg);
    cbStruct->cb1 = beginCB;
    cbStruct->cb2 = stripeCB;
    cbStruct->cb3 = totalCB;
    cbStruct->cb4 = completeCB;
    cbStruct->arg = arg;
    if (thread_state == NULL) {
        PyEval_InitThreads();
        thread_state = PyThreadState_New(PyThreadState_Get()->interp);
    }
    plugin = (globus_ftp_client_plugin_t*)globus_malloc(sizeof(globus_ftp_client_plugin_t));
    ret = globus_ftp_client_throughput_plugin_init(plugin, PythonThroughputPluginBeginCallback, 
						   PythonThroughputPluginStripeCallback,
						   PythonThroughputPluginTotalCallback,
						   PythonThroughputPluginCompleteCallback, (void*)cbStruct);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 2, Py_None);
        globus_libc_free(tmpstr);
        globus_libc_free(plugin);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)plugin, 
				 SWIGTYPE_p_globus_ftp_client_plugin_t, 0);
	ptr2 = SWIG_NewPointerObj((void*)cbStruct, 
				  SWIGTYPE_p_callbackStruct3, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
        PyTuple_SET_ITEM(tuple, 2, ptr2);
    }
    return tuple;
}
PyObject*
throughput_plugin_destroy(globus_ftp_client_plugin_t* plugin) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "throughput_plugin_destroy: PyTuple_New()");
        return NULL;
    }
    ret = globus_ftp_client_throughput_plugin_destroy(plugin);
    globus_free(plugin);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

/* ----------------------------- 
/* ----------------------------- 
PyObject*
netlogger_plugin_init(PyObject* nl_url, PyObject* prog_name, PyObject* opaque_string) {
    globus_ftp_client_plugin_t* plugin;
    globus_result_t ret;
    char* tmpstr;
    char* nl_url_p;
    char* prog_name_p;
    char* opaque_string_p;
    PyObject *tuple, *ptr;

    if (!(tuple=PyTuple_New(3)))
    {
        PyErr_SetString(PyExc_RuntimeError, "netlogger_plugin_init: PyTuple_New()");
        return NULL;
    }
#if defined(GLOBUS_BUILD_WITH_NETLOGGER)
    Py_XINCREF(nl_url);
    Py_XINCREF(prog_name);
    Py_XINCREF(opaque_string);
    nl_url_p = PyString_AsString(nl_url);
    prog_name_p = PyString_AsString(prog_name);
    opaque_string_p = PyString_AsString(opaque_string);
    plugin = (globus_ftp_client_plugin_t*)globus_malloc(sizeof(globus_ftp_client_plugin_t));
    ret = globus_ftp_client_throughput_nl_plugin_init(plugin, nl_url_p, 
						      prog_name_p, opaque_string_p);
    
    if (ret != GLOBUS_SUCCESS) {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
	globus_libc_free(tmpstr);
	globus_libc_free(plugin);
    }
    else {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	ptr = SWIG_NewPointerObj((void*)plugin, 
				 SWIGTYPE_p_globus_ftp_client_plugin_t, 0);
	PyTuple_SET_ITEM(tuple, 1, ptr);
    }
#else
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", "not compiled with support for netlogger"));
#endif  
    return tuple;
}

PyObject*
netlogger_plugin_destroy(globus_ftp_client_plugin_t* plugin) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "netlogger_plugin_destroy: PyTuple_New()");
        return NULL;
    }
#if defined(GLOBUS_BUILD_WITH_NETLOGGER)
    ret = globus_ftp_client_restart_marker_plugin_destroy(plugin);
    globus_free(plugin);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
#else
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", "not compiled with support for netlogger"));
#endif  
    return tuple;
}
*/
//---------------------------------------------------
/*
globus_ftp_client_plugin_t*
plugin_init(void) {
    globus_ftp_client_plugin_t* plugin;
    PyObject* list;
    int lsize = 15;
    int i;
    plugin = (globus_ftp_client_plugin_t*)globus_malloc(sizeof(globus_ftp_client_plugin_t));
    list = PyList_New(lsize);
    for(i = 0; i < lsize; i++) {
        Py_INCREF(Py_None);
        PyList_SetItem(list, i, Py_None);
    }
    plugin->plugin_name = GLOBUS_NULL;
    plugin->instantiate = GLOBUS_NULL;
    plugin->destroy = GLOBUS_NULL;
    plugin->get_func = GLOBUS_NULL;
    plugin->put_func = GLOBUS_NULL;
    plugin->transfer_func = GLOBUS_NULL;
    plugin->abort_func = GLOBUS_NULL;
    plugin->connect_func = GLOBUS_NULL;
    plugin->auth_func = GLOBUS_NULL;
    plugin->read_func = GLOBUS_NULL;
    plugin->write_func = GLOBUS_NULL;
    plugin->data_func = GLOBUS_NULL;
    plugin->command_func = GLOBUS_NULL;
    plugin->response_func = GLOBUS_NULL;
    plugin->fault_func = GLOBUS_NULL;
    plugin->command_mask = GLOBUS_FTP_CLIENT_CMD_MASK_NONE;
    plugin->plugin_specific = list;
    return plugin;
}

void
set_plugin_name(globus_ftp_client_plugin_t* plugin, char* name) {
    plugin->plugin_name = name;
}

void
set_plugin_instantiate(globus_ftp_client_plugin_t* plugin, PyObject* func) {
    PyObject* list = plugin->plugin_specific;
    plugin->instantiate = PythonPluginInstantiate;
    Py_INCREF(func);
    PyList_SetItem(list, INSTANTIATE, func);
}

void
set_plugin_destroy(globus_ftp_client_plugin_t* plugin, PyObject* func) {
    PyObject* list = plugin->plugin_specific;
    plugin->destroy = PythonPluginDestroy;
    Py_INCREF(func);
    PyList_SetItem(list, DESTROY, func);
}

void
set_plugin_get(globus_ftp_client_plugin_t* plugin, PyObject* func) {
    PyObject* list = plugin->plugin_specific;
    plugin->get_func = PythonPluginGet;
    Py_INCREF(func);
    PyList_SetItem(list, GET, func);
}

void
set_plugin_put(globus_ftp_client_plugin_t* plugin, PyObject* func) {
    PyObject* list = plugin->plugin_specific;
    plugin->put_func = PythonPluginPut;
    Py_INCREF(func);
    PyList_SetItem(list, PUT, func);
}

void
set_plugin_transfer(globus_ftp_client_plugin_t* plugin, PyObject* func) {
    PyObject* list = plugin->plugin_specific;
    plugin->transfer_func = PythonPluginTransfer;
    Py_INCREF(func);
    PyList_SetItem(list, TRANSFER, func);
}

void
set_plugin_abort(globus_ftp_client_plugin_t* plugin, PyObject* func) {
    PyObject* list = plugin->plugin_specific;
    plugin->abort_func = PythonPluginAbort;
    Py_INCREF(func);
    PyList_SetItem(list, ABORT, func);
}

void
set_plugin_connect(globus_ftp_client_plugin_t* plugin, PyObject* func) {
    PyObject* list = plugin->plugin_specific;
    plugin->connect_func = PythonPluginConnect;
    Py_INCREF(func);
    PyList_SetItem(list, CONNECT, func);
}

void
set_plugin_authenticate(globus_ftp_client_plugin_t* plugin, PyObject* func) {
    PyObject* list = plugin->plugin_specific;
    plugin->auth_func = PythonPluginAuth;
    Py_INCREF(func);
    PyList_SetItem(list, AUTH, func);
}

void
set_plugin_read(globus_ftp_client_plugin_t* plugin, PyObject* func) {
    PyObject* list = plugin->plugin_specific;
    plugin->read_func = PythonPluginRead;
    Py_INCREF(func);
    PyList_SetItem(list, READ, func);
}

void
set_plugin_write(globus_ftp_client_plugin_t* plugin, PyObject* func) {
    PyObject* list = plugin->plugin_specific;
    plugin->write_func = PythonPluginWrite;
    Py_INCREF(func);
    PyList_SetItem(list, WRITE, func);
}

void
set_plugin_data(globus_ftp_client_plugin_t* plugin, PyObject* func) {
    PyObject* list = plugin->plugin_specific;
    plugin->data_func = PythonPluginData;
    Py_INCREF(func);
    PyList_SetItem(list, DATA, func);
}

void
set_plugin_command(globus_ftp_client_plugin_t* plugin, PyObject* func) {
    PyObject* list = plugin->plugin_specific;
    plugin->command_func = PythonPluginCommand;
    Py_INCREF(func);
    PyList_SetItem(list, COMMAND, func);
}

void
set_plugin_response(globus_ftp_client_plugin_t* plugin, PyObject* func) {
    PyObject* list = plugin->plugin_specific;
    plugin->response_func = PythonPluginResponse;
    Py_INCREF(func);
    PyList_SetItem(list, RESPONSE, func);
}

void
set_plugin_fault(globus_ftp_client_plugin_t* plugin, PyObject* func) {
    PyObject* list = plugin->plugin_specific;
    plugin->fault_func = PythonPluginFault;
    Py_INCREF(func);
    PyList_SetItem(list, FAULT, func);
}

void
plugin_destroy(globus_ftp_client_plugin_t* plugin) {
    PyObject* list = (PyObject*)plugin->plugin_specific;
    int size = PyList_Size(list);
    int i;
    for(i = 0; i < size; i++) {
        Py_DECREF(PyList_GetItem(list, i));
    }
    globus_libc_free(plugin);
    return;
}
*/
%}
