/*############################################################################
# Keith R. Jackson, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################*/
%module gassFilec
%{
#include <stdio.h>
#include <fcntl.h>
#include "globus_common.h"
#include "globus_gass_file.h"
%}

%typemap(python, out) FILE* {
    int fd, mode, accmode;
    char md[10];
    int next = 0;
    if ($1 == NULL) {
        PyErr_SetString(PyExc_TypeError, "expected a FILE* but got NULL");
        return NULL;
    }
    fd = fileno($1); 
    mode = fcntl(fd, F_GETFL, 0);
    if (mode < 0) {
        PyErr_SetString(PyExc_SystemError, "fcntl error");
        return NULL;
    }
    accmode = mode & O_ACCMODE;
    if (accmode == O_RDONLY) {
       md[next] = 'r';
       next++;
    } 
    else if (accmode == O_WRONLY) {
       md[next] = 'w';
       next++;
    }
    else if (accmode == O_RDWR) {
       md[next] = 'r';
       next++;
       md[next] = '+';
       next++;
    }
    else {
       PyErr_SetString(PyExc_SystemError, "unknown access mode");
       return NULL;
    }
    if (mode == O_APPEND) {
       md[next] = 'a';
       next++;
    }
    if (mode == O_NONBLOCK) {
       md[next] = 'b';
       next++;
    } 
    md[next] = '\0';
    $result = PyFile_FromFile($1, "gass", md, NULL);
}

%typemap(python, in) FILE * {
    if (!PyFile_Check($input)) {
        PyErr_SetString(PyExc_TypeError, "expected PyFile");
        return NULL;
    }
    $1=PyFile_AsFile($input);
}

%inline %{
globus_module_descriptor_t* get_module(void) {
    return GLOBUS_GASS_FILE_MODULE;
}

int 
gass_open(char* file, int oflags) {
    int ret;
    Py_BEGIN_ALLOW_THREADS
    ret = globus_gass_open(file, oflags);
    Py_END_ALLOW_THREADS
    return ret;
}

int
gass_close(int fd) {
    int ret;
    Py_BEGIN_ALLOW_THREADS
    ret = globus_gass_close(fd);
    Py_END_ALLOW_THREADS
    return ret;
}

FILE*
gass_fopen(char* file, char* mode) {
    FILE* ret;
    Py_BEGIN_ALLOW_THREADS
    ret = globus_gass_fopen(file, mode);
    Py_END_ALLOW_THREADS
    return ret;
}

int
gass_fclose(FILE* f) {
    int ret;
    Py_BEGIN_ALLOW_THREADS
    ret = globus_gass_fclose(f);
    Py_END_ALLOW_THREADS
    return ret;
}
%}
