/*############################################################################
# Keith R. Jackson, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################*/
%module gsic
%{
#include "globus_common.h"
#include "globus_gss_assist.h"
#include "gssapi.h"
#include "globus_io.h"


/*The old sslutils includes*/


#include <wait.h>
#include <stdio.h>

#include "openssl/crypto.h"
#include "openssl/ssl.h"
#include "openssl/err.h"
#include "openssl/bio.h"
#include "openssl/pem.h"
#include "openssl/x509.h"

#if SSLEAY_VERSION_NUMBER >= 0x0090581fL
#include "openssl/x509v3.h"
#endif

#include "openssl/stack.h"

#include "globus_common.h"
#include "globus_gss_assist.h"
#include "globus_io.h"

#include "gssapi.h"
#include "sslutils.h"

#define GSS_I_CTX_INITIALIZED                       1
#define GSS_I_DISALLOW_ENCRYPTION                   2
#define GSS_I_PROTECTION_FAIL_ON_CONTEXT_EXPIRATION 4


#define GSS_C_QOP_GLOBUS_GSSAPI_SSLEAY_BIG 1
/*
 * Use the SSLeay error facility with the ERR_LIB_USER
 */

/* Offset from ERR_LIB_USER where GSSERR library will be stored */
#define ERR_USER_LIB_GSSERR_NUMBER                  ((ERR_LIB_USER) + 2)

#define GSSerr(f,r) ERR_PUT_error(ERR_USER_LIB_GSSERR_NUMBER,(f),(r),ERR_file_name,__LINE__)

/*
 * defines for function codes our minor error codes
 * These match strings defined in gsserr.c
 */

#define GSSERR_F_BASE                   100

#define GSSERR_F_ACCEPT_SEC             GSSERR_F_BASE + 0
#define GSSERR_F_ACQUIRE_CRED           GSSERR_F_BASE + 1
#define GSSERR_F_COMPARE_NAME           GSSERR_F_BASE + 2
#define GSSERR_F_DELETE_SEC             GSSERR_F_BASE + 3
#define GSSERR_F_EXPORT_NAME            GSSERR_F_BASE + 4
#define GSSERR_F_GLOBUSFILE             GSSERR_F_BASE + 5
#define GSSERR_F_IMPORT_NAME            GSSERR_F_BASE + 6
#define GSSERR_F_INIT_SEC               GSSERR_F_BASE + 7
#define GSSERR_F_RELEASE_BUFFER         GSSERR_F_BASE + 8
#define GSSERR_F_RELEASE_CRED           GSSERR_F_BASE + 9
#define GSSERR_F_RELEASE_NAME           GSSERR_F_BASE + 10
/* In gssutil.c: */
#define GSSERR_F_NAME_TO_NAME           GSSERR_F_BASE + 11
#define GSSERR_F_CREATE_FILL            GSSERR_F_BASE + 12
#define GSSERR_F_GS_HANDSHAKE           GSSERR_F_BASE + 13
#define GSSERR_F_GS_RETRIVE_PEER        GSSERR_F_BASE + 14

#define GSSERR_F_WRAP                   GSSERR_F_BASE + 15
#define GSSERR_F_UNWRAP                 GSSERR_F_BASE + 16
#define GSSERR_F_GET_MIC                GSSERR_F_BASE + 17
#define GSSERR_F_VERIFY_MIC             GSSERR_F_BASE + 18

#define GSSERR_F_IMPORT_SEC             GSSERR_F_BASE + 19
#define GSSERR_F_EXPORT_SEC             GSSERR_F_BASE + 20

#define GSSERR_F_IMPORT_CRED            GSSERR_F_BASE + 21
#define GSSERR_F_EXPORT_CRED            GSSERR_F_BASE + 22
#define GSSERR_F_READ                   GSSERR_F_BASE + 23
#define GSSERR_F_WRITE                  GSSERR_F_BASE + 24
#define GSSERR_F_INIT_DELEGATION        GSSERR_F_BASE + 25
#define GSSERR_F_ACCEPT_DELEGATION      GSSERR_F_BASE + 26
#define GSSERR_F_INQUIRE_BY_OID         GSSERR_F_BASE + 27
#define GSSERR_F_INQUIRE_CONTEXT        GSSERR_F_BASE + 28
#define GSSERR_F_ADD_OID_SET_MEMBER     GSSERR_F_BASE + 29
#define GSSERR_F_CREATE_EMPTY_OID_SET   GSSERR_F_BASE + 30
#define GSSERR_F_TEST_OID_SET_MEMBER    GSSERR_F_BASE + 31

/*
 * GSI minor error code
 *
 * There are three types of GSI minor error codes. There are the codes
 * defined in the gssapi library (as defined below starting at code 100).
 * There are the codes defined in the ssl_utils library (as defined in
 * sslutils.h starting at code 1000) and then there are codes that come up
 * from openssl. These codes are indicated by having the high bit set. See
 * below for more details.
 *
 * Defines for reasons in gssapi library.
 * The match strings defined in gsserr.c
 * These are also used for the minor_status codes.
 * Need to make sure these don't overlap with errors in sslutils.h
 */

#define GSSERR_R_BASE                           100

#define GSSERR_R_HANDSHAKE                      GSSERR_R_BASE + 0
#define GSSERR_R_NO_GLOBUSID                    GSSERR_R_BASE + 1
#define GSSERR_R_PROCESS_CERT                   GSSERR_R_BASE + 2
#define GSSERR_R_MUTUAL_AUTH                    GSSERR_R_BASE + 3
#define GSSERR_R_WRAP_BIO                       GSSERR_R_BASE + 4
#define GSSERR_R_PROXY_VIOLATION                GSSERR_R_BASE + 5
#define GSSERR_R_PROXY_NOT_RECEIVED             GSSERR_R_BASE + 6
#define GSSERR_R_BAD_ARGUMENT                   GSSERR_R_BASE + 7
#define GSSERR_R_IMPEXP_BIO_SSL                 GSSERR_R_BASE + 8
#define GSSERR_R_IMPEXP_NO_CIPHER               GSSERR_R_BASE + 9
#define GSSERR_R_IMPEXP_BAD_LEN                 GSSERR_R_BASE + 10
#define GSSERR_R_CLASS_ADD_EXT                  GSSERR_R_BASE + 11
#define GSSERR_R_EXPORT_FAIL                    GSSERR_R_BASE + 12
#define GSSERR_R_IMPORT_FAIL                    GSSERR_R_BASE + 13
#define GSSERR_R_READ_BIO                       GSSERR_R_BASE + 14
#define GSSERR_R_WRITE_BIO                      GSSERR_R_BASE + 15
#define GSSERR_R_UNEXPECTED_FORMAT              GSSERR_R_BASE + 17
#define GSSERR_R_BAD_DATE                       GSSERR_R_BASE + 20
#define GSSERR_R_BAD_MECH                       GSSERR_R_BASE + 21
#define GSSERR_R_ADD_EXT                        GSSERR_R_BASE + 22
#define GSSERR_R_REMOTE_CERT_VERIFY_FAILED      GSSERR_R_BASE + 23
#define GSSERR_R_OUT_OF_MEMORY                  GSSERR_R_BASE + 24
/* NOTE: Don't go over 1000 here or will conflict with errors in sslutils.h */

/* Old error codes in case anyone is using them */
#define GSSERR_R_IMPEXP_BAD_PARMS       GSSERR_R_BAD_ARGUMENT
#define GSSERR_R_PASSED_NULL_PARAMETER  GSSERR_R_BAD_ARGUMENT

/*
 * Flag used on minor_status to indicate an error from the openssl
 * libraries that we have caught and converted to an error code as
 * defined above, in sslutils.h or in scutils.h
 *
 * If this flag is set then the rest of the 32-bit minor status is
 * the error code as returned from the openssl library.
 */
#define GSI_SSL_ERROR_FLAG                      0x8000

#define GSI_IS_SSL_ERROR(minor)         (minor & GSI_SSL_ERROR_FLAG)


/*
 * we need to distinguish between a token
 * created by us using get_mic vs one using
 * the SSL application data
 * We use this in wrap and unwrap
 * Future versions of SSL may use this
 *
 * Our wraped buffer (integrity only) has
 *
 *  byte  type[1]          = SSL3_RT_GSSAPI_SSLEAY
 *  byte  version_major[1] = 0x03
 *  byte  version_minor[1] = 0
 *  byte  mic_length[2]    = 2 byte length of following mic
 *
 *  byte  mic_seq[8]           = 8 byte sequence number
 *  byte  mic_data_length[4]   = 4 byte length of data
 *  byte  hash[*]          = the hash of variable length
 *
 *  byte  data[*]          = the data being wrapped.
 */

#define SSL3_RT_GSSAPI_SSLEAY                   26



/* These conversions macros are taken from SSL */

#define l2n(l,c)   (*((c)++)=(unsigned char)(((l)>>24)&0xff), \
                    *((c)++)=(unsigned char)(((l)>>16)&0xff), \
                    *((c)++)=(unsigned char)(((l)>> 8)&0xff), \
                    *((c)++)=(unsigned char)(((l)    )&0xff))

#define n2l(c,l)   (l =((unsigned long)(*((c)++)))<<24, \
                    l|=((unsigned long)(*((c)++)))<<16, \
                    l|=((unsigned long)(*((c)++)))<< 8, \
                    l|=((unsigned long)(*((c)++))))

#define n2s(c,s)    (s =((unsigned int)(*((c)++)))<< 8, \
                                 s|=((unsigned int)(*((c)++))))

#define s2n(s,c)    (*((c)++)=(unsigned char)(((s)>> 8)&0xff), \
                     *((c)++)=(unsigned char)(((s)    )&0xff))

/* ssl_locl.h is not installed, so we define a few needed items */

#ifndef SSL_eNULL
#define SSL_eNULL       0x00010000L
#endif

#ifndef SSL_STRONG_MASK
#if  SSLEAY_VERSION_NUMBER >= 0x0090581fL
#define  SSL_STRONG_MASK     0x0000007cL
#else
#define SSL_STRONG_MASK     0x07000000L
#endif
#endif

#ifndef SSL_LOW
#if  SSLEAY_VERSION_NUMBER >= 0x0090581fL
#define SSL_LOW         0x00000010L
#else
#define SSL_LOW         0x01000000L
#endif
#endif
/* Compare OIDs */

#define g_OID_equal(o1,o2) \
        (((o1) == (o2)) || \
         ((o1) && (o2) && \
         ((o1)->length == (o2)->length) && \
         (memcmp((o1)->elements,(o2)->elements,(int) (o1)->length) == 0)))


/**********************************************************************
                               Type definitions
**********************************************************************/

/*
 * The SSL ssl_locl.h is a private headerfile which does
 * not get installed. The ssl3_enc_method is needed for
 * the import/export so we include it here, in orged to avoid
 * changes to the SSLeay code.
 *DEE This needs to be looked at.
 */

#ifndef HEADER_SSL_LOCL_H
typedef struct ssl3_enc_method
{
    int                                 (*enc)();
    int                                 (*mac)();
    int                                 (*setup_key_block)();
    int                                 (*generate_master_secret)();
    int                                 (*change_cipher_state)();
    int                                 (*final_finish_mac)();
    int                                 finish_mac_length;
    int                                 (*cert_verify_mac)();
    unsigned char                       client_finished[20];
    int                                 client_finished_len;
    unsigned char                       server_finished[20];
    int                                 server_finished_len;
    int                                 (*alert_value)();
} SSL3_ENC_METHOD;
#endif

typedef enum {
    GS_CON_ST_HANDSHAKE = 0,
    GS_CON_ST_FLAGS,
    GS_CON_ST_REQ,
    GS_CON_ST_CERT,
    GS_CON_ST_DONE
} gs_con_st_t;

typedef enum
{
    GS_DELEGATION_START,
    GS_DELEGATION_DONE,
    GS_DELEGATION_COMPLETE_CRED,
    GS_DELEGATION_SIGN_CERT
} gs_delegation_state_t;

typedef struct gss_name_desc_struct {
    /* gss_buffer_desc  name_buffer ; */
    gss_OID                             name_oid;
    X509_NAME *                         x509n;
} gss_name_desc ;

typedef struct gss_cred_id_desc_struct {
    proxy_cred_desc                  *pcd;
    gss_name_desc                    *globusid ;
    gss_cred_usage_t                  cred_usage ;
    BIO                              *gs_bio_err ;
} gss_cred_id_desc;

typedef struct gss_ctx_id_desc_struct{
    proxy_verify_desc                   pvd; /* used for verify_callback */
    proxy_verify_ctx_desc               pvxd;
    gss_name_desc *                     source_name;
    gss_name_desc *                     target_name;
    gss_cred_id_desc *                  cred_handle;
    OM_uint32                           ret_flags;
    OM_uint32                           req_flags;
    OM_uint32                           ctx_flags;
    int                                 cred_obtained;
    SSL *                               gs_ssl;
    BIO *                               gs_rbio;
    BIO *                               gs_wbio;
    BIO *                               gs_sslbio;
    gs_con_st_t                         gs_state;
    int                                 locally_initiated;
    time_t                              goodtill;
    /* following used during delegation */

    /* new key for delegated proxy - do we need this now that we have
     * init/accept-delegation
     */
    EVP_PKEY *                          dpkey;
    /* delegated cert */
    X509 *                              dcert;
    /* delegation state */
    gs_delegation_state_t               delegation_state;
} gss_ctx_id_desc ;

/**********************************************************************
                               Global variables
**********************************************************************/

extern
const gss_OID_desc * const gss_mech_globus_gssapi_ssleay;

extern
const gss_OID_desc * const gss_restrictions_extension;

/**********************************************************************
                               Function prototypes
**********************************************************************/

int
ERR_load_gsserr_strings(int);

OM_uint32
convert_minor_codes(const int lib, const int reason);



/**********************************************************************

gssutils.h

Description:
        This header file used internally  to define the gssutils

CVS Information:

    $Source: /home/g5/repository/pyGlobus/src/gsi.i,v $
    $Date: 2003/05/08 17:48:45 $
    $Revision: 1.8.2.2 $
    $Author: mateo $

**********************************************************************/

OM_uint32
gss_copy_name_to_name(
    gss_name_desc **                    output,
    const gss_name_desc *               input);


OM_uint32
gss_create_and_fill_context(
    gss_ctx_id_desc **                  context_handle,
    gss_cred_id_desc *                  cred_handle,
    const gss_cred_usage_t              cred_usage,
    OM_uint32                           req_flags);

OM_uint32
gss_create_and_fill_cred(
    gss_cred_id_t *                     output_cred_handle_P,
    const gss_cred_usage_t              cred_usage,
    X509   *                            ucert,
    EVP_PKEY *                          upkey,
    STACK_OF(X509)  *                   cert_chain,
    BIO *                               bp);

/* following added for ssleay */

OM_uint32
gs_handshake(
    gss_ctx_id_desc*                    context_handle);

OM_uint32
gs_get_token(
    const gss_ctx_id_desc *             context_handle,
    const gss_buffer_t                  output_token);

OM_uint32
gs_put_token(
    const gss_ctx_id_desc *             context_handle,
    const gss_buffer_t                  input_token);

OM_uint32
gs_retrieve_peer(
    gss_ctx_id_desc *                   context_handle,
    const gss_cred_usage_t              cred_usage);

%}
/*--------------------------------------------------------------------
 *  
 *  Start of Python Interface Code
 *
 *------------------------------------------------------------------*/


#define STACK_OF(A) STACK



typedef struct proxy_cred_desc_struct {
    X509 *                              ucert;
    EVP_PKEY *                          upkey;
    STACK_OF(X509) *                    cert_chain;
    SSL_CTX *                           gs_ctx;
    /* smart card session handle */
    unsigned long                       hSession;
    /* private key session handle */
    unsigned long                       hPrivKey;
    char *                              certdir;
    char *                              certfile;
    int                                 num_null_enc_ciphers;
    int                                 type;  
    int                                 owner;
} proxy_cred_desc;






/******************************************************************
****************End of old sslutils includes and typedef's ********
******************************************************************/

/*****************************************************************
***************Begin gsi includes and typedef's ******************
*****************************************************************/


%include ./src/pyGlobus-typemaps.i

typedef int globus_bool_t;
typedef void* gss_name_t;
typedef void* gss_cred_id_t;
typedef void* gss_ctx_id_t;
typedef int gss_cred_usage_t;
typedef unsigned int gss_qop_t; /* these next two will need to be ifdef'd */
typedef unsigned int OM_uint32; /* for other platforms where they're not uint's */

typedef struct gss_OID_desc_struct {
      OM_uint32 length;
      void       *elements;
} gss_OID_desc,  *gss_OID;

typedef struct gss_OID_set_desc_struct  {
      size_t  count;
      gss_OID elements;
} gss_OID_set_desc,  *gss_OID_set;


typedef struct gss_buffer_desc_struct {
      size_t length;
      void  *value;
} gss_buffer_desc,  *gss_buffer_t;

%constant GSS_EXPORT_CRED_TO_BUFFER=0;
%constant GSS_EXPORT_CRED_EXTERNAL=1;

typedef struct gss_buffer_set_desc_struct {
    size_t                              count;
    gss_buffer_desc *                   elements;
} gss_buffer_set_desc, *gss_buffer_set_t;

%types(gss_OID*);

%constant GSS_C_GSS_CODE=1;
%constant GSS_C_MECH_CODE=2;

%constant GSS_S_COMPLETE=0;
%constant GSS_C_BOTH=0;
%constant GSS_C_INITIATE=1;
%constant GSS_C_ACCEPT=2;

%constant GSS_C_DELEG_FLAG=1;
%constant GSS_C_MUTUAL_FLAG=2;
%constant GSS_C_REPLAY_FLAG=4;
%constant GSS_C_SEQUENCE_FLAG=8;
%constant GSS_C_CONF_FLAG=6;
%constant GSS_C_INTEG_FLAG=32;
%constant GSS_C_ANON_FLAG=64;
%constant GSS_C_PROT_READY_FLAG=128;
%constant GSS_C_TRANS_FLAG=256;

%constant GSS_C_GLOBUS_LIMITED_DELEG_PROXY_FLAG=4096;
%constant GSS_C_GLOBUS_LIMITED_PROXY_FLAG        = 192;
%constant GSS_C_GLOBUS_SSL_COMPATABLE          = 16384;

%constant GSS_C_CALLING_ERROR_OFFSET=24;
%constant GSS_C_ROUTINE_ERROR_OFFSET=16;
%constant GSS_C_SUPPLEMENTARY_OFFSET=0;
%constant GSS_C_CALLING_ERROR_MASK=0377ul;
%constant GSS_C_ROUTINE_ERROR_MASK=0377ul;
%constant GSS_C_SUPPLEMENTARY_MASK=0177777ul;
%constant GSS_S_BAD_MECH=(1ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_BAD_NAME=(2ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_BAD_NAMETYPE=(3ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_BAD_BINDINGS=(4ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_BAD_STATUS=(5ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_BAD_SIG=(6ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_BAD_MIC=GSS_S_BAD_SIG;
%constant GSS_S_NO_CRED=(7ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_NO_CONTEXT=(8ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_DEFECTIVE_TOKEN=(9ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_DEFECTIVE_CREDENTIAL=(10ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_CREDENTIALS_EXPIRED=(11ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_CONTEXT_EXPIRED=(12ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_FAILURE=(13ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_BAD_QOP=(14ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_UNAUTHORIZED=(15ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_UNAVAILABLE=(16ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_DUPLICATE_ELEMENT=(17ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_NAME_NOT_MN=(18ul << GSS_C_ROUTINE_ERROR_OFFSET);
%constant GSS_S_CONTINUE_NEEDED=(1ul << (GSS_C_SUPPLEMENTARY_OFFSET + 0));
%constant GSS_S_DUPLICATE_TOKEN=(1ul << (GSS_C_SUPPLEMENTARY_OFFSET + 1));
%constant GSS_S_OLD_TOKEN=(1ul << (GSS_C_SUPPLEMENTARY_OFFSET + 2));
%constant GSS_S_UNSEQ_TOKEN=(1ul << (GSS_C_SUPPLEMENTARY_OFFSET + 3));
%constant GSS_S_GAP_TOKEN=(1ul << (GSS_C_SUPPLEMENTARY_OFFSET + 4));

%apply char **OUTPUT { char **charRet };

extern gss_OID GSS_C_NT_USER_NAME;
extern gss_OID GSS_C_NT_MACHINE_UID_NAME;
extern gss_OID GSS_C_NT_STRING_UID_NAME;
extern gss_OID GSS_C_NT_HOSTBASED_SERVICE_X;
extern gss_OID GSS_C_NT_HOSTBASED_SERVICE;
extern gss_OID GSS_C_NT_ANONYMOUS;
extern gss_OID GSS_C_NT_EXPORT_NAME;

extern int
globus_gss_assist_gridmap(char * globusidp, char ** charRet);

extern int
globus_gss_assist_map_local_user(char *local_user, char** charRet);

extern int
globus_gss_assist_userok(char *globusid, char* userid);

%inline %{
/*************************************************************
************old sslutils.i functions ************************
**************************************************************/
void
kpcallback(int p, int n) {
    char c='B';
    if (p == 0) c='.';
    if (p == 1) c='+';
    if (p == 2) c='*';
    if (p == 3) c='\n';
    fputc(c,stderr);
}
void
free_X509(X509_REQ** req){
    if(req){
	X509_free((X509*)*req);
    }
    return;
}

void
free_EVP_PKEY(EVP_PKEY** pkey){
    if(pkey){
	EVP_PKEY_free(*pkey);
    }
}

EVP_PKEY**
malloc_EVP_PKEY() {
    return (EVP_PKEY**)malloc(sizeof(EVP_PKEY**));
}

X509_REQ**
malloc_X509_REQ() {
    return (X509_REQ**)malloc(sizeof(X509_REQ**));
}

gss_cred_id_t*
get_initialized_gss_cred() {
    gss_cred_id_desc     **output_cred_handle = NULL;

    output_cred_handle  = (gss_cred_id_desc**)globus_malloc(sizeof(gss_cred_id_desc**));
    if (output_cred_handle == NULL) {
        GSSerr(GSSERR_F_ACQUIRE_CRED, GSSERR_R_OUT_OF_MEMORY);
        goto err;
    }
    (*output_cred_handle) = (gss_cred_id_desc*)globus_malloc(sizeof(gss_cred_id_desc*));
    (*output_cred_handle)->globusid = NULL;
    if (((*output_cred_handle)->pcd = proxy_cred_desc_new()) == NULL) {
        goto err;
    }
    if( (*output_cred_handle)->pcd->ucert == NULL)
	printf("XXX NULL\n");
    return (gss_cred_id_t*)output_cred_handle;
 err:
    //return GSS_S_FAILURE ;
    return NULL;
}

int
genreq(gss_cred_id_t* issuer_cred, gss_cred_id_t* sub_cred,
       X509_REQ** req, EVP_PKEY **npkey, const int bits) {
    PyObject          *tuple;
    proxy_cred_desc   *sub_pcd            = NULL;
    //proxy_cred_desc   *sub_pcd            = (*(gss_cred_id_desc**)sub_cred)->pcd;
    proxy_cred_desc   *ipcd               = (*(gss_cred_id_desc**)issuer_cred)->pcd;
    gss_cred_id_desc **output_cred_handle = (gss_cred_id_desc**) sub_cred;


    *npkey = (EVP_PKEY*)malloc(sizeof(EVP_PKEY*));
    *req = (X509_REQ*)malloc(sizeof(X509_REQ*));
    if (proxy_genreq(ipcd->ucert, req, npkey, bits,(int (*)())kpcallback, sub_pcd)){
        goto err;
    }
    X509_REQ_print_fp(stderr, *req);
    return 0;
 err:
    return -1;
}


void
set_pcd_upkey(gss_cred_id_t* cred, EVP_PKEY** pkey) {
    (*(gss_cred_id_desc**)(cred))->pcd->upkey = *pkey;
    return;
}

int
set_pcd_outfile(const char* outfile, gss_cred_id_t* cred) {
    int   err = 0;
    char *certfile    = NULL;
    char *keyfile     = NULL;
    char *certdir     = NULL;
    char *certcafile  = NULL;
    char *outputfile  = NULL;
    gss_cred_id_desc **    output_cred_handle = (gss_cred_id_desc**) cred;
    (*output_cred_handle)->globusid = NULL;
    (*output_cred_handle)->pcd->certfile = NULL;
    if(outfile == NULL) {
        if ( proxy_get_filenames((*output_cred_handle)->pcd,
                             0,
                             &certcafile,
                             &certdir,
                             &outputfile,
                             &certfile,
                             &keyfile) )
            goto err;
	(*output_cred_handle)->pcd->certfile = (char*)malloc(strlen(outfile) + 1);
	strcpy((*output_cred_handle)->pcd->certfile, outfile);
        free(certfile);
        free(keyfile);
        free(certdir);
        free(certcafile);
	free(outputfile);
    }
    if(outfile) {
	(*output_cred_handle)->pcd->certfile = (char*)malloc(strlen(outfile) + 1);
	strcpy((*output_cred_handle)->pcd->certfile, outfile);
    }
    return 0;
err:
    return -1;
}

/*
 * verify if the cert is good, i.e. is signed by one of the
 * trusted CAs.
 */
int
verify_pcd(gss_cred_id_t* cred) {
    int debug = 1;
    proxy_verify_desc     pvd;
    proxy_verify_ctx_desc pvxd;
    gss_cred_id_desc **    output_cred_handle = (gss_cred_id_desc**) cred;


    proxy_verify_ctx_init(&pvxd);
    proxy_verify_init(&pvd, &pvxd);
    pvxd.certdir = (*output_cred_handle)->pcd->certdir;
    if (proxy_verify_cert_chain((*output_cred_handle)->pcd->ucert, 
				(*output_cred_handle)->pcd->cert_chain,
				&pvd)) {
        fprintf(stderr,"verify OK\n");
    } else {
        goto err;
    }

    return 0;
err:
    {
        unsigned long l;
        char buf[256];
#if SSLEAY_VERSION_NUMBER  >= 0x00904100L
        const char *file;
#else
        char *file;
#endif
        const char *data;
        int line;
        int flags;
        printf("Error: \n");
        while ( ERR_peek_error() != 0 )
        {
            printf("-->\n");
            l=ERR_get_error_line_data(&file,&line,&data,&flags);
            if (debug)

                fprintf(stderr,
                        "%s:%s:%d%s\n",
                        ERR_error_string(l,buf),
                        file,
                        line,
                        data);

            else
                fprintf(stderr,
                        "%s%s\nFunction: %s\n",
                        ERR_reason_error_string(l),
                        data,
                        ERR_func_error_string(l));
        }
    }
    return -1;
}

char*
get_pcd_certdir(gss_cred_id_t* cred) {
    int                    i;
    char                  *s = NULL;
    gss_cred_id_desc     **output_cred_handle = (gss_cred_id_desc**) cred;
    if((*output_cred_handle)->pcd->certdir != NULL) {
       s = (*output_cred_handle)->pcd->certdir;
    }
    return s;
}

void
set_pcd_certdir(gss_cred_id_t* cred, char* certdir) {
    gss_cred_id_desc     **output_cred_handle = (gss_cred_id_desc**) cred;
    (*output_cred_handle)->pcd->certdir = (char*)strdup(certdir);
    return;
}

int
write_pcd_to_file(gss_cred_id_t* cred) {
    int                   debug = 1;
    BIO *                 bp = NULL;
    BIO *                 bio_err;
    X509 *                ncert = NULL;
    FILE                 *fpout = NULL;
    gss_cred_id_desc    **output_cred_handle = (gss_cred_id_desc**)cred;

    printf("outfile = %s\n", (*output_cred_handle)->pcd->certfile);
    fpout=fopen((*output_cred_handle)->pcd->certfile,"w");
    if (fpout == NULL)
    {
        PRXYerr(PRXYERR_F_LOCAL_CREATE,PRXYERR_R_PROBLEM_PROXY_FILE);
        ERR_add_error_data(2,"\n        Open failed for File=", (*output_cred_handle)->pcd->certfile);
        goto err;
    }

    if ((bio_err=BIO_new(BIO_s_file())) != NULL)
        BIO_set_fp(bio_err,stderr,BIO_NOCLOSE);

    if ((bp=BIO_new(BIO_s_file())) != NULL)
    {
        BIO_set_fp(bp,fpout,BIO_NOCLOSE);
    }
    printf("marshal\n");
    if (proxy_marshal_bp(bp, (*output_cred_handle)->pcd->ucert, (*output_cred_handle)->pcd->upkey,
			 ncert, (*output_cred_handle)->pcd->cert_chain))
    {
        goto err;
    }
    return 0;
err:
    {
        unsigned long l;
        char buf[256];
#if SSLEAY_VERSION_NUMBER  >= 0x00904100L
        const char *file;
#else
        char *file;
#endif
        const char *data;
        int line;
        int flags;
        printf("Error: \n");
        while ( ERR_peek_error() != 0 )
        {
            printf("-->\n");
            l=ERR_get_error_line_data(&file,&line,&data,&flags);
            if (debug)

            fprintf(stderr,
                    "%s:%s:%d%s\n",
                    ERR_error_string(l,buf),
                    file,
                    line,
                    data);

            else
                fprintf(stderr,
                        "%s%s\nFunction: %s\n",
                        ERR_reason_error_string(l),
                        data,
                        ERR_func_error_string(l));
        }
    }
    return -1;
}

void
free_gpi(gss_cred_id_desc *cred)
{
globus_free(cred->pcd);
globus_free(cred);
}

PyObject*
grid_proxy_init(int verbose) {
    PyObject *            ptr;
    PyObject *            tuple;
    int                   debug       = verbose;
    int                   quiet       = (verbose)?0:1;
    int                   ret_status  = 0;
    int                   bits        = 512;   /* default proxy to 512 bits */
    int                   hours       = 12;    /* default to a 12 hour cert */
    int                   limit_proxy = 0;     /* dont restrict the proxy */
    int                   verify      = 1;
    int                   i;
    int                   j;
    int                   cert_chain_length = 0;
    char *                certfile    = NULL;
    char *                keyfile     = NULL;
    char *                outfile     = NULL;
    char *                certdir     = NULL;
    char *                certcafile  = NULL;
    char *                pin         = NULL;
    char *                argp;
    char *                program;
    proxy_cred_desc *     pcd         = NULL;
    proxy_cred_desc *     pcd2        = NULL;
    gss_cred_id_desc *    output_cred_handle = NULL;
    proxy_verify_desc     pvd;
    proxy_verify_ctx_desc pvxd;
    BIO *                 bio_err;
    X509 *                xcert;
    time_t                time_after;
    time_t                time_now;
    time_t                time_diff;
    time_t                time_after_proxy;
    ASN1_UTCTIME *        asn1_time = NULL;
#ifdef CLASS_ADD
    void *                class_add_buf = NULL;
    size_t                class_add_buf_len = 0;
#endif
    X509_STORE *         cert_ctx    = NULL;
    X509_LOOKUP *        lookup      = NULL;
    X509_STORE_CTX       csc;

    int                                 status = -1;
    FILE *                              fpout = NULL;
    X509 *                              ncert = NULL;
    EVP_PKEY *                          npkey, npkey2;
    X509_REQ *                          req, req2;
    BIO *                               bp = NULL;
    STACK_OF(X509_EXTENSION) *          extensions = NULL;
    X509_EXTENSION *                    ex = NULL;

    int (*pw_cb)() = NULL;
    output_cred_handle = (gss_cred_id_desc*)globus_malloc(sizeof(gss_cred_id_desc));

    ERR_load_prxyerr_strings(0);
    SSLeay_add_ssl_algorithms();

    EVP_set_pw_prompt("Enter GRID pass phrase:");
    
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "grid_proxy_init: PyTuple_New()");
        return NULL;
    }

    if ((pcd = proxy_cred_desc_new()) == NULL)
        goto err;

    if ( proxy_get_filenames(pcd,
                             0,
                             &certcafile,
                             &certdir,
                             &outfile,
                             &certfile,
                             &keyfile) )
        goto err;

        if (debug) {
                printf("Files being used:\n");
                printf("    cert_file: %s\n", certcafile ? certcafile:"none");
                printf("    cert_dir : %s\n", certdir ? certdir:"none");
                printf("    proxy    : %s\n", outfile ? outfile:"none");
                printf("    user_cert: %s\n", certfile ? certfile:"none");
                printf("    user_key : %s\n", keyfile ? keyfile:"none");
        }

    if (certdir)
                pcd->certdir = (char*)strdup(certdir);

    if (!strncmp(certfile,"SC:",3))
        EVP_set_pw_prompt("Enter card pin:");
    else
        EVP_set_pw_prompt("Enter GRID pass phrase:");

    if (proxy_load_user_cert(pcd, certfile, pw_cb, NULL))
                goto err;

    if (!quiet)
    {
        char *s = NULL;
        s = X509_NAME_oneline(X509_get_subject_name(pcd->ucert),NULL,0);
        printf("Your identity: %s\n", s);
        free(s);
    }

    if (!strncmp(keyfile,"SC:",3))
        EVP_set_pw_prompt("Enter card pin:");

    if (proxy_load_user_key(pcd, keyfile, pw_cb, NULL))
        goto err;
    else
        printf("user_key loaded\n");

    //if (strncmp("SC:",certfile,3) && !strcmp(certfile, keyfile))
    //{
        if (pcd->cert_chain == NULL)
            pcd->cert_chain = sk_X509_new_null();
        if (proxy_load_user_proxy(pcd->cert_chain, certfile, NULL) < 0)
            goto err;
        else
            printf("user_proxy cert_chain\n");
    //}

    /*
     * verify if the cert is good, i.e. is signed by one of the
     * trusted CAs.
     */
    if (verify)
    {
        proxy_verify_ctx_init(&pvxd);
        proxy_verify_init(&pvd, &pvxd);
        pvxd.certdir = certdir;
        if (proxy_verify_cert_chain(pcd->ucert,pcd->cert_chain,&pvd)) {
            fprintf(stderr,"verify OK\n");
        } else {
            fprintf(stderr,"verify failed\n");
            goto err;
        }
    }

    asn1_time = ASN1_UTCTIME_new();
    X509_gmtime_adj(asn1_time,0);
    time_now = ASN1_UTCTIME_mktime(asn1_time);
    if (!quiet)
    {
        printf("Creating proxy ");
        fflush(stdout);
    }

    if (proxy_genreq(pcd->ucert,&req,&npkey,bits,
                     (int (*)())kpcallback, pcd))
    {
        goto err;
    }
    printf("generated request\n");
    /*
     * Add proxy extensions
     */
#ifdef DEBUG
    fprintf(stderr,"Adding Extensions to request\n");
#endif
    if ((extensions = sk_X509_EXTENSION_new_null()) == NULL)
    {
        PRXYerr(PRXYERR_F_PROXY_SIGN,PRXYERR_R_CLASS_ADD_EXT);
        goto err;
    }

#ifdef CLASS_ADD
    if (class_add_buf && class_add_buf_len > 0)
    {
        if ((ex = proxy_extension_class_add_create(class_add_buf,
                                                   class_add_buf_len)) == NULL)
        {
            PRXYerr(PRXYERR_F_PROXY_SIGN,PRXYERR_R_CLASS_ADD_EXT);
            goto err;
        }

        if (!sk_X509_EXTENSION_push(extensions, ex))
        {
            PRXYerr(PRXYERR_F_PROXY_SIGN,PRXYERR_R_CLASS_ADD_EXT);
            goto err;
        }
    }
#endif

    /*
     *   req   - request
     *   ncert - certificate
     *   npkey - key
     *   pcd->ucert,  pcd->cert_chain
     */
    if (proxy_sign(pcd->ucert,
                   pcd->upkey,
                   req,
                   &ncert,
                   hours*60*60,
                   extensions,
                   limit_proxy))
    {
        printf("proxy_sign\n");
        goto err;
    }
    else
        printf("gpi signed cert\n");
       
  
    output_cred_handle->pcd = (proxy_cred_desc *)globus_malloc(sizeof(proxy_cred_desc));  
    output_cred_handle->pcd->ucert = ncert;
    output_cred_handle->pcd->upkey = npkey;
    //insert cert chain
    
    	
    output_cred_handle->pcd->cert_chain = sk_X509_new_null();
    
    
    i = sk_X509_insert(output_cred_handle->pcd->cert_chain, 
		       pcd->ucert, 
		       sk_X509_num(pcd->cert_chain));

      
    ptr = SWIG_NewPointerObj((void*)output_cred_handle, 
				 SWIGTYPE_p_gss_cred_id_desc, 0);
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
    PyTuple_SET_ITEM(tuple, 1, ptr);
    printf("Returning tuple\n");
    return tuple;
err:
    {
        unsigned long l;
        char buf[256];
	char tmpstr[356];
#if SSLEAY_VERSION_NUMBER  >= 0x00904100L
        const char *file;
#else
        char *file;
#endif
        const char *data;
        int line;
        int flags;
        //printf("Error: \n");
        while ( ERR_peek_error() != 0 )
        {
            //printf("-->\n");
            l=ERR_get_error_line_data(&file,&line,&data,&flags);
	    
            if (debug)

            sprintf(tmpstr,
                    "%s:%s:%d%s\n",
                    ERR_error_string(l,buf),
                    file,
                    line,
                    data);

            
        }
    
    
    PyTuple_SET_ITEM(tuple,0, Py_BuildValue("i", -1));
    PyTuple_SET_ITEM(tuple,1, Py_BuildValue("s", tmpstr));
    return tuple;
    }
}

int
sign_request(const gss_cred_id_t* issuer_cred,
	     gss_cred_id_t*       sub_cred,
	     X509_REQ**           req)
{
    int bits = 512;
    int limit_proxy = 0;
    int hours = 12;
    int i;
    int debug = 1;
    int                   cert_chain_length = 0;
    char *                certfile    = NULL;
    char *                keyfile     = NULL;
    char *                outfile     = NULL;
    char *                certdir     = NULL;
    char *                certcafile  = NULL;
    char *                s           = NULL;

    X509 *                              ncert = NULL;
    X509 *                              icert = NULL;
    STACK_OF(X509_EXTENSION) *          extensions = NULL;
    X509_EXTENSION *                    ex = NULL;
    proxy_cred_desc* ipcd    = (*(gss_cred_id_desc**)issuer_cred)->pcd;
    proxy_cred_desc* sub_pcd = (*(gss_cred_id_desc**)sub_cred)->pcd;

    // SIGN
    // int sign_request(issuer, sub_cred, request, hours, limited)
    fprintf(stderr,"Adding Extensions to request\n");
    if ((extensions = sk_X509_EXTENSION_new_null()) == NULL)
    {
        PRXYerr(PRXYERR_F_PROXY_SIGN,PRXYERR_R_CLASS_ADD_EXT);
        goto err;
    }

#ifdef CLASS_ADD
    if (class_add_buf && class_add_buf_len > 0)
    {
        if ((ex = proxy_extension_class_add_create(class_add_buf,
                                                   class_add_buf_len)) == NULL)
        {
            PRXYerr(PRXYERR_F_PROXY_SIGN,PRXYERR_R_CLASS_ADD_EXT);
            goto err;
        }

        if (!sk_X509_EXTENSION_push(extensions, ex))
        {
            PRXYerr(PRXYERR_F_PROXY_SIGN,PRXYERR_R_CLASS_ADD_EXT);
            goto err;
        }
    }
#endif
    if (proxy_sign(ipcd->ucert,
                   ipcd->upkey,
                   (X509_REQ*)*req,
                   &ncert,
                   hours*60*60,
                   extensions,
                   limit_proxy))
    {
	printf("proxy_sign\n");
        goto err;
    }
    else
	printf("signed cert\n");

//********************************************
// CONSTRUCT PCD
    if( (*(gss_cred_id_desc**)sub_cred)->pcd == NULL )
	printf("1 NULL\n");
    printf("1 X\n");
    if( (*(gss_cred_id_desc**)sub_cred)->pcd->ucert == NULL )
	printf("2 NULL\n");
    printf("2 X\n");
    if( sub_pcd->ucert == NULL )
	printf("NULL\n");
    printf(" sub_pcd->ucert = ncert\n");
    sub_pcd->ucert = ncert;
    printf(" sub_pcd->ucert = ncert\n");

//********************************************
//build_cert_chain(issuer, sub_cred)

    if (sub_pcd->cert_chain == NULL)
	sub_pcd->cert_chain = sk_X509_new_null();
    printf(" sub_pcd->cert_chain = sk_X509_new_null()\n");


    if(ipcd->cert_chain != NULL)
    {
	cert_chain_length = sk_X509_num(ipcd->cert_chain);
	printf("XCredPath: %d\n", cert_chain_length);
    }
    for(i=cert_chain_length-1;i>=0;i--)
    {
	icert = sk_X509_value(ipcd->cert_chain,i);
	s = X509_NAME_oneline(X509_get_subject_name(icert), NULL, 0);
	printf("\t%d) %s\n", i, s);
	free(s);
	sk_X509_push(sub_pcd->cert_chain, icert);
	//i = sk_X509_insert(sub_pcd->cert_chain, pcd->ucert, sk_X509_num(pcd->cert_chain));
    }
    sk_X509_push(sub_pcd->cert_chain, ipcd->ucert);
    (*(gss_cred_id_desc**)sub_cred)->pcd = sub_pcd;
    return 0;

err:
    {
        unsigned long l;
        char buf[256];
#if SSLEAY_VERSION_NUMBER  >= 0x00904100L
        const char *file;
#else
        char *file;
#endif
        const char *data;
        int line;
        int flags;
	printf("Error: \n");
        while ( ERR_peek_error() != 0 )
        {
	    printf("-->\n");
            l=ERR_get_error_line_data(&file,&line,&data,&flags);
            if (debug)

            fprintf(stderr,
                    "%s:%s:%d%s\n",
                    ERR_error_string(l,buf),
                    file,
                    line,
                    data);

            else
                fprintf(stderr,
                        "%s%s\nFunction: %s\n",
                        ERR_reason_error_string(l),
                        data,
                        ERR_func_error_string(l));
        }
    }
    return -1;
}

PyObject*
get_cred_path(gss_cred_id_t* cred) {
    char                  *s;
    int                    i, cert_chain_length;
    X509                  *cert = NULL;
    gss_cred_id_desc     **output_cred_handle = (gss_cred_id_desc**) cred;
    PyObject              *tuple;


    if((*output_cred_handle)->pcd->cert_chain != NULL) {
        cert_chain_length = sk_X509_num((*output_cred_handle)->pcd->cert_chain);
        printf("CredPath: %d\n", cert_chain_length);
    }
    else {
        printf("cert chain is null\n");
    }
    if (!(tuple=PyTuple_New(cert_chain_length+1)))
    {
        PyErr_SetString(PyExc_RuntimeError,
                        "import_name: PyTuple_New()");
        return NULL;
    }
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", cert_chain_length));
    for(i=cert_chain_length-1;i>=0;i--) {
        cert = sk_X509_value((*output_cred_handle)->pcd->cert_chain,i);
        s = X509_NAME_oneline(X509_get_subject_name(cert), NULL, 0);
        printf("\t%d) %s\n", i, s);
        PyTuple_SET_ITEM(tuple, (i+1), Py_BuildValue("s#", s));
        free(s);
    }
    return tuple;
}

int
set_cred_globusuid(gss_cred_id_t* cred)
{
    gss_cred_id_desc     **output_cred_handle = (gss_cred_id_desc**) cred;
    /*
     * get the globusid, which is the subject name - any proxy entries
     */

    (*output_cred_handle)->globusid = (gss_name_desc*) malloc(sizeof(gss_name_desc)) ;
    if ((*output_cred_handle)->globusid == NULL)
    {
        GSSerr(GSSERR_F_ACQUIRE_CRED, GSSERR_R_OUT_OF_MEMORY);
        goto err;
    }
    (*output_cred_handle)->globusid->name_oid = GSS_C_NO_OID;
    (*output_cred_handle)->globusid->x509n =
        X509_NAME_dup(X509_get_subject_name((*output_cred_handle)->pcd->ucert));

    if ((*output_cred_handle)->globusid->x509n == NULL)
    {
        GSSerr(GSSERR_F_ACQUIRE_CRED,GSSERR_R_PROCESS_CERT);
        goto err;
    }
    printf("name : %s\n", (*output_cred_handle)->globusid->x509n);
    return 0;

    err:
    return -1;
}

/***********old gsi.i functions****************/
gss_cred_id_t
get_no_cred_const() {
    return GSS_C_NO_CREDENTIAL;
}

gss_cred_id_t*
get_no_cred_const_ptr() {
    gss_cred_id_t* cred;
    cred = (gss_cred_id_t*)globus_malloc(sizeof(gss_cred_id_t));
    *cred = GSS_C_NO_CREDENTIAL;
    return cred;
}

PyObject*
release_cred(gss_cred_id_t* cred) {
    OM_uint32 major_status;
    OM_uint32 minor_status;
    PyObject* tuple;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, 
                        "release_cred: PyTuple_New()");
        return NULL;
    }
    major_status = gss_release_cred(&minor_status, cred);
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major_status));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor_status));
    return tuple;
} 

	
PyObject*
assist_acquire_cred(gss_cred_usage_t usage, gss_cred_id_t* cred) {
    OM_uint32 major_status;
    OM_uint32 minor_status; 
    PyObject* tuple;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, 
                        "assist_acquire_cred: PyTuple_New()");
        return NULL;
    }
    major_status = globus_gss_assist_acquire_cred(&minor_status,
						  usage, cred);
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major_status));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor_status));
    return tuple;
}

PyObject*
import_name(PyObject* nameStr, const gss_OID nameType) {
    OM_uint32 major_status;
    OM_uint32 minor_status;
    PyObject *tuple, *ptr;
    gss_buffer_t buf;
    gss_name_t* name;
    if (!(tuple=PyTuple_New(3)))
    {
        PyErr_SetString(PyExc_RuntimeError, 
                        "import_name: PyTuple_New()");
        return NULL;
    }
    if (!PyString_Check(nameStr)) {
	PyErr_SetString(PyExc_TypeError, "not a string");
	return NULL;
    }
    buf = (gss_buffer_desc*)globus_malloc(sizeof(gss_buffer_desc));
    buf->length = PyString_Size(nameStr);
    buf->value = PyString_AsString(nameStr);
    major_status = gss_import_name(&minor_status, buf, nameType, name);
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major_status));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor_status));
    if (major_status != GSS_S_COMPLETE) {
	Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 2, Py_None);
    }
    else {
	ptr = SWIG_NewPointerObj((void*)name, SWIGTYPE_p_gss_name_t, 0);
	PyTuple_SET_ITEM(tuple, 2, ptr);
    }
    globus_free(buf);
    return tuple;
}



PyObject*
display_name(gss_name_t* name) {
    PyObject* tuple;
    OM_uint32         major = 0;
    OM_uint32         minor = 0;
    gss_buffer_desc   buffer_desc = GSS_C_EMPTY_BUFFER;
    gss_buffer_t      buffer = &buffer_desc;

    /*
    //gss_buffer_desc sname;
    //gss_buffer_desc sname = GSS_C_EMPTY_BUFFER;
    //gss_OID* type = NULL;

    //gss_display_name
    gss_buffer_t output_name;
    const gss_name_desc* input_name = (gss_name_desc*) name;

    if (!(input_name) || !(input_name->x509n) ||
	!(output_name)) {
	printf("BAD INPUT\n");
	major = GSS_S_FAILURE;
    }
    else{
	printf("GOOD INPUT\n");
	output_name->value = X509_NAME_oneline(input_name->x509n,NULL,0);
	output_name->length = strlen(output_name->value);
	major = GSS_S_COMPLETE ;
    }
    */


    if (!(tuple=PyTuple_New(3)))
    {
        PyErr_SetString(PyExc_RuntimeError, "display_name: PyTuple_New()");
        return NULL;
    }
    major = gss_display_name(&minor, *name, buffer, NULL);
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor));
    if (major != GSS_S_COMPLETE) {
	Py_INCREF(Py_None);
	PyTuple_SET_ITEM(tuple, 2, Py_None);
    }
    else {
	PyTuple_SET_ITEM(tuple, 2, Py_BuildValue("s#", (char*)buffer_desc.value, 
						 (int)buffer_desc.length));
	gss_release_buffer(&minor, buffer);
    }
    return tuple;
}
 
PyObject*
release_name(gss_name_t* name) {
    PyObject* tuple;
    OM_uint32 major_status;
    OM_uint32 minor_status;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, 
                        "release_name: PyTuple_New()");
        return NULL;
    }
    major_status = gss_release_name(&minor_status, name);
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major_status));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor_status));
    return tuple;
}

PyObject*
init_sec_context(const gss_cred_id_t initiator_cred, gss_ctx_id_t* ctx,
		 const gss_name_t target_name, const gss_OID mech_type,
		 OM_uint32 req_flags, OM_uint32 time_req, gss_buffer_t input_token) {
    PyObject* tuple;
    OM_uint32 major_status;
    OM_uint32 minor_status;
    gss_buffer_desc output_token;
    gss_OID actual_mech;
    OM_uint32 ret_flags;
    OM_uint32 time_rec;
    if (!(tuple=PyTuple_New(3)))
    {
        PyErr_SetString(PyExc_RuntimeError, 
                        "init_sec_context: PyTuple_New()");
        return NULL;
    }
    major_status = gss_init_sec_context(&minor_status, initiator_cred, ctx,
					target_name, mech_type, req_flags,
					time_req, GSS_C_NO_CHANNEL_BINDINGS, 
					input_token, &actual_mech, &output_token,
					&ret_flags, &time_rec);
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major_status));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor_status));
    if (major_status != GSS_S_COMPLETE && major_status != GSS_S_CONTINUE_NEEDED) {
	Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 2, Py_None);
    }
    else {
	PyTuple_SET_ITEM(tuple, 2, Py_BuildValue("s#", output_token.value,
						 output_token.length));
    }
    return tuple;
} 

PyObject *
accept_sec_context(gss_ctx_id_t * context_handle,gss_cred_id_t acceptor_cred_handle, 
		   gss_buffer_t input_token,gss_channel_bindings_t  input_chan_bindings,
		   gss_name_t *src_name, gss_OID* mech_type, gss_buffer_t output_token,
		   gss_cred_id_t * delegated_cred_handle){

    PyObject * tuple;
    OM_uint32 major_status;
    OM_uint32 minor_status;
    OM_uint32 ret_flags;
    OM_uint32 time_rec;

    if (!(tuple=PyTuple_New(3)))
    {
        PyErr_SetString(PyExc_RuntimeError, 
                        "accept_sec_context: PyTuple_New()");
        return NULL;
    }
   
    major_status = gss_accept_sec_context(&minor_status, context_handle, acceptor_cred_handle,
    					  input_token, input_chan_bindings, src_name,
					  mech_type, output_token, &ret_flags,
					  &time_rec,delegated_cred_handle);

    				  
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major_status));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor_status));


	return tuple;
}


gss_ctx_id_t*
create_context() {
    PyObject* tuple;
    gss_ctx_id_t* ctx;
    ctx = (gss_ctx_id_t*)globus_malloc(sizeof(gss_ctx_id_t));
    *ctx = GSS_C_NO_CONTEXT;
    return ctx;
}

PyObject*
delete_context(gss_ctx_id_t* ctx) {
    PyObject* tuple;
    OM_uint32 major, minor;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, 
                        "delete_context: PyTuple_New()");
        return NULL;
    }
    major = gss_delete_sec_context(&minor, ctx, GSS_C_NO_BUFFER);
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor));
    return tuple;
}
 
PyObject*
inquire_context(gss_ctx_id_t* ctx) {
    PyObject *tuple, *ptr, *ptr2, *ptr3;
    OM_uint32 major, minor, lifetime, ctxFlags;
    int local, open;
    gss_name_t *srcName, *targetName;
    gss_OID* mech;
    if (!(tuple=PyTuple_New(9)))
    {
        PyErr_SetString(PyExc_RuntimeError, 
                        "inquire_context: PyTuple_New()");
        return NULL;
    }
    mech = (gss_OID*)globus_malloc(sizeof(gss_OID));
    srcName = (gss_name_t*)globus_malloc(sizeof(gss_name_t));
    targetName = (gss_name_t*)globus_malloc(sizeof(gss_name_t));
    major = gss_inquire_context(&minor, *ctx, srcName, targetName,
				&lifetime, mech, &ctxFlags, &local, &open);
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor));
    if (major != GSS_S_COMPLETE) {
	Py_INCREF(Py_None);
	PyTuple_SET_ITEM(tuple, 2, Py_None);
	Py_INCREF(Py_None);
	PyTuple_SET_ITEM(tuple, 3, Py_None);
	Py_INCREF(Py_None);
	PyTuple_SET_ITEM(tuple, 4, Py_None);
	Py_INCREF(Py_None);
	PyTuple_SET_ITEM(tuple, 5, Py_None);
	Py_INCREF(Py_None);
	PyTuple_SET_ITEM(tuple, 6, Py_None);
	Py_INCREF(Py_None);
	PyTuple_SET_ITEM(tuple, 7, Py_None);
	Py_INCREF(Py_None);
	PyTuple_SET_ITEM(tuple, 8, Py_None);
    }
    else {
	ptr = SWIG_NewPointerObj((void*)srcName, SWIGTYPE_p_gss_name_t, 0);
	PyTuple_SET_ITEM(tuple, 2, ptr);
	ptr2 = SWIG_NewPointerObj((void*)targetName, SWIGTYPE_p_gss_name_t, 0);
	PyTuple_SET_ITEM(tuple, 3, ptr2);
	PyTuple_SET_ITEM(tuple, 4, Py_BuildValue("i", lifetime));
	ptr3 = SWIG_NewPointerObj((void*)mech, SWIGTYPE_p_gss_OID, 0);
	PyTuple_SET_ITEM(tuple, 5, ptr3);
	PyTuple_SET_ITEM(tuple, 6, Py_BuildValue("i", ctxFlags));
	PyTuple_SET_ITEM(tuple, 7, Py_BuildValue("i", local));
	PyTuple_SET_ITEM(tuple, 8, Py_BuildValue("i", open));
    }
    return tuple;
} 
	
gss_buffer_t
get_no_buffer_const() {
    return GSS_C_NO_BUFFER;
}
// JRB Argument minor_status is not used
PyObject*
display_status(OM_uint32 major_status, OM_uint32 minor_status)
{
    PyObject* list;
    OM_uint32 maj, min, status_type, msgCtx;
    gss_buffer_desc status;
    msgCtx = 0;
    list = PyList_New(0);
    status_type = GSS_C_GSS_CODE;
    do {
	maj = gss_display_status(&min, major_status, status_type,
				 GSS_C_NULL_OID, &msgCtx, &status);
	PyList_Append(list, Py_BuildValue("s#", 
                                         (char*)status.value,
					  status.length));
	gss_release_buffer(&min, &status);
    } while (msgCtx != 0 || 
	(status_type == GSS_C_GSS_CODE && (status_type = GSS_C_MECH_CODE )));
    return list;
}

#ifdef CAS
 
PyObject*
inquire_context_by_oid(gss_ctx_id_t* ctx, PyObject* pyOID) {
    PyObject *tuple, *list;
    OM_uint32 major, minor;
    gss_buffer_set_t set;
    gss_OID_desc oid;
    if (!(tuple=PyTuple_New(3))) {
        PyErr_SetString(PyExc_RuntimeError, 
                        "inquire_context_by_oid: PyTuple_New()");
        return NULL;
    }
    PyString_AsStringAndSize(pyOID,
                             (char **) &(oid.elements),
                             (int *) &(oid.length));
    major = gss_inquire_sec_context_by_oid(&minor, *ctx, &oid, &set);
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor));
    if (major != GSS_S_COMPLETE) {
	Py_INCREF(Py_None);
	PyTuple_SET_ITEM(tuple, 2, Py_None);
    }
    else {
        int i;
	list = PyList_New(0);
	for (i=0; i < set.count; i++) {
	    gss_buffer_desc buf = set->elements[i];
	    PyList_Append(list, Py_BuildValue("s#", 
					      (char*)buf.value,
					      buf.length));
	}
	PyTuple_SET_ITEM(tuple, 2, list);
    }
    return tuple;
} 

PyObject*
inquire_cred_by_oid(gss_cred_id_t* cred, PyObject* pyOID) {
    PyObject *tuple, *list;
    OM_uint32 major, minor;
    gss_buffer_set_t set;
    gss_OID_desc oid;
    if (!(tuple=PyTuple_New(3))) {
        PyErr_SetString(PyExc_RuntimeError, 
                        "inquire_cred_by_oid: PyTuple_New()");
        return NULL;
    }
    PyString_AsStringAndSize(pyOID,
                             (char **) &(oid.elements),
                             (int *) &(oid.length));
    major = gss_inquire_cred_by_oid(&minor, *cred, &oid, &set);
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor));
    if (major != GSS_S_COMPLETE) {
	Py_INCREF(Py_None);
	PyTuple_SET_ITEM(tuple, 2, Py_None);
    }
    else {
        int i;
	list = PyList_New(0);
	for (i=0; i < set.count; i++) {
	    gss_buffer_desc buf = set->elements[i];
	    PyList_Append(list, Py_BuildValue("s#", 
					      (char*)buf.value,
					      buf.length));
	}
	PyTuple_SET_ITEM(tuple, 2, list);
    }
    return tuple;
}

PyObject*
accept_delegation(gss_ctx_id_t* ctx, PyObject* oids, 
		  PyObject* buffers, OM_uint32 time_req,
		  globus_io_handle_t* io_handle) {
    PyObject *item, *ptr;
    globus_result_t ret;
    OM_uint32 major, minor;
    OM_uint32 time_rec_buf;
    gss_OID_set_desc oidSet;
    gss_buffer_set_desc bufferSet;
    gss_buffer_t buf;
    gss_buffer_desc output_token;
    gss_buffer_desc input_token;
    gss_cred_id_t *cred;
    gss_OID oid, mech;
    globus_size_t read, lenSize, written;
    int oidsLen, i, bufsLen;
    long sendLen;
    long readLen;
    char *globus_error_string = NULL;

    cred = (gss_cred_id_t*)globus_malloc(sizeof(gss_cred_id_t));
    *cred = GSS_C_NO_CREDENTIAL;

    lenSize = sizeof(input_token.length);
    oidsLen = PyList_Size(oids);
    mech = NULL;
    if (oidsLen > 0)
    {
        oidSet.count = oidsLen;
        bufsLen = PyList_Size(buffers);
        bufferSet.count = bufsLen;
        oidSet.elements = (gss_OID_desc *)globus_malloc(sizeof(gss_OID_desc) *
                                                        oidsLen);
        for (i=0; i < oidsLen; i++) {
            oid = &(oidSet.elements[i]);
            item = PyList_GetItem(oids, i);
            PyString_AsStringAndSize(item,
                                     (char **) &(oid->elements),
                                     (int *) &(oid->length));
        }

        bufferSet.elements = (gss_buffer_desc*)globus_malloc(
            sizeof(gss_buffer_desc) * bufsLen);
        for (i=0; i < bufsLen; i++) {
            buf = &(bufferSet.elements[i]);
            item = PyList_GetItem(buffers, i);
            PyString_AsStringAndSize(item,
                                     (char **) &(buf->value),
                                     (int *) &(buf->length));
        }
    }
    else
    {
        oidSet.count = 0;
        oidSet.elements = NULL;
        bufferSet.count = 0;
        bufferSet.elements = NULL;
    }
    do {
	Py_BEGIN_ALLOW_THREADS
#ifdef DEBUG
        printf("Reading input token length\n");
#endif
	ret = globus_io_read(io_handle, (globus_byte_t*)&readLen,
			     sizeof(readLen), sizeof(readLen), &read);
	Py_END_ALLOW_THREADS
	if (!(ret == GLOBUS_SUCCESS)) {
            globus_error_string = globus_object_printable_to_string(globus_error_get(ret));
	    PyErr_SetString(PyExc_RuntimeError, globus_error_string);
            globus_libc_free(globus_error_string);
	    if (oidsLen > 0) {
		globus_free(oidSet.elements);
	    }
	    return NULL;
	}
	Py_BEGIN_ALLOW_THREADS
	input_token.length = ntohl(readLen);
#ifdef DEBUG
        printf("Reading %d byte input token\n", input_token.length);
#endif
	input_token.value = globus_malloc(input_token.length);
	ret = globus_io_read(io_handle, (globus_byte_t*)input_token.value,
			     input_token.length, input_token.length, &read);
	Py_END_ALLOW_THREADS
	if (!(ret == GLOBUS_SUCCESS)) {
	    globus_error_string = globus_object_printable_to_string(globus_error_get(ret));
	    PyErr_SetString(PyExc_RuntimeError, globus_error_string);
	    globus_libc_free(globus_error_string);
	    if (oidsLen > 0) {
		globus_free(oidSet.elements);
	    }
	    return NULL;
	}
	Py_BEGIN_ALLOW_THREADS
#ifdef DEBUG
        printf("Calling gss_accept_delegation()\n");
#endif
	major = gss_accept_delegation(&minor, *ctx, 
				      &oidSet, &bufferSet, &input_token,
				      time_req, &time_rec_buf,
				      cred, &mech, &output_token);
	Py_END_ALLOW_THREADS
	if (major != GSS_S_COMPLETE && major != GSS_S_CONTINUE_NEEDED) {
            globus_gss_assist_display_status_str(&globus_error_string,
                                                 "Accepting delegation",
                                                 major, minor, 0);
            gss_release_buffer(&minor, &input_token);
	    PyErr_SetString(PyExc_RuntimeError, globus_error_string);
	    if (oidsLen > 0) {
		globus_free(oidSet.elements);
	    }
	    return NULL;
	}
        gss_release_buffer(&minor, &input_token);
#ifdef DEBUG
        printf("gss_accept_delegation() returned %d\n", major);
#endif
        if (output_token.length > 0) 
        {
	    Py_BEGIN_ALLOW_THREADS
#ifdef DEBUG
            printf("Writing %d byte token\n", output_token.length);
#endif
            sendLen = htonl(output_token.length);
            ret = globus_io_write(io_handle, (globus_byte_t*)&sendLen,
                                  sizeof(sendLen), &written);
	    Py_END_ALLOW_THREADS
            if (!(ret == GLOBUS_SUCCESS)) {
                globus_error_string = globus_object_printable_to_string(globus_error_get(ret));
                PyErr_SetString(PyExc_RuntimeError, globus_error_string);
                globus_libc_free(globus_error_string);
		if (oidsLen > 0) {
		    globus_free(oidSet.elements);
		}
                return NULL;
            }
	    Py_BEGIN_ALLOW_THREADS
            ret = globus_io_write(io_handle, output_token.value, 
                                  output_token.length, &written);
	    Py_END_ALLOW_THREADS
            if (!(ret == GLOBUS_SUCCESS)) {
                globus_error_string = globus_object_printable_to_string(globus_error_get(ret));
                PyErr_SetString(PyExc_RuntimeError, globus_error_string);
                globus_libc_free(globus_error_string);
		if (oidsLen > 0) {
		    globus_free(oidSet.elements);
		}
                return NULL;
            }
            gss_release_buffer(&minor, &output_token);
        }
    } while (major == GSS_S_CONTINUE_NEEDED);
#ifdef DEBUG
    printf("Returning good credential %x\n", cred);
#endif
    ptr = SWIG_NewPointerObj((void*)cred, SWIGTYPE_p_gss_cred_id_t, 0);
    if (oidsLen > 0) {
	globus_free(oidSet.elements);
    }
    return ptr;
}

PyObject*
init_delegation(gss_ctx_id_t* ctx, gss_cred_id_t* cred, 
		PyObject* oids, PyObject* buffers,
		OM_uint32 time_req,
		globus_io_handle_t* io_handle) {
    PyObject *item;
    globus_result_t ret;
    OM_uint32 major, minor;
    gss_OID_set_desc oidSet;
    gss_buffer_set_desc bufferSet;
    gss_buffer_t buf;
    gss_buffer_desc output_token;
    gss_buffer_t input_token = GSS_C_NO_BUFFER;
    gss_buffer_desc input_token_buf;
    gss_OID oid;
    gss_OID mech = NULL;
    globus_size_t lenSize;
    globus_size_t written;
    globus_size_t read;
    int oidsLen, i, bufsLen;
    long sendLen;
    long readLen;
    char *globus_error_string;
    
    oidsLen = PyList_Size(oids);

    if (oidsLen > 0) 
    {
        oidSet.count = oidsLen;
        bufsLen = PyList_Size(buffers);
        bufferSet.count = bufsLen;
        oidSet.elements = (gss_OID_desc*)globus_malloc(sizeof(gss_OID_desc) *
                                                       oidsLen);
        for (i=0; i < oidsLen; i++) {
            oid = &(oidSet.elements[i]);
            item = PyList_GetItem(oids, i);
            PyString_AsStringAndSize(item,
                                     (char **) &(oid->elements),
                                     (int *) &(oid->length));
        }

        bufferSet.elements = (gss_buffer_desc*)globus_malloc(
            sizeof(gss_buffer_desc) * bufsLen);
        for (i=0; i < bufsLen; i++) {
            buf = &(bufferSet.elements[i]);
            item = PyList_GetItem(buffers, i);
            PyString_AsStringAndSize(item,
                                     (char **) &(buf->value),
                                     (int *) &(buf->length));
        }
    }
    else
    {
        oidSet.count = 0;
        oidSet.elements = NULL;
        bufferSet.count = 0;
        bufferSet.elements = NULL;
    }
    do {
	Py_BEGIN_ALLOW_THREADS
#ifdef DEBUG
        printf("DEBUG: Calling gss_init_delegation\n");
#endif
	major = gss_init_delegation(&minor, *ctx, *cred, mech,
				    &oidSet, &bufferSet, 
				    input_token, time_req, &output_token);
#ifdef DEBUG
        printf("DEBUG: gss_init_delegation returned %d\n", major);
#endif
	Py_END_ALLOW_THREADS
	if (major != GSS_S_COMPLETE && major != GSS_S_CONTINUE_NEEDED) {
            globus_gss_assist_display_status_str(&globus_error_string,
                                                 "Initing delegation",
                                                 major, minor, 0);
            gss_release_buffer(&minor, input_token);
	    PyErr_SetString(PyExc_RuntimeError, globus_error_string);
	    if (oidsLen > 0) {
		globus_free(oidSet.elements);
	    }
	    return NULL;
	}

        /* This call is ok if input_token == GSS_C_NO_BUFFER */
        gss_release_buffer(&minor, input_token);

        if (output_token.length > 0)
        {
	    Py_BEGIN_ALLOW_THREADS
#ifdef DEBUG
            printf("Writing %d byte token\n", output_token.length);
#endif
            sendLen = htonl(output_token.length);
            ret = globus_io_write(io_handle, (globus_byte_t*)&sendLen,
                                  sizeof(sendLen), &written);
	    Py_END_ALLOW_THREADS
            if (!(ret == GLOBUS_SUCCESS)) {
                globus_error_string = globus_object_printable_to_string(globus_error_get(ret));
                PyErr_SetString(PyExc_RuntimeError, globus_error_string);
                globus_libc_free(globus_error_string);
		if (oidsLen > 0) {
		    globus_free(oidSet.elements);
		}
                return NULL;
            }
	    Py_BEGIN_ALLOW_THREADS
            ret = globus_io_write(io_handle, output_token.value, 
                                  output_token.length, &written);
	    Py_END_ALLOW_THREADS
            if (!(ret == GLOBUS_SUCCESS)) {
                globus_error_string = globus_object_printable_to_string(globus_error_get(ret));
                PyErr_SetString(PyExc_RuntimeError, globus_error_string);
                globus_libc_free(globus_error_string);
		if (oidsLen > 0) {
		    globus_free(oidSet.elements);
		}
                return NULL;
            }
            gss_release_buffer(&minor, &output_token);
        }
        
	if (major == GSS_S_CONTINUE_NEEDED) {
	    Py_BEGIN_ALLOW_THREADS
	    ret = globus_io_read(io_handle, (globus_byte_t*)&readLen,
				 sizeof(readLen), sizeof(readLen), &read);
	    Py_END_ALLOW_THREADS
	    if (!(ret == GLOBUS_SUCCESS)) {
                globus_error_string = globus_object_printable_to_string(globus_error_get(ret));
                PyErr_SetString(PyExc_RuntimeError, globus_error_string);
                globus_libc_free(globus_error_string);
		if (oidsLen > 0) {
		    globus_free(oidSet.elements);
		}
                return NULL;
	    }
	    Py_BEGIN_ALLOW_THREADS
	    input_token_buf.length = ntohl(readLen);
#ifdef DEBUG
            printf("Reading %d byte input token\n", input_token_buf.length);
#endif
	    input_token_buf.value = globus_malloc(input_token_buf.length);  
	    ret = globus_io_read(io_handle,
                                 (globus_byte_t*)input_token_buf.value,
				 input_token_buf.length,
                                 input_token_buf.length,
                                 &read);
	    Py_END_ALLOW_THREADS
	    if (!(ret == GLOBUS_SUCCESS)) {
                globus_error_string = globus_object_printable_to_string(globus_error_get(ret));
                PyErr_SetString(PyExc_RuntimeError, globus_error_string);
                globus_libc_free(globus_error_string);
		if (oidsLen > 0) {
		    globus_free(oidSet.elements);
		}
                return NULL;
	    }
            input_token = &input_token_buf;
	}
    } while (major == GSS_S_CONTINUE_NEEDED);
    Py_INCREF(Py_None);
    return Py_None;
}

#endif /* CAS */

PyObject*
export_cred(gss_cred_id_t *cred,
            int option_req)
{
    OM_uint32 major;
    OM_uint32 minor;
    gss_OID p_desired_mech = NULL;
    gss_buffer_desc export_buffer;
    char *globus_error_string = NULL;
    PyObject* tuple;

#ifdef DEBUG
    printf("Exporting cred %x (%x)\n", cred, *cred);
#endif

    if (!(tuple=PyTuple_New(3))) {
	PyErr_SetString(PyExc_RuntimeError, 
			"export_cred: PyTuple_New()");
	return NULL;
    }
    major = gss_export_cred(&minor,
                            *cred,
                            p_desired_mech,
                            option_req,
                            &export_buffer);

    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor));

#ifdef DEBUG
    printf("Export done. major = %d\n", major);
#endif
    if (major == GLOBUS_SUCCESS)
    {
        PyTuple_SET_ITEM(tuple, 2,
                         Py_BuildValue("s#", (char *) export_buffer.value,
                                       export_buffer.length));
        gss_release_buffer(&minor, &export_buffer);
    }
    else
    {
	Py_INCREF(Py_None);
	PyTuple_SET_ITEM(tuple, 2, Py_None);
    }

    return tuple;
}
                
PyObject*
import_cred(int option_req, PyObject* credHandle, int time)
{    
    OM_uint32 major;
    OM_uint32 minor;
    OM_uint32 timeRec;
    gss_OID p_desired_mech = NULL;
    gss_buffer_t buf;
    char *globus_error_string = NULL;
    PyObject* tuple;
    gss_cred_id_t* cred;
    PyObject* ptr;
    if (!(tuple=PyTuple_New(4))) {
	PyErr_SetString(PyExc_RuntimeError, 
			"import_cred: PyTuple_New()");
	return NULL;
    }
    cred = (gss_cred_id_t*)globus_malloc(sizeof(gss_cred_id_t));
    buf = (gss_buffer_desc*)globus_malloc(sizeof(gss_buffer_desc));
    buf->length = PyString_Size(credHandle);
    buf->value = PyString_AsString(credHandle);
    printf("about to call c import cred\n");
    major = gss_import_cred(&minor,
                            cred,
                            p_desired_mech,
                            option_req,
                            buf,
                            (OM_uint32)time,
                            &timeRec);
    printf("completed c imort cred call\n");
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor));
    if (major == GLOBUS_SUCCESS)
    {
        ptr = SWIG_NewPointerObj((void*)cred, SWIGTYPE_p_gss_cred_id_t, 0);
        PyTuple_SET_ITEM(tuple, 2, ptr);
        PyTuple_SET_ITEM(tuple, 3, Py_BuildValue("i", timeRec));
    }
    else
    {
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 2, Py_None);
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 3, Py_None);
    }
    globus_free(buf);
    return tuple;
}

PyObject*
inquire_cred(gss_cred_id_t* cred) {
    PyObject *tuple, *list, *name_p, life_p;
    OM_uint32 major;
    OM_uint32 minor;
    gss_name_t name;
    OM_uint32 life;
    gss_cred_usage_t cred_usage;
    //gss_buffer_set_t set;
    //gss_OID_desc *oid;
    gss_OID_set_desc *oidSet;
    if (!(tuple=PyTuple_New(4))) {
        PyErr_SetString(PyExc_RuntimeError, 
                        "inquire_cred: PyTuple_New()");
        return NULL;
    }
    name = (gss_name_t*)globus_malloc(sizeof(gss_name_t));
    major = gss_inquire_cred(&minor, *cred, &name, &life, 
                              &cred_usage, &oidSet);

    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor));
    if (major != GSS_S_COMPLETE) {
	Py_INCREF(Py_None);
	PyTuple_SET_ITEM(tuple, 2, Py_None);
	Py_INCREF(Py_None);
	PyTuple_SET_ITEM(tuple, 3, Py_None);
    }
    else {
        //int i;
	PyTuple_SET_ITEM(tuple, 2, Py_BuildValue("i", life));
	name_p = SWIG_NewPointerObj((void*)name, SWIGTYPE_p_gss_name_t, 0);
	Py_INCREF(name_p);
	PyTuple_SET_ITEM(tuple, 3, name_p);
/*
	list = PyList_New(0);
	for (i=0; i < oidSet.count; i++) {
	    gss_OID_desc buf = oidSet.elements[i];
	    PyList_Append(list, Py_BuildValue("s#", 
					      (char*)buf.elements,
					      buf.length));
	}
	PyTuple_SET_ITEM(tuple, 3, list);
*/
    }
    return tuple;
}

PyObject*
get_name(gss_cred_id_t* cred) {
    PyObject         *tuple, *list, *name_p, life_p;
    OM_uint32         major;
    OM_uint32         minor;
    gss_name_t        name;
    OM_uint32         life;
    gss_cred_usage_t  cred_usage;
    gss_OID_set_desc *oidSet;
    gss_buffer_desc   buffer_desc = GSS_C_EMPTY_BUFFER;
    gss_buffer_t      buffer = &buffer_desc;


    if (!(tuple=PyTuple_New(3))) {
        PyErr_SetString(PyExc_RuntimeError, 
                        "inquire_cred: PyTuple_New()");
        return NULL;
    }
    name = (gss_name_t*)globus_malloc(sizeof(gss_name_t));
    major = gss_inquire_cred(&minor, *cred, &name, &life, 
                              &cred_usage, &oidSet);

    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major));
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor));
    if (major != GSS_S_COMPLETE) {
	Py_INCREF(Py_None);
	PyTuple_SET_ITEM(tuple, 2, Py_None);
    }
    else {
	major = 0;
	minor = 0;
	major = gss_display_name(&minor, name, buffer, NULL);
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", major));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", minor));
	if (major != GSS_S_COMPLETE) {
	    PyTuple_SET_ITEM(tuple, 2, Py_BuildValue("i", minor));
	}
	else {
	    PyTuple_SET_ITEM(tuple, 2, Py_BuildValue("s#", (char*)buffer_desc.value, 
						     (int)buffer_desc.length));
	    gss_release_buffer(&minor, buffer);
	}
    }
    return tuple;
}
%}
