/*############################################################################
# Keith R. Jackson, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################*/
%module replicaCatalogc
%{
#include "globus_common.h"
#include "globus_replica_catalog.h"
#include "globus_replica_catalog_sets.h"
%}

%include ./src/pyGlobus-typemaps.i

typedef int globus_bool_t;
typedef unsigned char globus_byte_t;

typedef enum
{
    GLOBUS_REPLICA_CATALOG_AUTHENTICATION_MODE_SECURE,
    GLOBUS_REPLICA_CATALOG_AUTHENTICATION_MODE_CLEARTEXT
} globus_replica_catalog_authentication_mode_t;

%constant COLLECTION_OBJECTCLASS = "GlobusReplicaLogicalCollection";
%constant LOCATION_OBJECTCLASS = "GlobusReplicaInfo";
%constant LOGICALFILE_OBJECTCLASS = "GlobusReplicaLogicalFile";
%constant FILEPARENT_OBJECTCLASS = "GlobusReplicaFileInfo";
%constant FILEPARENT_RDN = "fin=LFROOT";
%constant FILENAME_ATTRIBUTE_TYPE = "filename";

%{

%}

%inline %{
globus_module_descriptor_t*
get_module(void) {
    return GLOBUS_REPLICA_CATALOG_MODULE;
}

PyObject*
handleattr_init() {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_replica_catalog_collection_handleattr_t* attr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_init: PyTuple_New()");
        return NULL;
    }
    attr = (globus_replica_catalog_collection_handleattr_t*)
	globus_malloc(sizeof(globus_replica_catalog_collection_handleattr_t));
    ret = globus_replica_catalog_collection_handleattr_init(attr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(attr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)attr, 
		   SWIGTYPE_p_globus_replica_catalog_collection_handleattr_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
handleattr_copy(globus_replica_catalog_collection_handleattr_t* srcAttr) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_replica_catalog_collection_handleattr_t* attr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_copy: PyTuple_New()");
        return NULL;
    }
    attr = (globus_replica_catalog_collection_handleattr_t*)
	globus_malloc(sizeof(globus_replica_catalog_collection_handleattr_t));
    ret = globus_replica_catalog_collection_handleattr_copy(attr, srcAttr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(attr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)attr, 
		     SWIGTYPE_p_globus_replica_catalog_collection_handleattr_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
handleattr_destroy(globus_replica_catalog_collection_handleattr_t* attr) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_destroy: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_collection_handleattr_destroy(attr);
    globus_free(attr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
set_authentication_mode(globus_replica_catalog_collection_handleattr_t* attr,
			globus_replica_catalog_authentication_mode_t mode,
			PyObject* pyUsr, PyObject* pyPasswd) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    char* usr;
    char* passwd;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "set_authentication_mode: PyTuple_New()");
        return NULL;
    }
    if (mode == GLOBUS_REPLICA_CATALOG_AUTHENTICATION_MODE_SECURE)
        ret = globus_replica_catalog_collection_handleattr_set_authentication_mode(attr, mode);
    else {
        if (pyUsr == Py_None) {
            usr = GLOBUS_NULL;
            passwd = GLOBUS_NULL;
        }
        else {
            usr = PyString_AsString(pyUsr);
            passwd = PyString_AsString(pyPasswd);
        }
        ret = globus_replica_catalog_collection_handleattr_set_authentication_mode(attr, 
                                                                                   mode, usr,
                                                                                   passwd);
    }
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;	
}

PyObject*
get_authentication_mode(globus_replica_catalog_collection_handleattr_t* attr) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    char** usr;
    char** passwd;
    globus_replica_catalog_authentication_mode_t mode;
    if (!(tuple=PyTuple_New(4)))
    {
        PyErr_SetString(PyExc_RuntimeError, "set_authentication_mode: PyTuple_New()");
        return NULL;
    }
    *usr = NULL;
    *passwd = NULL;
    ret = globus_replica_catalog_collection_handleattr_get_authentication_mode(attr, 
                                                                               &mode, usr, passwd);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);								
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 2, Py_None);
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 3, Py_None);
    }
    else if (*usr == NULL) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", (int)mode));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 2, Py_None);
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 3, Py_None);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", (int)mode));
        PyTuple_SET_ITEM(tuple, 2, Py_BuildValue("s", *usr));
        PyTuple_SET_ITEM(tuple, 3, Py_BuildValue("s", *passwd));
    }
    return tuple;
}

PyObject*
collection_open(globus_replica_catalog_collection_handleattr_t* attr, char* url) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_replica_catalog_collection_handle_t* handle;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "collection_open: PyTuple_New()");
        return NULL;
    }
    handle = (globus_replica_catalog_collection_handle_t*)globus_malloc(
                                                   sizeof(globus_replica_catalog_collection_handle_t));
    ret = globus_replica_catalog_collection_open(handle, attr, url);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(handle);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)handle, 
			SWIGTYPE_p_globus_replica_catalog_collection_handle_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
list_collections(globus_replica_catalog_collection_handle_t* handle,
                 PyObject* attrNames) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    char** attrs;
    int len, i;
    globus_replica_catalog_entry_set_t* set;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "list_collections: PyTuple_New()");
        return NULL;
    }
    set = (globus_replica_catalog_entry_set_t*)globus_malloc(
                                                           sizeof(globus_replica_catalog_entry_set_t));
    ret = globus_replica_catalog_entry_set_init(set);
     if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(set);
    }
    else {
        if (attrNames != Py_None) {
            len = PyList_Size(attrNames);
            attrs = (char**)globus_malloc(len * sizeof(char*) + 1);
            for(i=0; i <len; i++) {
                attrs[i] = PyString_AsString(PyList_GetItem(attrNames, i));
            }
            attrs[len] = "\0";
            ret = globus_replica_catalog_list_collections(handle, attrs, set);
        }
        else {
            ret = globus_replica_catalog_list_collections(handle, GLOBUS_NULL, set);
        }
        if (ret != GLOBUS_SUCCESS) {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
            tmpstr = globus_object_printable_to_string(globus_error_get(ret));
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
            globus_libc_free(set);
        }
        else {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            ptr = SWIG_NewPointerObj((void*)set, 
                                     SWIGTYPE_p_globus_replica_catalog_entry_set_t, 0);
            PyTuple_SET_ITEM(tuple, 1, ptr);
        }
    }       
     if (attrNames != Py_None)
         globus_libc_free(attrs);
     return tuple;
}

PyObject*
list_rollback_entries(char* url, PyObject* pyAttr, PyObject* attrNames) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    char** attrs;
    int len, i;
    globus_replica_catalog_entry_set_t* set;
    globus_replica_catalog_collection_handleattr_t* attr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "list_collections: PyTuple_New()");
        return NULL;
    }
    set = (globus_replica_catalog_entry_set_t*)globus_malloc(
                                                           sizeof(globus_replica_catalog_entry_set_t));
    ret = globus_replica_catalog_entry_set_init(set);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(set);
    }
    else {
        if (pyAttr == Py_None) 
            attr = GLOBUS_NULL;
        else
            SWIG_ConvertPtr(pyAttr, (void**)&attr, 
                            SWIGTYPE_p_globus_replica_catalog_collection_handleattr_t, 0);
        if (attrNames != Py_None) {
            len = PyList_Size(attrNames);
            attrs = (char**)globus_malloc(len * sizeof(char*) + 1);
            for(i=0; i <len; i++) {
                attrs[i] = PyString_AsString(PyList_GetItem(attrNames, i));
            }
            attrs[len] = "\0";
            ret = globus_replica_catalog_list_rollback_entries(url, attr, attrs, set);
        }
        else
            ret = globus_replica_catalog_list_rollback_entries(url, attr,
                                                               GLOBUS_NULL, set);
        if (ret != GLOBUS_SUCCESS) {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
            tmpstr = globus_object_printable_to_string(globus_error_get(ret));
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
            globus_libc_free(set);
        }
        else {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            ptr = SWIG_NewPointerObj((void*)set, 
                                     SWIGTYPE_p_globus_replica_catalog_entry_set_t, 0);
            PyTuple_SET_ITEM(tuple, 1, ptr);
        }
    }       
    if (attrNames != Py_None)
        globus_libc_free(attrs);
    return tuple;
}

PyObject*
collection_set_dn(globus_replica_catalog_collection_handle_t* handle, char* dn) {
    PyObject* tuple;
    globus_result_t ret;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "collection_set_dn: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_collection_set_dn(handle, dn);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
collection_create(globus_replica_catalog_collection_handleattr_t* attr, 
                  char* url, PyObject* fileList, char* objectclass) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    char** files;
    int len, i;
    globus_replica_catalog_collection_handle_t* handle;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "collection_create: PyTuple_New()");
        return NULL;
    }
    handle = (globus_replica_catalog_collection_handle_t*)globus_malloc(
                                                   sizeof(globus_replica_catalog_collection_handle_t));
    if (fileList != Py_None) {
        len = PyList_Size(fileList);
        files = (char**)globus_malloc(len * sizeof(char*) + 1);
        for(i=0; i <len; i++) {
            files[i] = PyString_AsString(PyList_GetItem(fileList, i));
        }
        files[len] = "\0";
        ret = globus_replica_catalog_collection_create(handle, attr, url, files, objectclass);
    }
    else {
        ret = globus_replica_catalog_collection_create(handle, attr, 
                                                       url, GLOBUS_NULL, objectclass);
    }
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(handle);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)handle, 
			SWIGTYPE_p_globus_replica_catalog_collection_handle_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    if (fileList != Py_None)
        globus_libc_free(files);
    return tuple;
}

PyObject*
collection_close(globus_replica_catalog_collection_handle_t* handle) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "collection_close: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_collection_close(handle);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        globus_libc_free(handle);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
collection_delete(globus_replica_catalog_collection_handle_t* handle) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "collection_delete: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_collection_delete(handle);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        globus_libc_free(handle);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
collection_add_filenames(globus_replica_catalog_collection_handle_t* handle,
                         PyObject* fileList, globus_bool_t add) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    char** files;
    int len, i;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "collection_add_filenames: PyTuple_New()");
        return NULL;
    }
    len = PyList_Size(fileList);
    files = (char**)globus_malloc(len * sizeof(char*) + 1);
    for(i=0; i <len; i++) {
        files[i] = PyString_AsString(PyList_GetItem(fileList, i));
    }
    files[len] = "\0";
    ret = globus_replica_catalog_collection_add_filenames(handle, files, add);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    globus_libc_free(files);
    return tuple;
}

PyObject*
collection_add_attribute(globus_replica_catalog_collection_handle_t* handle,
                         char* name, char* value) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "collection_add_attribute: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_collection_add_attribute(handle, name, value);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
collection_delete_filenames(globus_replica_catalog_collection_handle_t* handle,
                            PyObject* fileList) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    char** files;
    int len, i;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "collection_delete_filenames: PyTuple_New()");
        return NULL;
    }
    len = PyList_Size(fileList);
    files = (char**)globus_malloc(len * sizeof(char*) + 1);
    for(i=0; i <len; i++) {
        files[i] = PyString_AsString(PyList_GetItem(fileList, i));
    }
    files[len] = "\0";
    ret = globus_replica_catalog_collection_delete_filenames(handle, files);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    globus_libc_free(files);
    return tuple;
}

PyObject*
collection_delete_attribute(globus_replica_catalog_collection_handle_t* handle,
                            char* name, PyObject* value) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    char* tmpPtr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "collection_delete_attributes: PyTuple_New()");
        return NULL;
    }
    if (value != Py_None) {
        tmpPtr = PyString_AsString(value);
        ret = globus_replica_catalog_collection_delete_attribute(handle, name, tmpPtr);
    }
    else 
        ret = globus_replica_catalog_collection_delete_attribute(handle,
                                                                 name, GLOBUS_NULL);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
collection_list_filenames(globus_replica_catalog_collection_handle_t* handle) {
    globus_result_t ret;
    PyObject *tuple, *fileList;
    char* tmpstr;
    char** files;
    int i;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "collection_list_filenames: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_collection_list_filenames(handle, &files);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        fileList = PyList_New(0);
        for (i=0; files[i]; i++) {
            PyList_Append(fileList, Py_BuildValue("s", files[i]));
        }
        PyTuple_SET_ITEM(tuple, 1, fileList);
    }
    return tuple;
}

PyObject*
collection_list_attributes(globus_replica_catalog_collection_handle_t* handle,
                           PyObject* attrNames) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    char** attrs;
    int len, i;
    globus_replica_catalog_attribute_set_t* set;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "list_attributes: PyTuple_New()");
        return NULL;
    }
    set = (globus_replica_catalog_attribute_set_t*)globus_malloc(
                                                           sizeof(globus_replica_catalog_attribute_set_t));
    ret = globus_replica_catalog_attribute_set_init(set);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(set);
    }
    else {
        if (attrNames != Py_None) {
            len = PyList_Size(attrNames);
            attrs = (char**)globus_malloc(len * sizeof(char*) + 1);
            for(i=0; i <len; i++) {
                attrs[i] = PyString_AsString(PyList_GetItem(attrNames, i));
            }
            attrs[len] = "\0";
            ret = globus_replica_catalog_collection_list_attributes(handle, attrs, set);
        }
        else
            ret = globus_replica_catalog_collection_list_attributes(handle,
                                                               GLOBUS_NULL, set);
        if (ret != GLOBUS_SUCCESS) {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
            tmpstr = globus_object_printable_to_string(globus_error_get(ret));
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
            globus_libc_free(set);
        }
        else {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            ptr = SWIG_NewPointerObj((void*)set, 
                                     SWIGTYPE_p_globus_replica_catalog_attribute_set_t, 0);
            PyTuple_SET_ITEM(tuple, 1, ptr);
        }
    }       
    if (attrNames != Py_None)
        globus_libc_free(attrs);
    return tuple;
}

PyObject*
collection_find_locations(globus_replica_catalog_collection_handle_t* handle,
                          PyObject* fileList, globus_bool_t all, PyObject* attrNames) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    char** attrs;
    char** files;
    int len, i;
    globus_replica_catalog_entry_set_t* set;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "find_locations: PyTuple_New()");
        return NULL;
    }
    set = (globus_replica_catalog_entry_set_t*)globus_malloc(
                                                           sizeof(globus_replica_catalog_entry_set_t));
    ret = globus_replica_catalog_entry_set_init(set);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(set);
    }
    else {
        len = PyList_Size(fileList);
        files = (char**)globus_malloc(len * sizeof(char*) + 1);
        for (i=0; i<len; i++) {
            files[i] = PyString_AsString(PyList_GetItem(fileList, i));
        }
        files[len] = "\0";
        if (attrNames != Py_None) {
            len = PyList_Size(attrNames);
            attrs = (char**)globus_malloc(len * sizeof(char*) + 1);
            for(i=0; i <len; i++) {
                attrs[i] = PyString_AsString(PyList_GetItem(attrNames, i));
            }
            attrs[len] = "\0";
            ret = globus_replica_catalog_collection_find_locations(handle, files, all, attrs, set);
        }
        else
            ret = globus_replica_catalog_collection_find_locations(handle, files, all,
                                                               GLOBUS_NULL, set);
        if (ret != GLOBUS_SUCCESS) {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
            tmpstr = globus_object_printable_to_string(globus_error_get(ret));
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
            globus_libc_free(set);
        }
        else {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            ptr = SWIG_NewPointerObj((void*)set, 
                                     SWIGTYPE_p_globus_replica_catalog_entry_set_t, 0);
            PyTuple_SET_ITEM(tuple, 1, ptr);
        }
        globus_libc_free(files);
    }       
    if (attrNames != Py_None)
        globus_libc_free(attrs);
    return tuple;
}

PyObject*
collection_list_locations(globus_replica_catalog_collection_handle_t* handle,
                                                                                  PyObject* attrNames) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    char** attrs;
    int len, i;
    globus_replica_catalog_entry_set_t* set;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "list_locations: PyTuple_New()");
        return NULL;
    }
    set = (globus_replica_catalog_entry_set_t*)globus_malloc(
                                                           sizeof(globus_replica_catalog_entry_set_t));
    ret = globus_replica_catalog_entry_set_init(set);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(set);
    }
    else {
        if (attrNames != Py_None) {
            len = PyList_Size(attrNames);
            attrs = (char**)globus_malloc(len * sizeof(char*) + 1);
            for(i=0; i <len; i++) {
                attrs[i] = PyString_AsString(PyList_GetItem(attrNames, i));
            }
            attrs[len] = "\0";
            ret = globus_replica_catalog_collection_list_locations(handle, attrs, set);
        }
        else
            ret = globus_replica_catalog_collection_list_locations(handle,
                                                               GLOBUS_NULL, set);
        if (ret != GLOBUS_SUCCESS) {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
            tmpstr = globus_object_printable_to_string(globus_error_get(ret));
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
            globus_libc_free(set);
        }
        else {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            ptr = SWIG_NewPointerObj((void*)set, 
                                     SWIGTYPE_p_globus_replica_catalog_entry_set_t, 0);
            PyTuple_SET_ITEM(tuple, 1, ptr);
        }
    }       
    if (attrNames != Py_None)
        globus_libc_free(attrs);
    return tuple;
}

PyObject*
collection_list_logicalfiles(globus_replica_catalog_collection_handle_t* handle,
                                                                                  PyObject* attrNames) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    char** attrs;
    int len, i;
    globus_replica_catalog_entry_set_t* set;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "list_logicalfiles: PyTuple_New()");
        return NULL;
    }
    set = (globus_replica_catalog_entry_set_t*)globus_malloc(
                                                           sizeof(globus_replica_catalog_entry_set_t));
    ret = globus_replica_catalog_entry_set_init(set);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(set);
    }
    else {
        if (attrNames != Py_None) {
            len = PyList_Size(attrNames);
            attrs = (char**)globus_malloc(len * sizeof(char*) + 1);
            for(i=0; i <len; i++) {
                attrs[i] = PyString_AsString(PyList_GetItem(attrNames, i));
            }
            attrs[len] = "\0";
            ret = globus_replica_catalog_collection_list_logicalfiles(handle, attrs, set);
        }
        else
            ret = globus_replica_catalog_collection_list_logicalfiles(handle,
                                                               GLOBUS_NULL, set);
        if (ret != GLOBUS_SUCCESS) {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
            tmpstr = globus_object_printable_to_string(globus_error_get(ret));
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
            globus_libc_free(set);
        }
        else {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            ptr = SWIG_NewPointerObj((void*)set, 
                                     SWIGTYPE_p_globus_replica_catalog_entry_set_t, 0);
            PyTuple_SET_ITEM(tuple, 1, ptr);
        }
    }       
    if (attrNames != Py_None)
        globus_libc_free(attrs);
    return tuple;
}

PyObject*
location_create(globus_replica_catalog_collection_handle_t* handle,
                char* location, char* url, PyObject* fileNames, char* objectclass) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    char** files;
    int len, i;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "location_create: PyTuple_New()");
        return NULL;
    }
    if (fileNames != Py_None) {
        len = PyList_Size(fileNames);
        files = (char**)globus_malloc(len * sizeof(char*) + 1);
        for(i=0; i <len; i++) {
            files[i] = PyString_AsString(PyList_GetItem(fileNames, i));
        }
        files[len] = "\0";
        ret = globus_replica_catalog_location_create(handle, location,
                                                     url, files, objectclass);
    }
    else
        ret = globus_replica_catalog_location_create(handle,
                                                     location, url, GLOBUS_NULL, objectclass);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    if (fileNames != Py_None)
        globus_libc_free(files);
    return tuple;
}

PyObject*
location_add_filenames(globus_replica_catalog_collection_handle_t* handle,
                char* location, PyObject* fileNames, globus_bool_t add) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    char** files;
    int len, i;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "location_add_filenames: PyTuple_New()");
        return NULL;
    }
    len = PyList_Size(fileNames);
    files = (char**)globus_malloc(len * sizeof(char*) + 1);
    for(i=0; i <len; i++) {
        files[i] = PyString_AsString(PyList_GetItem(fileNames, i));
    }
    files[len] = "\0";
    ret = globus_replica_catalog_location_add_filenames(handle, location, files, add);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    if (fileNames != Py_None)
        globus_libc_free(files);
    return tuple;
}

PyObject*
location_add_attribute(globus_replica_catalog_collection_handle_t* handle,
                       char* location, char* name, char* value) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "location_add_attribute: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_location_add_attribute(handle, location, name, value);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
location_delete(globus_replica_catalog_collection_handle_t* handle,
                char* location) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "location_delete: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_location_delete(handle, location);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
location_delete_filenames(globus_replica_catalog_collection_handle_t* handle,
                          char* location, PyObject* fileNames) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    char** files;
    int len, i;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "location_delete_filenames: PyTuple_New()");
        return NULL;
    }
    if (fileNames != Py_None) {
        len = PyList_Size(fileNames);
        files = (char**)globus_malloc(len * sizeof(char*) + 1);
        for(i=0; i <len; i++) {
            files[i] = PyString_AsString(PyList_GetItem(fileNames, i));
        }
        files[len] = "\0";
        ret = globus_replica_catalog_location_delete_filenames(handle, location, files);
    }
    else
        ret = globus_replica_catalog_location_delete_filenames(handle,
                                                               location, GLOBUS_NULL);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    if (fileNames != Py_None)
        globus_libc_free(files);
    return tuple;
}

PyObject*
location_delete_attribute(globus_replica_catalog_collection_handle_t* handle,
                       char* location, char* name, char* value) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "location_delete_attribute: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_location_delete_attribute(handle, location, name, value);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
location_list_filenames(globus_replica_catalog_collection_handle_t* handle,
                        char* location) {
    globus_result_t ret;
    PyObject *tuple, *fileList;
    char* tmpstr;
    char** files;
    int i;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "location_list_filenames: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_location_list_filenames(handle, location, &files);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        if (&files == GLOBUS_NULL) {
            Py_INCREF(Py_None);
            fileList = Py_None;
        }
        else {
            fileList = PyList_New(0);
            for (i=0; files[i]; i++) {
                PyList_Append(fileList, Py_BuildValue("s", files[i]));
            }
            globus_replica_catalog_attribute_set_free_values(files);
        }
        PyTuple_SET_ITEM(tuple, 1, fileList);
    }
    return tuple;
}

PyObject*
location_list_attributes(globus_replica_catalog_collection_handle_t* handle,
                         char* location, PyObject* attrNames) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    char** attrs;
    int len, i;
    globus_replica_catalog_attribute_set_t* set;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "location_list_attributes: PyTuple_New()");
        return NULL;
    }
    set = (globus_replica_catalog_attribute_set_t*)globus_malloc(
                                                           sizeof(globus_replica_catalog_attribute_set_t));
    ret = globus_replica_catalog_attribute_set_init(set);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(set);
    }
    else {
        if (attrNames != Py_None) {
            len = PyList_Size(attrNames);
            attrs = (char**)globus_malloc(len * sizeof(char*) + 1);
            for(i=0; i <len; i++) {
                attrs[i] = PyString_AsString(PyList_GetItem(attrNames, i));
            }
            attrs[len] = "\0";
            ret = globus_replica_catalog_location_list_attributes(handle, location, attrs, set);
        }
        else
            ret = globus_replica_catalog_location_list_attributes(handle,
                                                                  location, GLOBUS_NULL, set);
        if (ret != GLOBUS_SUCCESS) {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
            tmpstr = globus_object_printable_to_string(globus_error_get(ret));
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
            globus_libc_free(set);
        }
        else {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            ptr = SWIG_NewPointerObj((void*)set, 
                                     SWIGTYPE_p_globus_replica_catalog_attribute_set_t, 0);
            PyTuple_SET_ITEM(tuple, 1, ptr);
        }
    }       
    if (attrNames != Py_None)
        globus_libc_free(attrs);
    return tuple;
}

PyObject*
logicalfile_create(globus_replica_catalog_collection_handle_t* handle,
                char* logicalfile, int size, char* objectclass) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "logicalfile_create: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_logicalfile_create(handle,
                                                     logicalfile, size, objectclass);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
logicalfile_add_attribute(globus_replica_catalog_collection_handle_t* handle,
                       char* logicalfile, char* name, char* value) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "logicalfile_add_attribute: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_logicalfile_add_attribute(handle, logicalfile, name, value);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
logicalfile_delete(globus_replica_catalog_collection_handle_t* handle,
                char* logicalfile) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "logicalfile_delete: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_logicalfile_delete(handle, logicalfile);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
logicalfile_delete_attribute(globus_replica_catalog_collection_handle_t* handle,
                       char* logicalfile, char* name, char* value) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "logicalfile_delete_attribute: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_logicalfile_delete_attribute(handle, logicalfile, name, value);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
logicalfile_list_attributes(globus_replica_catalog_collection_handle_t* handle,
                         char* logicalfile, PyObject* attrNames) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    char** attrs;
    int len, i;
    globus_replica_catalog_attribute_set_t* set;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "logicalfile_list_attributes: PyTuple_New()");
        return NULL;
    }
    set = (globus_replica_catalog_attribute_set_t*)globus_malloc(
                                                           sizeof(globus_replica_catalog_attribute_set_t));
    ret = globus_replica_catalog_attribute_set_init(set);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(set);
    }
    else {
        if (attrNames != Py_None) {
            len = PyList_Size(attrNames);
            attrs = (char**)globus_malloc(len * sizeof(char*) + 1);
            for(i=0; i <len; i++) {
                attrs[i] = PyString_AsString(PyList_GetItem(attrNames, i));
            }
            attrs[len] = "\0";
            ret = globus_replica_catalog_logicalfile_list_attributes(handle, logicalfile, attrs, set);
        }
        else
            ret = globus_replica_catalog_logicalfile_list_attributes(handle,
                                                                  logicalfile, GLOBUS_NULL, set);
        if (ret != GLOBUS_SUCCESS) {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
            tmpstr = globus_object_printable_to_string(globus_error_get(ret));
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
            globus_libc_free(set);
        }
        else {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            ptr = SWIG_NewPointerObj((void*)set, 
                                     SWIGTYPE_p_globus_replica_catalog_attribute_set_t, 0);
            PyTuple_SET_ITEM(tuple, 1, ptr);
        }
    }       
    if (attrNames != Py_None)
        globus_libc_free(attrs);
    return tuple;
}

PyObject*
attribute_set_destroy(globus_replica_catalog_attribute_set_t* set) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "attribute_set_destroy: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_attribute_set_destroy(set);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
attribute_set_first(globus_replica_catalog_attribute_set_t* set) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "attribute_set_first: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_attribute_set_first(set);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
attribute_set_next(globus_replica_catalog_attribute_set_t* set) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "attribute_set_next: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_attribute_set_next(set);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}


extern globus_bool_t
globus_replica_catalog_attribute_set_more(globus_replica_catalog_attribute_set_t* set);

PyObject*
attribute_set_get_name(globus_replica_catalog_attribute_set_t* set) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    char* name;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "attribute_set_get_name: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_attribute_set_get_name(set, &name);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", name));
        globus_libc_free(name);
    }
    return tuple;
}

PyObject*
attribute_set_get_values(globus_replica_catalog_attribute_set_t* set) {
    globus_result_t ret;
    PyObject *tuple, *valList;
    char* tmpstr;
    char** values;
    int i;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "attribute_set_get_values: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_attribute_set_get_values(set, &values);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        valList = PyList_New(0);
        for (i=0; values[i]; i++) {
            PyList_Append(valList, Py_BuildValue("s", values[i]));
        }
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, valList);
        globus_replica_catalog_attribute_set_free_values(values);
    }
    return tuple;
}

PyObject*
entry_set_destroy(globus_replica_catalog_entry_set_t* set) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "entry_set_destroy: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_entry_set_destroy(set);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
entry_set_first(globus_replica_catalog_entry_set_t* set) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "entry_set_first: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_entry_set_first(set);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
entry_set_next(globus_replica_catalog_entry_set_t* set) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "entry_set_next: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_entry_set_next(set);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}


extern globus_bool_t
globus_replica_catalog_entry_set_more(globus_replica_catalog_entry_set_t* set);

PyObject*
entry_set_count(globus_replica_catalog_entry_set_t* set) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    int count;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "entry_set_count: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_entry_set_count(set, &count);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", count));
    }
    return tuple;
}

PyObject*
entry_set_get_dn(globus_replica_catalog_entry_set_t* set) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    char** dn;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "entry_set_get_dn: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_entry_set_get_dn(set, dn);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", dn));
        globus_replica_catalog_entry_set_free_dn(*dn);
    }
    return tuple;
}

PyObject*
entry_set_get_url(globus_replica_catalog_entry_set_t* set) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    char** url;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "entry_set_get_url: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_entry_set_get_url(set, url);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", url));
        globus_libc_free(*url);
    }
    return tuple;
}

PyObject*
entry_set_get_name(globus_replica_catalog_entry_set_t* set) {
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    char** name;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "entry_set_get_name: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_catalog_entry_set_get_name(set, name);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", *name));
        globus_libc_free(*name);
    }
    return tuple;
}

PyObject*
entry_set_get_attributes(globus_replica_catalog_entry_set_t* entrySet) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_replica_catalog_attribute_set_t* set;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "entry_set_get_attributes: PyTuple_New()");
        return NULL;
    }
    set = (globus_replica_catalog_attribute_set_t*)globus_malloc(
                                    sizeof(globus_replica_catalog_attribute_set_t));
    ret = globus_replica_catalog_attribute_set_init(set);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(set);
    }
    else {
        ret = globus_replica_catalog_entry_set_get_attributes(entrySet, set);
        if (ret != GLOBUS_SUCCESS) {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
            tmpstr = globus_object_printable_to_string(globus_error_get(ret));
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
            globus_libc_free(set);
        }
        else {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            ptr = SWIG_NewPointerObj((void*)set, 
                                     SWIGTYPE_p_globus_replica_catalog_attribute_set_t, 0);
            PyTuple_SET_ITEM(tuple, 1, ptr);
        }
    }       
    return tuple;
}
%}
