/*############################################################################
# Keith R. Jackson, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################*/
%module replicaManagementc
%{
#include "globus_common.h"
#include "globus_replica_management.h"
%}

%include ./src/pyGlobus-typemaps.i

%types(globus_ftp_client_operationattr_t*);

typedef int globus_bool_t;
typedef unsigned char globus_byte_t;

typedef enum
{
    GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_DISALLOW,
    GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_SAFE,
    GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_UNSAFE
} globus_replica_management_overwrite_behavior_t;

%apply globus_bool_t *boolRet { globus_bool_t *caching, 
                                    globus_bool_t *third_party_transfer };
%apply unsigned int *uintRet { unsigned int *frequency };

%{
static PyThreadState *thread_state = NULL;

typedef struct {
    PyObject* callback;
    PyObject* arg;
} callbackStruct;

static globus_result_t PythonCallback(void* user_data, char* url) {
    PyObject *func, *arglist, *result, *arg;
    globus_result_t ret;
    int i;
    callbackStruct* cbStruct = (callbackStruct*)user_data;
    PyEval_AcquireThread(thread_state);
    func = cbStruct->callback;
    arg = cbStruct->arg;
    arglist = Py_BuildValue("(Os)", arg, url);
    result = PyEval_CallObject(func, arglist);
    i = (int)PyInt_AsLong(result);
    if (i != 0) 
        ret = GLOBUS_REPLICA_MANAGEMENT_ERROR("Error from preregister callback",
                                              "PythonCallback");
    else
        ret = GLOBUS_SUCCESS;
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(thread_state);
    return ret; 
}

%}

%inline %{
globus_module_descriptor_t*
get_module(void) {
    return GLOBUS_REPLICA_MANAGEMENT_MODULE;
}

PyObject*
handleattr_init() {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_replica_management_handleattr_t* attr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_init: PyTuple_New()");
        return NULL;
    }
    attr = (globus_replica_management_handleattr_t*)
	globus_malloc(sizeof(globus_replica_management_handleattr_t));
    ret = globus_replica_management_handleattr_init(attr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(attr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)attr, 
		   SWIGTYPE_p_globus_replica_management_handleattr_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
handleattr_copy(globus_replica_management_handleattr_t* srcAttr) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_replica_management_handleattr_t* attr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_copy: PyTuple_New()");
        return NULL;
    }
    attr = (globus_replica_management_handleattr_t*)
	globus_malloc(sizeof(globus_replica_management_handleattr_t));
    ret = globus_replica_management_handleattr_copy(attr, srcAttr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(attr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)attr, 
		     SWIGTYPE_p_globus_replica_management_handleattr_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
handleattr_destroy(globus_replica_management_handleattr_t* attr) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_destroy: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_management_handleattr_destroy(attr);
    globus_free(attr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
handleattr_set_caching(globus_replica_management_handleattr_t* attr,
                       globus_bool_t caching) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_set_caching: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_management_handleattr_set_caching(attr, caching);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
handleattr_get_caching(globus_replica_management_handleattr_t* attr,
                       globus_bool_t *caching) {
    globus_result_t ret;
    PyObject* list;
    char* tmpstr;
    list = PyList_New(0);
    ret = globus_replica_management_handleattr_get_caching(attr, caching);
    if (ret != GLOBUS_SUCCESS) {
        PyList_Append(list, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyList_Append(list, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyList_Append(list, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyList_Append(list, Py_None);
    }
    return list;
}

PyObject*
handleattr_set_ftp_update_frequency(globus_replica_management_handleattr_t* attr,
                       unsigned int frequency) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_set_ftp_update_frequency: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_management_handleattr_set_ftp_update_frequency(attr, frequency);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
handleattr_get_ftp_update_frequency(globus_replica_management_handleattr_t* attr,
                       unsigned int *frequency) {
    globus_result_t ret;
    PyObject* list;
    char* tmpstr;
    list = PyList_New(0);
    ret = globus_replica_management_handleattr_get_ftp_update_frequency(attr, frequency);
    if (ret != GLOBUS_SUCCESS) {
        PyList_Append(list, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyList_Append(list, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyList_Append(list, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyList_Append(list, Py_None);
    }
    return list;
}

PyObject*
handleattr_set_ftp_handleattr(globus_replica_management_handleattr_t* attr,
                              globus_ftp_client_handleattr_t* ftpAttr) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_set_ftp_handleattr: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_management_handleattr_set_ftp_handleattr(attr, ftpAttr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
handleattr_get_ftp_handleattr(globus_replica_management_handleattr_t* attr) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_ftp_client_handleattr_t* ftpAttr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_get_ftp_handleattr: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_management_handleattr_get_ftp_handleattr(attr, ftpAttr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)ftpAttr,
                                 SWIGTYPE_p_globus_ftp_client_handleattr_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
handleattr_set_third_party_transfer(globus_replica_management_handleattr_t* attr,
                       globus_bool_t third_party_transfer) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handleattr_set_third_party_transfer: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_management_handleattr_set_third_party_transfer(attr, 
                                                                        third_party_transfer);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
handleattr_get_third_party_transfer(globus_replica_management_handleattr_t* attr,
                       globus_bool_t *third_party_transfer) {
    globus_result_t ret;
    PyObject* list;
    char* tmpstr;
    list = PyList_New(0);
    ret = globus_replica_management_handleattr_get_third_party_transfer(attr, 
                                                                        third_party_transfer);
    if (ret != GLOBUS_SUCCESS) {
        PyList_Append(list, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyList_Append(list, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyList_Append(list, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyList_Append(list, Py_None);
    }
    return list;
}

PyObject*
handle_init(globus_replica_management_handleattr_t* attr) {
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_replica_management_handle_t* handle;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handle_init: PyTuple_New()");
        return NULL;
    }
    handle = (globus_replica_management_handle_t*)globus_malloc(sizeof(globus_replica_management_handle_t));
    ret = globus_replica_management_handle_init(handle, attr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
        globus_libc_free(handle);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)handle, 
				 SWIGTYPE_p_globus_replica_management_handle_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
handle_destroy(globus_replica_management_handle_t* handle) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handle_destroy: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_management_handle_destroy(handle);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    globus_free(handle);
    return tuple;
}

PyObject*
restart(globus_replica_management_handle_t* handle,
        char* rollbackInfo, PyObject* pyHandleAttr, 
        PyObject* pySrcAttr, PyObject* pyDestAttr,
        PyObject *pyfunc, PyObject *arg) {
    globus_result_t ret;
    PyObject *tuple, *cbPtr;
    callbackStruct* cbStruct;
    char* tmpstr;
    globus_replica_catalog_collection_handleattr_t* handleAttr;
    globus_ftp_client_operationattr_t* srcAttr;
    globus_ftp_client_operationattr_t* destAttr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "restart: PyTuple_New()");
        return NULL;
    }
    if (pyHandleAttr == Py_None) 
        handleAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pyHandleAttr, (void**)&handleAttr,
                        SWIGTYPE_p_globus_replica_catalog_collection_handleattr_t, 0);
    if (pySrcAttr == Py_None)
        srcAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pySrcAttr, (void**)&srcAttr,
                        SWIGTYPE_p_globus_ftp_client_operationattr_t, 0);
    if (pyDestAttr == Py_None)
        destAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pyDestAttr, (void**)&destAttr,
                        SWIGTYPE_p_globus_ftp_client_operationattr_t, 0);
    if (pyfunc != Py_None) {
        cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
        Py_XINCREF(pyfunc);
        Py_XINCREF(arg);
        cbStruct->callback = pyfunc;
        cbStruct->arg = arg;
        if (thread_state == NULL) {
            PyEval_InitThreads();
            thread_state = PyThreadState_New(PyThreadState_Get()->interp);
        }
        ret = globus_replica_management_restart(handle, rollbackInfo, handleAttr,
                                                srcAttr, destAttr, PythonCallback, cbStruct);
        if (ret != GLOBUS_SUCCESS) {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
            tmpstr = globus_object_printable_to_string(globus_error_get(ret));
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
            globus_libc_free(cbStruct);
        }
        else {
            cbPtr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            PyTuple_SET_ITEM(tuple, 1, cbPtr);
        }
    }
    else {
        ret = globus_replica_management_restart(handle, rollbackInfo, handleAttr,
                                                srcAttr, destAttr, GLOBUS_NULL, GLOBUS_NULL);
        if (ret != GLOBUS_SUCCESS) {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
            tmpstr = globus_object_printable_to_string(globus_error_get(ret));
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
        }
        else {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            Py_INCREF(Py_None);
            PyTuple_SET_ITEM(tuple, 1, Py_None);
        }
    }
    return tuple; 
}

PyObject*
rollback(globus_replica_management_handle_t* handle, char* info, 
         PyObject* pyColAttr, PyObject* pyFtpAttr) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    globus_replica_catalog_collection_handleattr_t* colAttr;
    globus_ftp_client_operationattr_t* ftpAttr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handle_destroy: PyTuple_New()");
        return NULL;
    }
    if (pyColAttr == Py_None) 
        colAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pyColAttr, (void**)&colAttr,
                        SWIGTYPE_p_globus_replica_catalog_collection_handleattr_t, 0);
    if (pyFtpAttr == Py_None)
        ftpAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pyFtpAttr, (void**)&ftpAttr,
                        SWIGTYPE_p_globus_ftp_client_operationattr_t, 0);
    ret = globus_replica_management_rollback(handle, info, colAttr, ftpAttr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
get_rollback_contact(globus_replica_management_handle_t* handle) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    char** str;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "get_rollback_contact: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_management_get_rollback_contact(handle, str);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", str));
        globus_libc_free(*str);
    }
    return tuple;
}

PyObject*
collection_create(globus_replica_management_handle_t* handle,
                  char* url, globus_replica_catalog_collection_handleattr_t* colAttr) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "collection_create: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_management_collection_create(handle, url, colAttr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
collection_delete(globus_replica_management_handle_t* handle,
                  char* url, globus_replica_catalog_collection_handleattr_t* colAttr) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "collection_delete: PyTuple_New()");
        return NULL;
    }
    ret = globus_replica_management_collection_delete(handle, url, colAttr);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
location_create(globus_replica_management_handle_t* handle,
                char* url, PyObject* pyColAttr, char* name, 
                PyObject* pyFtpAttr, char* rootUrl) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    globus_replica_catalog_collection_handleattr_t* colAttr;
    globus_ftp_client_operationattr_t* ftpAttr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "location_create: PyTuple_New()");
        return NULL;
    }
    if (pyColAttr == Py_None) 
        colAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pyColAttr, (void**)&colAttr,
                        SWIGTYPE_p_globus_replica_catalog_collection_handleattr_t, 0);
    if (pyFtpAttr == Py_None)
        ftpAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pyFtpAttr, (void**)&ftpAttr,
                        SWIGTYPE_p_globus_ftp_client_operationattr_t, 0);
    ret = globus_replica_management_location_create(handle, url, 
                                                    colAttr, name, ftpAttr, rootUrl);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
location_delete(globus_replica_management_handle_t* handle,
                  char* url, PyObject* pyColAttr, char* name) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    globus_replica_catalog_collection_handleattr_t* colAttr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "location_delete: PyTuple_New()");
        return NULL;
    }
    if (pyColAttr == Py_None) 
        colAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pyColAttr, (void**)&colAttr,
                        SWIGTYPE_p_globus_replica_catalog_collection_handleattr_t, 0);
    ret = globus_replica_management_location_delete(handle, url, colAttr, name);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
file_register(globus_replica_management_handle_t* handle,
                char* url, PyObject* pyColAttr, char* name, 
                PyObject* pyFtpAttr, char* filename) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    globus_replica_catalog_collection_handleattr_t* colAttr;
    globus_ftp_client_operationattr_t* ftpAttr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "file_register: PyTuple_New()");
        return NULL;
    }
    if (pyColAttr == Py_None) 
        colAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pyColAttr, (void**)&colAttr,
                        SWIGTYPE_p_globus_replica_catalog_collection_handleattr_t, 0);
    if (pyFtpAttr == Py_None)
        ftpAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pyFtpAttr, (void**)&ftpAttr,
                        SWIGTYPE_p_globus_ftp_client_operationattr_t, 0);
    ret = globus_replica_management_file_register(handle, url, 
                                                    colAttr, name, ftpAttr, filename);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
file_publish(globus_replica_management_handle_t* handle,
             char* srcUrl, PyObject* pySrcAttr, char* destUrl, 
             PyObject* pyColAttr, char* locationName, 
             PyObject* pyDestAttr, char* filename,
             globus_replica_management_overwrite_behavior_t behavior,
             PyObject *pyfunc, PyObject *arg) {
    globus_result_t ret;
    PyObject *tuple, *cbPtr;
    callbackStruct* cbStruct;
    char* tmpstr;
    globus_replica_catalog_collection_handleattr_t* colAttr;
    globus_ftp_client_operationattr_t* srcAttr;
    globus_ftp_client_operationattr_t* destAttr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "file_publish: PyTuple_New()");
        return NULL;
    }
    if (pyColAttr == Py_None) 
        colAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pyColAttr, (void**)&colAttr,
                        SWIGTYPE_p_globus_replica_catalog_collection_handleattr_t, 0);
    if (pySrcAttr == Py_None)
        srcAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pySrcAttr, (void**)&srcAttr,
                        SWIGTYPE_p_globus_ftp_client_operationattr_t, 0);
    if (pyDestAttr == Py_None)
        destAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pyDestAttr, (void**)&destAttr,
                        SWIGTYPE_p_globus_ftp_client_operationattr_t, 0);
    if (pyfunc != Py_None) {
        cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
        Py_XINCREF(pyfunc);
        Py_XINCREF(arg);
        cbStruct->callback = pyfunc;
        cbStruct->arg = arg;
        if (thread_state == NULL) {
            PyEval_InitThreads();
            thread_state = PyThreadState_New(PyThreadState_Get()->interp);
        }
        ret = globus_replica_management_file_publish(handle, srcUrl, srcAttr,
                                                     destUrl, colAttr, locationName, destAttr, 
                                                     filename, behavior, PythonCallback, cbStruct);
        if (ret != GLOBUS_SUCCESS) {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
            tmpstr = globus_object_printable_to_string(globus_error_get(ret));
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
            globus_libc_free(cbStruct);
        }
        else {
            cbPtr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            PyTuple_SET_ITEM(tuple, 1, cbPtr);
        }
    }
    else {
        ret = globus_replica_management_file_publish(handle, srcUrl, srcAttr,
                                                     destUrl, colAttr, locationName, destAttr, 
                                                     filename, behavior, GLOBUS_NULL, GLOBUS_NULL);
        if (ret != GLOBUS_SUCCESS) {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
            tmpstr = globus_object_printable_to_string(globus_error_get(ret));
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
        }
        else {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            Py_INCREF(Py_None);
            PyTuple_SET_ITEM(tuple, 1, Py_None);
        }
    }
    return tuple; 
}

PyObject*
file_copy(globus_replica_management_handle_t* handle,
          char* url, PyObject* pyColAttr, char* srcLocationName, 
          PyObject* pySrcAttr, char* destLocationName, 
          PyObject* pyDestAttr, char* filename,
          globus_replica_management_overwrite_behavior_t behavior,
          PyObject *pyfunc, PyObject *arg) {
    globus_result_t ret;
    PyObject *tuple, *cbPtr;
    callbackStruct* cbStruct;
    char* tmpstr;
    globus_replica_catalog_collection_handleattr_t* colAttr;
    globus_ftp_client_operationattr_t* srcAttr;
    globus_ftp_client_operationattr_t* destAttr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "file_copy: PyTuple_New()");
        return NULL;
    }
    if (pyColAttr == Py_None) 
        colAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pyColAttr, (void**)&colAttr,
                        SWIGTYPE_p_globus_replica_catalog_collection_handleattr_t, 0);
    if (pySrcAttr == Py_None)
        srcAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pySrcAttr, (void**)&srcAttr,
                        SWIGTYPE_p_globus_ftp_client_operationattr_t, 0);
    if (pyDestAttr == Py_None)
        destAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pyDestAttr, (void**)&destAttr,
                        SWIGTYPE_p_globus_ftp_client_operationattr_t, 0);
    if (pyfunc != Py_None) {
        cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
        Py_XINCREF(pyfunc);
        Py_XINCREF(arg);
        cbStruct->callback = pyfunc;
        cbStruct->arg = arg;
        if (thread_state == NULL) {
            PyEval_InitThreads();
            thread_state = PyThreadState_New(PyThreadState_Get()->interp);
        }
        ret = globus_replica_management_file_copy(handle, url, colAttr,
                                                  srcLocationName, srcAttr, destLocationName, destAttr, 
                                                  filename, behavior, PythonCallback, cbStruct);
        if (ret != GLOBUS_SUCCESS) {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
            tmpstr = globus_object_printable_to_string(globus_error_get(ret));
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
            globus_libc_free(cbStruct);
        }
        else {
            cbPtr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            PyTuple_SET_ITEM(tuple, 1, cbPtr);
        }
    }
    else {
        ret = globus_replica_management_file_copy(handle, url, colAttr,
                                                     srcLocationName, srcAttr, destLocationName, destAttr, 
                                                     filename, behavior, GLOBUS_NULL, GLOBUS_NULL);
        if (ret != GLOBUS_SUCCESS) {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
            tmpstr = globus_object_printable_to_string(globus_error_get(ret));
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
        }
        else {
            PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            Py_INCREF(Py_None);
            PyTuple_SET_ITEM(tuple, 1, Py_None);
        }
    }
    return tuple; 
}

PyObject*
file_delete(globus_replica_management_handle_t* handle,
                char* url, PyObject* pyColAttr, char* name, 
                PyObject* pyFtpAttr, char* filename, globus_bool_t delete) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    globus_replica_catalog_collection_handleattr_t* colAttr;
    globus_ftp_client_operationattr_t* ftpAttr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "file_delete: PyTuple_New()");
        return NULL;
    }
    if (pyColAttr == Py_None) 
        colAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pyColAttr, (void**)&colAttr,
                        SWIGTYPE_p_globus_replica_catalog_collection_handleattr_t, 0);
    if (pyFtpAttr == Py_None)
        ftpAttr = GLOBUS_NULL;
    else
        SWIG_ConvertPtr(pyFtpAttr, (void**)&ftpAttr,
                        SWIGTYPE_p_globus_ftp_client_operationattr_t, 0);
    ret = globus_replica_management_file_delete(handle, url, 
                                                colAttr, name, ftpAttr, filename, delete);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_libc_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

int
free_callback(callbackStruct* cbStruct) {
    int ret;
    Py_XDECREF(cbStruct->callback);
    Py_XDECREF(cbStruct->arg);
    globus_free(cbStruct);
    return ret;
}

%}
