/*############################################################################
# Joshua R. Boverhof, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################*/
/* 
 *  Parent Cred is only needed for gss_name
 *
 * pcd->ucert = NULL;        *1*
 * pcd->upkey = NULL;        *2*
 * pcd->cert_chain = NULL;   *3*
 * pcd->gs_ctx = NULL;
 * pcd->hSession = 0;
 * pcd->hPrivKey = 0;
 * pcd->certdir = NULL;  
 * pcd->certfile = NULL;
 * pcd->num_null_enc_ciphers = 0;
 * pcd->type = CRED_TYPE_PERMANENT;
 * pcd->owner = CRED_OWNER_USER;
 */

//#include "gssutils.h"
//#include "gssapi_ssleay.h"

%module sslutilsc
%include ./src/pyGlobus-typemaps.i
%{
#include <wait.h>
#include <stdio.h>

#include "openssl/crypto.h"
#include "openssl/ssl.h"
#include "openssl/err.h"
#include "openssl/bio.h"
#include "openssl/pem.h"
#include "openssl/x509.h"

#if SSLEAY_VERSION_NUMBER >= 0x0090581fL
#include "openssl/x509v3.h"
#endif

#include "openssl/stack.h"

#include "globus_common.h"
#include "globus_gss_assist.h"
#include "globus_io.h"

#include "gssapi.h"
#include "sslutils.h"

/**********************************************************************
gssapi_ssleay.h:

Description:
        This header file used internally by the gssapi_ssleay
        routines

CVS Information:

        $Source: /home/g5/repository/pyGlobus/src/Attic/sslutils.i,v $
        $Date: 2003/05/08 17:48:46 $
        $Revision: 1.4.2.1 $
        $Author: mateo $

**********************************************************************/

/**********************************************************************
                               Define constants
**********************************************************************/

#define GSS_I_CTX_INITIALIZED                       1
#define GSS_I_DISALLOW_ENCRYPTION                   2
#define GSS_I_PROTECTION_FAIL_ON_CONTEXT_EXPIRATION 4


#define GSS_C_QOP_GLOBUS_GSSAPI_SSLEAY_BIG 1
/*
 * Use the SSLeay error facility with the ERR_LIB_USER
 */

/* Offset from ERR_LIB_USER where GSSERR library will be stored */
#define ERR_USER_LIB_GSSERR_NUMBER                  ((ERR_LIB_USER) + 2)

#define GSSerr(f,r) ERR_PUT_error(ERR_USER_LIB_GSSERR_NUMBER,(f),(r),ERR_file_name,__LINE__)

/*
 * defines for function codes our minor error codes
 * These match strings defined in gsserr.c
 */

#define GSSERR_F_BASE                   100

#define GSSERR_F_ACCEPT_SEC             GSSERR_F_BASE + 0
#define GSSERR_F_ACQUIRE_CRED           GSSERR_F_BASE + 1
#define GSSERR_F_COMPARE_NAME           GSSERR_F_BASE + 2
#define GSSERR_F_DELETE_SEC             GSSERR_F_BASE + 3
#define GSSERR_F_EXPORT_NAME            GSSERR_F_BASE + 4
#define GSSERR_F_GLOBUSFILE             GSSERR_F_BASE + 5
#define GSSERR_F_IMPORT_NAME            GSSERR_F_BASE + 6
#define GSSERR_F_INIT_SEC               GSSERR_F_BASE + 7
#define GSSERR_F_RELEASE_BUFFER         GSSERR_F_BASE + 8
#define GSSERR_F_RELEASE_CRED           GSSERR_F_BASE + 9
#define GSSERR_F_RELEASE_NAME           GSSERR_F_BASE + 10
/* In gssutil.c: */
#define GSSERR_F_NAME_TO_NAME           GSSERR_F_BASE + 11
#define GSSERR_F_CREATE_FILL            GSSERR_F_BASE + 12
#define GSSERR_F_GS_HANDSHAKE           GSSERR_F_BASE + 13
#define GSSERR_F_GS_RETRIVE_PEER        GSSERR_F_BASE + 14

#define GSSERR_F_WRAP                   GSSERR_F_BASE + 15
#define GSSERR_F_UNWRAP                 GSSERR_F_BASE + 16
#define GSSERR_F_GET_MIC                GSSERR_F_BASE + 17
#define GSSERR_F_VERIFY_MIC             GSSERR_F_BASE + 18

#define GSSERR_F_IMPORT_SEC             GSSERR_F_BASE + 19
#define GSSERR_F_EXPORT_SEC             GSSERR_F_BASE + 20

#define GSSERR_F_IMPORT_CRED            GSSERR_F_BASE + 21
#define GSSERR_F_EXPORT_CRED            GSSERR_F_BASE + 22
#define GSSERR_F_READ                   GSSERR_F_BASE + 23
#define GSSERR_F_WRITE                  GSSERR_F_BASE + 24
#define GSSERR_F_INIT_DELEGATION        GSSERR_F_BASE + 25
#define GSSERR_F_ACCEPT_DELEGATION      GSSERR_F_BASE + 26
#define GSSERR_F_INQUIRE_BY_OID         GSSERR_F_BASE + 27
#define GSSERR_F_INQUIRE_CONTEXT        GSSERR_F_BASE + 28
#define GSSERR_F_ADD_OID_SET_MEMBER     GSSERR_F_BASE + 29
#define GSSERR_F_CREATE_EMPTY_OID_SET   GSSERR_F_BASE + 30
#define GSSERR_F_TEST_OID_SET_MEMBER    GSSERR_F_BASE + 31

/*
 * GSI minor error code
 *
 * There are three types of GSI minor error codes. There are the codes
 * defined in the gssapi library (as defined below starting at code 100).
 * There are the codes defined in the ssl_utils library (as defined in
 * sslutils.h starting at code 1000) and then there are codes that come up
 * from openssl. These codes are indicated by having the high bit set. See
 * below for more details.
 *
 * Defines for reasons in gssapi library.
 * The match strings defined in gsserr.c
 * These are also used for the minor_status codes.
 * Need to make sure these don't overlap with errors in sslutils.h
 */

#define GSSERR_R_BASE                           100

#define GSSERR_R_HANDSHAKE                      GSSERR_R_BASE + 0
#define GSSERR_R_NO_GLOBUSID                    GSSERR_R_BASE + 1
#define GSSERR_R_PROCESS_CERT                   GSSERR_R_BASE + 2
#define GSSERR_R_MUTUAL_AUTH                    GSSERR_R_BASE + 3
#define GSSERR_R_WRAP_BIO                       GSSERR_R_BASE + 4
#define GSSERR_R_PROXY_VIOLATION                GSSERR_R_BASE + 5
#define GSSERR_R_PROXY_NOT_RECEIVED             GSSERR_R_BASE + 6
#define GSSERR_R_BAD_ARGUMENT                   GSSERR_R_BASE + 7
#define GSSERR_R_IMPEXP_BIO_SSL                 GSSERR_R_BASE + 8
#define GSSERR_R_IMPEXP_NO_CIPHER               GSSERR_R_BASE + 9
#define GSSERR_R_IMPEXP_BAD_LEN                 GSSERR_R_BASE + 10
#define GSSERR_R_CLASS_ADD_EXT                  GSSERR_R_BASE + 11
#define GSSERR_R_EXPORT_FAIL                    GSSERR_R_BASE + 12
#define GSSERR_R_IMPORT_FAIL                    GSSERR_R_BASE + 13
#define GSSERR_R_READ_BIO                       GSSERR_R_BASE + 14
#define GSSERR_R_WRITE_BIO                      GSSERR_R_BASE + 15
#define GSSERR_R_UNEXPECTED_FORMAT              GSSERR_R_BASE + 17
#define GSSERR_R_BAD_DATE                       GSSERR_R_BASE + 20
#define GSSERR_R_BAD_MECH                       GSSERR_R_BASE + 21
#define GSSERR_R_ADD_EXT                        GSSERR_R_BASE + 22
#define GSSERR_R_REMOTE_CERT_VERIFY_FAILED      GSSERR_R_BASE + 23
#define GSSERR_R_OUT_OF_MEMORY                  GSSERR_R_BASE + 24
/* NOTE: Don't go over 1000 here or will conflict with errors in sslutils.h */

/* Old error codes in case anyone is using them */
#define GSSERR_R_IMPEXP_BAD_PARMS       GSSERR_R_BAD_ARGUMENT
#define GSSERR_R_PASSED_NULL_PARAMETER  GSSERR_R_BAD_ARGUMENT

/*
 * Flag used on minor_status to indicate an error from the openssl
 * libraries that we have caught and converted to an error code as
 * defined above, in sslutils.h or in scutils.h
 *
 * If this flag is set then the rest of the 32-bit minor status is
 * the error code as returned from the openssl library.
 */
#define GSI_SSL_ERROR_FLAG                      0x8000

#define GSI_IS_SSL_ERROR(minor)         (minor & GSI_SSL_ERROR_FLAG)


/*
 * we need to distinguish between a token
 * created by us using get_mic vs one using
 * the SSL application data
 * We use this in wrap and unwrap
 * Future versions of SSL may use this
 *
 * Our wraped buffer (integrity only) has
 *
 *  byte  type[1]          = SSL3_RT_GSSAPI_SSLEAY
 *  byte  version_major[1] = 0x03
 *  byte  version_minor[1] = 0
 *  byte  mic_length[2]    = 2 byte length of following mic
 *
 *  byte  mic_seq[8]           = 8 byte sequence number
 *  byte  mic_data_length[4]   = 4 byte length of data
 *  byte  hash[*]          = the hash of variable length
 *
 *  byte  data[*]          = the data being wrapped.
 */

#define SSL3_RT_GSSAPI_SSLEAY                   26



/* These conversions macros are taken from SSL */

#define l2n(l,c)   (*((c)++)=(unsigned char)(((l)>>24)&0xff), \
                    *((c)++)=(unsigned char)(((l)>>16)&0xff), \
                    *((c)++)=(unsigned char)(((l)>> 8)&0xff), \
                    *((c)++)=(unsigned char)(((l)    )&0xff))

#define n2l(c,l)   (l =((unsigned long)(*((c)++)))<<24, \
                    l|=((unsigned long)(*((c)++)))<<16, \
                    l|=((unsigned long)(*((c)++)))<< 8, \
                    l|=((unsigned long)(*((c)++))))

#define n2s(c,s)    (s =((unsigned int)(*((c)++)))<< 8, \
                                 s|=((unsigned int)(*((c)++))))

#define s2n(s,c)    (*((c)++)=(unsigned char)(((s)>> 8)&0xff), \
                     *((c)++)=(unsigned char)(((s)    )&0xff))

/* ssl_locl.h is not installed, so we define a few needed items */

#ifndef SSL_eNULL
#define SSL_eNULL       0x00010000L
#endif

#ifndef SSL_STRONG_MASK
#if  SSLEAY_VERSION_NUMBER >= 0x0090581fL
#define  SSL_STRONG_MASK     0x0000007cL
#else
#define SSL_STRONG_MASK     0x07000000L
#endif
#endif

#ifndef SSL_LOW
#if  SSLEAY_VERSION_NUMBER >= 0x0090581fL
#define SSL_LOW         0x00000010L
#else
#define SSL_LOW         0x01000000L
#endif
#endif
/* Compare OIDs */

#define g_OID_equal(o1,o2) \
        (((o1) == (o2)) || \
         ((o1) && (o2) && \
         ((o1)->length == (o2)->length) && \
         (memcmp((o1)->elements,(o2)->elements,(int) (o1)->length) == 0)))


/**********************************************************************
                               Type definitions
**********************************************************************/

/*
 * The SSL ssl_locl.h is a private headerfile which does
 * not get installed. The ssl3_enc_method is needed for
 * the import/export so we include it here, in orged to avoid
 * changes to the SSLeay code.
 *DEE This needs to be looked at.
 */

#ifndef HEADER_SSL_LOCL_H
typedef struct ssl3_enc_method
{
    int                                 (*enc)();
    int                                 (*mac)();
    int                                 (*setup_key_block)();
    int                                 (*generate_master_secret)();
    int                                 (*change_cipher_state)();
    int                                 (*final_finish_mac)();
    int                                 finish_mac_length;
    int                                 (*cert_verify_mac)();
    unsigned char                       client_finished[20];
    int                                 client_finished_len;
    unsigned char                       server_finished[20];
    int                                 server_finished_len;
    int                                 (*alert_value)();
} SSL3_ENC_METHOD;
#endif

typedef enum {
    GS_CON_ST_HANDSHAKE = 0,
    GS_CON_ST_FLAGS,
    GS_CON_ST_REQ,
    GS_CON_ST_CERT,
    GS_CON_ST_DONE
} gs_con_st_t;

typedef enum
{
    GS_DELEGATION_START,
    GS_DELEGATION_DONE,
    GS_DELEGATION_COMPLETE_CRED,
    GS_DELEGATION_SIGN_CERT
} gs_delegation_state_t;

typedef struct gss_name_desc_struct {
    /* gss_buffer_desc  name_buffer ; */
    gss_OID                             name_oid;
    X509_NAME *                         x509n;
} gss_name_desc ;

typedef struct gss_cred_id_desc_struct {
    proxy_cred_desc *                   pcd;
    gss_name_desc *                     globusid;
    gss_cred_usage_t                    cred_usage;
    BIO *                               gs_bio_err;
} gss_cred_id_desc ;

typedef struct gss_ctx_id_desc_struct{
    proxy_verify_desc                   pvd; /* used for verify_callback */
    proxy_verify_ctx_desc               pvxd;
    gss_name_desc *                     source_name;
    gss_name_desc *                     target_name;
    gss_cred_id_desc *                  cred_handle;
    OM_uint32                           ret_flags;
    OM_uint32                           req_flags;
    OM_uint32                           ctx_flags;
    int                                 cred_obtained;
    SSL *                               gs_ssl;
    BIO *                               gs_rbio;
    BIO *                               gs_wbio;
    BIO *                               gs_sslbio;
    gs_con_st_t                         gs_state;
    int                                 locally_initiated;
    time_t                              goodtill;
    /* following used during delegation */

    /* new key for delegated proxy - do we need this now that we have
     * init/accept-delegation
     */
    EVP_PKEY *                          dpkey;
    /* delegated cert */
    X509 *                              dcert;
    /* delegation state */
    gs_delegation_state_t               delegation_state;
} gss_ctx_id_desc ;

/**********************************************************************
                               Global variables
**********************************************************************/

extern
const gss_OID_desc * const gss_mech_globus_gssapi_ssleay;

extern
const gss_OID_desc * const gss_restrictions_extension;

/**********************************************************************
                               Function prototypes
**********************************************************************/

int
ERR_load_gsserr_strings(int);

OM_uint32
convert_minor_codes(const int lib, const int reason);



/**********************************************************************

gssutils.h

Description:
        This header file used internally  to define the gssutils

CVS Information:

    $Source: /home/g5/repository/pyGlobus/src/Attic/sslutils.i,v $
    $Date: 2003/05/08 17:48:46 $
    $Revision: 1.4.2.1 $
    $Author: mateo $

**********************************************************************/

OM_uint32
gss_copy_name_to_name(
    gss_name_desc **                    output,
    const gss_name_desc *               input);


OM_uint32
gss_create_and_fill_context(
    gss_ctx_id_desc **                  context_handle,
    gss_cred_id_desc *                  cred_handle,
    const gss_cred_usage_t              cred_usage,
    OM_uint32                           req_flags);

OM_uint32
gss_create_and_fill_cred(
    gss_cred_id_t *                     output_cred_handle_P,
    const gss_cred_usage_t              cred_usage,
    X509   *                            ucert,
    EVP_PKEY *                          upkey,
    STACK_OF(X509)  *                   cert_chain,
    BIO *                               bp);

/* following added for ssleay */

OM_uint32
gs_handshake(
    gss_ctx_id_desc*                    context_handle);

OM_uint32
gs_get_token(
    const gss_ctx_id_desc *             context_handle,
    const gss_buffer_t                  output_token);

OM_uint32
gs_put_token(
    const gss_ctx_id_desc *             context_handle,
    const gss_buffer_t                  input_token);

OM_uint32
gs_retrieve_peer(
    gss_ctx_id_desc *                   context_handle,
    const gss_cred_usage_t              cred_usage);
%}

/*--------------------------------------------------------------------
 *  
 *  Start of Python Interface Code
 *
 *------------------------------------------------------------------*/


typedef int globus_bool_t;
typedef void* gss_name_t;
typedef void* gss_cred_id_t;
typedef void* gss_ctx_id_t;
typedef int gss_cred_usage_t;
typedef unsigned int gss_qop_t; /* these next two will need to be ifdef'd */
typedef unsigned int OM_uint32; /* for other platforms where they're not uint's */

#define STACK_OF(A) STACK

typedef struct gss_cred_id_desc_struct {
    proxy_cred_desc                  *pcd;
    gss_name_desc                    *globusid ;
    gss_cred_usage_t                  cred_usage ;
    BIO                              *gs_bio_err ;
} gss_cred_id_desc;

typedef struct proxy_cred_desc_struct {
    X509 *                              ucert;
    EVP_PKEY *                          upkey;
    STACK_OF(X509) *                    cert_chain;
    SSL_CTX *                           gs_ctx;
    /* smart card session handle */
    unsigned long                       hSession;
    /* private key session handle */
    unsigned long                       hPrivKey;
    char *                              certdir;
    char *                              certfile;
    int                                 num_null_enc_ciphers;
    int                                 type;  /* for gsi error messages */
    int                                 owner; /* for gsi error messages */
} *proxy_cred_desc;

%apply char **OUTPUT { char **charRet };

%inline %{
void
kpcallback(int p, int n) {
    char c='B';
    if (p == 0) c='.';
    if (p == 1) c='+';
    if (p == 2) c='*';
    if (p == 3) c='\n';
    fputc(c,stderr);
}

void
free_X509(X509_REQ** req){
    if(req){
	X509_free((X509*)*req);
    }
    return;
}

void
free_EVP_PKEY(EVP_PKEY** pkey){
    if(pkey){
	EVP_PKEY_free(*pkey);
    }
}

EVP_PKEY**
malloc_EVP_PKEY() {
    return (EVP_PKEY**)malloc(sizeof(EVP_PKEY**));
}

X509_REQ**
malloc_X509_REQ() {
    return (X509_REQ**)malloc(sizeof(X509_REQ**));
}

gss_cred_id_t*
get_initialized_gss_cred() {
    gss_cred_id_desc     **output_cred_handle = NULL;

    output_cred_handle  = (gss_cred_id_desc**)globus_malloc(sizeof(gss_cred_id_desc**));
    if (output_cred_handle == NULL) {
        GSSerr(GSSERR_F_ACQUIRE_CRED, GSSERR_R_OUT_OF_MEMORY);
        goto err;
    }
    (*output_cred_handle) = (gss_cred_id_desc*)globus_malloc(sizeof(gss_cred_id_desc*));
    (*output_cred_handle)->globusid = NULL;
    if (((*output_cred_handle)->pcd = proxy_cred_desc_new()) == NULL) {
        goto err;
    }
    if( (*output_cred_handle)->pcd->ucert == NULL)
	printf("XXX NULL\n");
    return (gss_cred_id_t*)output_cred_handle;
 err:
    //return GSS_S_FAILURE ;
    return NULL;
}

int
genreq(gss_cred_id_t* issuer_cred, gss_cred_id_t* sub_cred,
       X509_REQ** req, EVP_PKEY **npkey, const int bits) {
    PyObject          *tuple;
    proxy_cred_desc   *sub_pcd            = NULL;
    //proxy_cred_desc   *sub_pcd            = (*(gss_cred_id_desc**)sub_cred)->pcd;
    proxy_cred_desc   *ipcd               = (*(gss_cred_id_desc**)issuer_cred)->pcd;
    gss_cred_id_desc **output_cred_handle = (gss_cred_id_desc**) sub_cred;


    *npkey = (EVP_PKEY*)malloc(sizeof(EVP_PKEY*));
    *req = (X509_REQ*)malloc(sizeof(X509_REQ*));
    if (proxy_genreq(ipcd->ucert, req, npkey, bits,(int (*)())kpcallback, sub_pcd)){
        goto err;
    }
    X509_REQ_print_fp(stderr, *req);
    return 0;
 err:
    return -1;
}


void
set_pcd_upkey(gss_cred_id_t* cred, EVP_PKEY** pkey) {
    (*(gss_cred_id_desc**)(cred))->pcd->upkey = *pkey;
    return;
}

int
set_pcd_outfile(const char* outfile, gss_cred_id_t* cred) {
    int   err = 0;
    char *certfile    = NULL;
    char *keyfile     = NULL;
    char *certdir     = NULL;
    char *certcafile  = NULL;
    char *outputfile  = NULL;
    gss_cred_id_desc **    output_cred_handle = (gss_cred_id_desc**) cred;
    (*output_cred_handle)->globusid = NULL;
    (*output_cred_handle)->pcd->certfile = NULL;
    if(outfile == NULL) {
        if ( proxy_get_filenames((*output_cred_handle)->pcd,
                             0,
                             &certcafile,
                             &certdir,
                             &outputfile,
                             &certfile,
                             &keyfile) )
            goto err;
	(*output_cred_handle)->pcd->certfile = (char*)malloc(strlen(outfile) + 1);
	strcpy((*output_cred_handle)->pcd->certfile, outfile);
        free(certfile);
        free(keyfile);
        free(certdir);
        free(certcafile);
	free(outputfile);
    }
    if(outfile) {
	(*output_cred_handle)->pcd->certfile = (char*)malloc(strlen(outfile) + 1);
	strcpy((*output_cred_handle)->pcd->certfile, outfile);
    }
    return 0;
err:
    return -1;
}

/*
 * verify if the cert is good, i.e. is signed by one of the
 * trusted CAs.
 */
int
verify_pcd(gss_cred_id_t* cred) {
    int debug = 1;
    proxy_verify_desc     pvd;
    proxy_verify_ctx_desc pvxd;
    gss_cred_id_desc **    output_cred_handle = (gss_cred_id_desc**) cred;


    proxy_verify_ctx_init(&pvxd);
    proxy_verify_init(&pvd, &pvxd);
    pvxd.certdir = (*output_cred_handle)->pcd->certdir;
    if (proxy_verify_cert_chain((*output_cred_handle)->pcd->ucert, 
				(*output_cred_handle)->pcd->cert_chain,
				&pvd)) {
        fprintf(stderr,"verify OK\n");
    } else {
        goto err;
    }

    return 0;
err:
    {
        unsigned long l;
        char buf[256];
#if SSLEAY_VERSION_NUMBER  >= 0x00904100L
        const char *file;
#else
        char *file;
#endif
        const char *data;
        int line;
        int flags;
        printf("Error: \n");
        while ( ERR_peek_error() != 0 )
        {
            printf("-->\n");
            l=ERR_get_error_line_data(&file,&line,&data,&flags);
            if (debug)

                fprintf(stderr,
                        "%s:%s:%d%s\n",
                        ERR_error_string(l,buf),
                        file,
                        line,
                        data);

            else
                fprintf(stderr,
                        "%s%s\nFunction: %s\n",
                        ERR_reason_error_string(l),
                        data,
                        ERR_func_error_string(l));
        }
    }
    return -1;
}

char*
get_pcd_certdir(gss_cred_id_t* cred) {
    int                    i;
    char                  *s = NULL;
    gss_cred_id_desc     **output_cred_handle = (gss_cred_id_desc**) cred;
    if((*output_cred_handle)->pcd->certdir != NULL) {
       s = (*output_cred_handle)->pcd->certdir;
    }
    return s;
}

void
set_pcd_certdir(gss_cred_id_t* cred, char* certdir) {
    gss_cred_id_desc     **output_cred_handle = (gss_cred_id_desc**) cred;
    (*output_cred_handle)->pcd->certdir = (char*)strdup(certdir);
    return;
}

int
write_pcd_to_file(gss_cred_id_t* cred) {
    int                   debug = 1;
    BIO *                 bp = NULL;
    BIO *                 bio_err;
    X509 *                ncert = NULL;
    FILE                 *fpout = NULL;
    gss_cred_id_desc    **output_cred_handle = (gss_cred_id_desc**)cred;

    printf("outfile = %s\n", (*output_cred_handle)->pcd->certfile);
    fpout=fopen((*output_cred_handle)->pcd->certfile,"w");
    if (fpout == NULL)
    {
        PRXYerr(PRXYERR_F_LOCAL_CREATE,PRXYERR_R_PROBLEM_PROXY_FILE);
        ERR_add_error_data(2,"\n        Open failed for File=", (*output_cred_handle)->pcd->certfile);
        goto err;
    }

    if ((bio_err=BIO_new(BIO_s_file())) != NULL)
        BIO_set_fp(bio_err,stderr,BIO_NOCLOSE);

    if ((bp=BIO_new(BIO_s_file())) != NULL)
    {
        BIO_set_fp(bp,fpout,BIO_NOCLOSE);
    }
    printf("marshal\n");
    if (proxy_marshal_bp(bp, (*output_cred_handle)->pcd->ucert, (*output_cred_handle)->pcd->upkey,
			 ncert, (*output_cred_handle)->pcd->cert_chain))
    {
        goto err;
    }
    return 0;
err:
    {
        unsigned long l;
        char buf[256];
#if SSLEAY_VERSION_NUMBER  >= 0x00904100L
        const char *file;
#else
        char *file;
#endif
        const char *data;
        int line;
        int flags;
        printf("Error: \n");
        while ( ERR_peek_error() != 0 )
        {
            printf("-->\n");
            l=ERR_get_error_line_data(&file,&line,&data,&flags);
            if (debug)

            fprintf(stderr,
                    "%s:%s:%d%s\n",
                    ERR_error_string(l,buf),
                    file,
                    line,
                    data);

            else
                fprintf(stderr,
                        "%s%s\nFunction: %s\n",
                        ERR_reason_error_string(l),
                        data,
                        ERR_func_error_string(l));
        }
    }
    return -1;
}


PyObject*
grid_proxy_init(int verbose) {
    PyObject *            ptr;
    PyObject *            tuple;
    int                   debug       = verbose;
    int                   quiet       = (verbose)?0:1;
    int                   ret_status  = 0;
    int                   bits        = 512;   /* default proxy to 512 bits */
    int                   hours       = 12;    /* default to a 12 hour cert */
    int                   limit_proxy = 0;     /* dont restrict the proxy */
    int                   verify      = 1;
    int                   i;
    int                   j;
    int                   cert_chain_length = 0;
    char *                certfile    = NULL;
    char *                keyfile     = NULL;
    char *                outfile     = NULL;
    char *                certdir     = NULL;
    char *                certcafile  = NULL;
    char *                pin         = NULL;
    char *                argp;
    char *                program;
    proxy_cred_desc *     pcd         = NULL;
    proxy_cred_desc *     pcd2        = NULL;
    gss_cred_id_desc *    output_cred_handle = NULL;
    proxy_verify_desc     pvd;
    proxy_verify_ctx_desc pvxd;
    BIO *                 bio_err;
    X509 *                xcert;
    time_t                time_after;
    time_t                time_now;
    time_t                time_diff;
    time_t                time_after_proxy;
    ASN1_UTCTIME *        asn1_time = NULL;
#ifdef CLASS_ADD
    void *                class_add_buf = NULL;
    size_t                class_add_buf_len = 0;
#endif
    X509_STORE *         cert_ctx    = NULL;
    X509_LOOKUP *        lookup      = NULL;
    X509_STORE_CTX       csc;

    int                                 status = -1;
    FILE *                              fpout = NULL;
    X509 *                              ncert = NULL;
    EVP_PKEY *                          npkey, npkey2;
    X509_REQ *                          req, req2;
    BIO *                               bp = NULL;
    STACK_OF(X509_EXTENSION) *          extensions = NULL;
    X509_EXTENSION *                    ex = NULL;

    int (*pw_cb)() = NULL;
    output_cred_handle = (gss_cred_id_desc*)globus_malloc(sizeof(gss_cred_id_desc));

    ERR_load_prxyerr_strings(0);
    SSLeay_add_ssl_algorithms();

    EVP_set_pw_prompt("Enter GRID pass phrase:");
    
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "grid_proxy_init: PyTuple_New()");
        return NULL;
    }

    if ((pcd = proxy_cred_desc_new()) == NULL)
        goto err;

    if ( proxy_get_filenames(pcd,
                             0,
                             &certcafile,
                             &certdir,
                             &outfile,
                             &certfile,
                             &keyfile) )
        goto err;

        if (debug) {
                printf("Files being used:\n");
                printf("    cert_file: %s\n", certcafile ? certcafile:"none");
                printf("    cert_dir : %s\n", certdir ? certdir:"none");
                printf("    proxy    : %s\n", outfile ? outfile:"none");
                printf("    user_cert: %s\n", certfile ? certfile:"none");
                printf("    user_key : %s\n", keyfile ? keyfile:"none");
        }

    if (certdir)
                pcd->certdir = (char*)strdup(certdir);

    if (!strncmp(certfile,"SC:",3))
        EVP_set_pw_prompt("Enter card pin:");
    else
        EVP_set_pw_prompt("Enter GRID pass phrase:");

    if (proxy_load_user_cert(pcd, certfile, pw_cb, NULL))
                goto err;

    if (!quiet)
    {
        char *s = NULL;
        s = X509_NAME_oneline(X509_get_subject_name(pcd->ucert),NULL,0);
        printf("Your identity: %s\n", s);
        free(s);
    }

    if (!strncmp(keyfile,"SC:",3))
        EVP_set_pw_prompt("Enter card pin:");

    if (proxy_load_user_key(pcd, keyfile, pw_cb, NULL))
        goto err;
    else
        printf("user_key loaded\n");

    //if (strncmp("SC:",certfile,3) && !strcmp(certfile, keyfile))
    //{
        if (pcd->cert_chain == NULL)
            pcd->cert_chain = sk_X509_new_null();
        if (proxy_load_user_proxy(pcd->cert_chain, certfile, NULL) < 0)
            goto err;
        else
            printf("user_proxy cert_chain\n");
    //}

    /*
     * verify if the cert is good, i.e. is signed by one of the
     * trusted CAs.
     */
    if (verify)
    {
        proxy_verify_ctx_init(&pvxd);
        proxy_verify_init(&pvd, &pvxd);
        pvxd.certdir = certdir;
        if (proxy_verify_cert_chain(pcd->ucert,pcd->cert_chain,&pvd)) {
            fprintf(stderr,"verify OK\n");
        } else {
            fprintf(stderr,"verify failed\n");
            goto err;
        }
    }

    asn1_time = ASN1_UTCTIME_new();
    X509_gmtime_adj(asn1_time,0);
    time_now = ASN1_UTCTIME_mktime(asn1_time);
    if (!quiet)
    {
        printf("Creating proxy ");
        fflush(stdout);
    }

    if (proxy_genreq(pcd->ucert,&req,&npkey,bits,
                     (int (*)())kpcallback, pcd))
    {
        goto err;
    }
    printf("generated request\n");
    /*
     * Add proxy extensions
     */
#ifdef DEBUG
    fprintf(stderr,"Adding Extensions to request\n");
#endif
    if ((extensions = sk_X509_EXTENSION_new_null()) == NULL)
    {
        PRXYerr(PRXYERR_F_PROXY_SIGN,PRXYERR_R_CLASS_ADD_EXT);
        goto err;
    }

#ifdef CLASS_ADD
    if (class_add_buf && class_add_buf_len > 0)
    {
        if ((ex = proxy_extension_class_add_create(class_add_buf,
                                                   class_add_buf_len)) == NULL)
        {
            PRXYerr(PRXYERR_F_PROXY_SIGN,PRXYERR_R_CLASS_ADD_EXT);
            goto err;
        }

        if (!sk_X509_EXTENSION_push(extensions, ex))
        {
            PRXYerr(PRXYERR_F_PROXY_SIGN,PRXYERR_R_CLASS_ADD_EXT);
            goto err;
        }
    }
#endif

    /*
     *   req   - request
     *   ncert - certificate
     *   npkey - key
     *   pcd->ucert,  pcd->cert_chain
     */
    if (proxy_sign(pcd->ucert,
                   pcd->upkey,
                   req,
                   &ncert,
                   hours*60*60,
                   extensions,
                   limit_proxy))
    {
        printf("proxy_sign\n");
        goto err;
    }
    else
        printf("gpi signed cert\n");
       
  
    output_cred_handle->pcd = (proxy_cred_desc *)globus_malloc(sizeof(proxy_cred_desc));  
    output_cred_handle->pcd->ucert = ncert;
    output_cred_handle->pcd->upkey = npkey;
    //insert cert chain
    
    	
    output_cred_handle->pcd->cert_chain = sk_X509_new_null();
    
    
    i = sk_X509_insert(output_cred_handle->pcd->cert_chain, 
		       pcd->ucert, 
		       sk_X509_num(pcd->cert_chain));

      
    ptr = SWIG_NewPointerObj((void*)output_cred_handle, 
				 SWIGTYPE_p_gss_cred_id_t, 0);
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
    PyTuple_SET_ITEM(tuple, 1, ptr);
    return tuple;
err:
    {
        unsigned long l;
        char buf[256];
	char tmpstr[356];
#if SSLEAY_VERSION_NUMBER  >= 0x00904100L
        const char *file;
#else
        char *file;
#endif
        const char *data;
        int line;
        int flags;
        //printf("Error: \n");
        while ( ERR_peek_error() != 0 )
        {
            //printf("-->\n");
            l=ERR_get_error_line_data(&file,&line,&data,&flags);
	    
            if (debug)

            sprintf(tmpstr,
                    "%s:%s:%d%s\n",
                    ERR_error_string(l,buf),
                    file,
                    line,
                    data);

            
        }
    
    
    PyTuple_SET_ITEM(tuple,0, Py_BuildValue("i", -1));
    PyTuple_SET_ITEM(tuple,1, Py_BuildValue("s", tmpstr));
    return tuple;
    }
}

int
sign_request(const gss_cred_id_t* issuer_cred,
	     gss_cred_id_t*       sub_cred,
	     X509_REQ**           req)
{
    int bits = 512;
    int limit_proxy = 0;
    int hours = 12;
    int i;
    int debug = 1;
    int                   cert_chain_length = 0;
    char *                certfile    = NULL;
    char *                keyfile     = NULL;
    char *                outfile     = NULL;
    char *                certdir     = NULL;
    char *                certcafile  = NULL;
    char *                s           = NULL;

    X509 *                              ncert = NULL;
    X509 *                              icert = NULL;
    STACK_OF(X509_EXTENSION) *          extensions = NULL;
    X509_EXTENSION *                    ex = NULL;
    proxy_cred_desc* ipcd    = (*(gss_cred_id_desc**)issuer_cred)->pcd;
    proxy_cred_desc* sub_pcd = (*(gss_cred_id_desc**)sub_cred)->pcd;

    // SIGN
    // int sign_request(issuer, sub_cred, request, hours, limited)
    fprintf(stderr,"Adding Extensions to request\n");
    if ((extensions = sk_X509_EXTENSION_new_null()) == NULL)
    {
        PRXYerr(PRXYERR_F_PROXY_SIGN,PRXYERR_R_CLASS_ADD_EXT);
        goto err;
    }

#ifdef CLASS_ADD
    if (class_add_buf && class_add_buf_len > 0)
    {
        if ((ex = proxy_extension_class_add_create(class_add_buf,
                                                   class_add_buf_len)) == NULL)
        {
            PRXYerr(PRXYERR_F_PROXY_SIGN,PRXYERR_R_CLASS_ADD_EXT);
            goto err;
        }

        if (!sk_X509_EXTENSION_push(extensions, ex))
        {
            PRXYerr(PRXYERR_F_PROXY_SIGN,PRXYERR_R_CLASS_ADD_EXT);
            goto err;
        }
    }
#endif
    if (proxy_sign(ipcd->ucert,
                   ipcd->upkey,
                   (X509_REQ*)*req,
                   &ncert,
                   hours*60*60,
                   extensions,
                   limit_proxy))
    {
	printf("proxy_sign\n");
        goto err;
    }
    else
	printf("signed cert\n");

//********************************************
// CONSTRUCT PCD
    if( (*(gss_cred_id_desc**)sub_cred)->pcd == NULL )
	printf("1 NULL\n");
    printf("1 X\n");
    if( (*(gss_cred_id_desc**)sub_cred)->pcd->ucert == NULL )
	printf("2 NULL\n");
    printf("2 X\n");
    if( sub_pcd->ucert == NULL )
	printf("NULL\n");
    printf(" sub_pcd->ucert = ncert\n");
    sub_pcd->ucert = ncert;
    printf(" sub_pcd->ucert = ncert\n");

//********************************************
//build_cert_chain(issuer, sub_cred)

    if (sub_pcd->cert_chain == NULL)
	sub_pcd->cert_chain = sk_X509_new_null();
    printf(" sub_pcd->cert_chain = sk_X509_new_null()\n");


    if(ipcd->cert_chain != NULL)
    {
	cert_chain_length = sk_X509_num(ipcd->cert_chain);
	printf("XCredPath: %d\n", cert_chain_length);
    }
    for(i=cert_chain_length-1;i>=0;i--)
    {
	icert = sk_X509_value(ipcd->cert_chain,i);
	s = X509_NAME_oneline(X509_get_subject_name(icert), NULL, 0);
	printf("\t%d) %s\n", i, s);
	free(s);
	sk_X509_push(sub_pcd->cert_chain, icert);
	//i = sk_X509_insert(sub_pcd->cert_chain, pcd->ucert, sk_X509_num(pcd->cert_chain));
    }
    sk_X509_push(sub_pcd->cert_chain, ipcd->ucert);
    (*(gss_cred_id_desc**)sub_cred)->pcd = sub_pcd;
    return 0;

err:
    {
        unsigned long l;
        char buf[256];
#if SSLEAY_VERSION_NUMBER  >= 0x00904100L
        const char *file;
#else
        char *file;
#endif
        const char *data;
        int line;
        int flags;
	printf("Error: \n");
        while ( ERR_peek_error() != 0 )
        {
	    printf("-->\n");
            l=ERR_get_error_line_data(&file,&line,&data,&flags);
            if (debug)

            fprintf(stderr,
                    "%s:%s:%d%s\n",
                    ERR_error_string(l,buf),
                    file,
                    line,
                    data);

            else
                fprintf(stderr,
                        "%s%s\nFunction: %s\n",
                        ERR_reason_error_string(l),
                        data,
                        ERR_func_error_string(l));
        }
    }
    return -1;
}

PyObject*
get_cred_path(gss_cred_id_t* cred) {
    char                  *s;
    int                    i, cert_chain_length;
    X509                  *cert = NULL;
    gss_cred_id_desc     **output_cred_handle = (gss_cred_id_desc**) cred;
    PyObject              *tuple;


    if((*output_cred_handle)->pcd->cert_chain != NULL) {
        cert_chain_length = sk_X509_num((*output_cred_handle)->pcd->cert_chain);
        printf("CredPath: %d\n", cert_chain_length);
    }
    else {
        printf("cert chain is null\n");
    }
    if (!(tuple=PyTuple_New(cert_chain_length+1)))
    {
        PyErr_SetString(PyExc_RuntimeError,
                        "import_name: PyTuple_New()");
        return NULL;
    }
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", cert_chain_length));
    for(i=cert_chain_length-1;i>=0;i--) {
        cert = sk_X509_value((*output_cred_handle)->pcd->cert_chain,i);
        s = X509_NAME_oneline(X509_get_subject_name(cert), NULL, 0);
        printf("\t%d) %s\n", i, s);
        PyTuple_SET_ITEM(tuple, (i+1), Py_BuildValue("s#", s));
        free(s);
    }
    return tuple;
}

int
set_cred_globusuid(gss_cred_id_t* cred)
{
    gss_cred_id_desc     **output_cred_handle = (gss_cred_id_desc**) cred;
    /*
     * get the globusid, which is the subject name - any proxy entries
     */

    (*output_cred_handle)->globusid = (gss_name_desc*) malloc(sizeof(gss_name_desc)) ;
    if ((*output_cred_handle)->globusid == NULL)
    {
        GSSerr(GSSERR_F_ACQUIRE_CRED, GSSERR_R_OUT_OF_MEMORY);
        goto err;
    }
    (*output_cred_handle)->globusid->name_oid = GSS_C_NO_OID;
    (*output_cred_handle)->globusid->x509n =
        X509_NAME_dup(X509_get_subject_name((*output_cred_handle)->pcd->ucert));

    if ((*output_cred_handle)->globusid->x509n == NULL)
    {
        GSSerr(GSSERR_F_ACQUIRE_CRED,GSSERR_R_PROCESS_CERT);
        goto err;
    }
    printf("name : %s\n", (*output_cred_handle)->globusid->x509n);
    return 0;

    err:
    return -1;
}
%}
