/*############################################################################
# Keith R. Jackson, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################*/
%module utilc
%{
#include "globus_common.h"
#include "globus_module.h"
#include "globus_time.h"
%}

%typemap(python, in) PyObject * {
    $1 = $input;
}

%typemap(python, out) PyObject * {
    $result=$1;
}

%inline%{
globus_module_descriptor_t*
get_module(void) {
    return GLOBUS_COMMON_MODULE;
}

PyObject*
create_buffer(int size) {
    globus_byte_t* buf;
    PyObject *tuple, *ptr;
    int ret;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_get: PyTuple_New()");
        return NULL;
    }
    ret = globus_module_activate(GLOBUS_COMMON_MODULE);
    if (ret != GLOBUS_SUCCESS) {
        PyErr_SetString(PyExc_RuntimeError, "create_buffer: module failure");
        return NULL;
    }
    buf = globus_malloc(sizeof(globus_byte_t) * size);
    if (buf == NULL) {
        PyErr_SetString(PyExc_RuntimeError, "create_buffer: Memory Error");
        return NULL;
    }
    memset(buf, 0, size);
    ptr = SWIG_NewPointerObj((void*)buf, SWIGTYPE_p_globus_byte_t, 0);
    PyTuple_SET_ITEM(tuple, 0, ptr);
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", size));
    return tuple;
}

void 
destroy_buffer(globus_byte_t* buf) {
    int ret;
    globus_libc_free(buf);
    ret = globus_module_deactivate(GLOBUS_COMMON_MODULE);
    return;
}

PyObject*
buffer_to_string(globus_byte_t* buf, int size) {
    PyObject* ret;
    int s;
#ifdef DEBUG
    printf("in the C wrapper, size is %d\n", size);
#endif
    ret = PyString_FromStringAndSize((char*)buf, size);
    s = PyString_Size(ret);
#ifdef DEBUG
    printf("PyString_Size is: %d\n", s);
#endif
    return ret;
}
 
PyObject*
get_hostname(int len) {
    int ret;
    PyObject *tuple;
    char* str;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "client_get: PyTuple_New()");
        return NULL;
    }
    ret = globus_module_activate(GLOBUS_COMMON_MODULE);
    if (ret != GLOBUS_SUCCESS) {
        PyErr_SetString(PyExc_RuntimeError, "create_buffer: module failure");
        return NULL;
    }
    str = globus_malloc(sizeof(char) * len);
    ret = globus_libc_gethostname(str, len);
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", ret));
    if (ret != 0) {
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    else {
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", str));
    }
    globus_libc_free(str);
    return tuple;
}

void
free_off_t(globus_off_t * size)
    {
	globus_free(size);
    }

PyObject * 
convert_off_t(globus_off_t * size)
     {
	 PyObject * ptr;	 
	 ptr = PyLong_FromLongLong(*size);
	 return ptr;
     }

void 
free_abstime(globus_abstime_t * t)
    {
	globus_free(t);
    }
	
PyObject*
getSecs(globus_abstime_t * modTime)
    {
	PyObject * ptr;
	long seconds;
	seconds = modTime->tv_sec;
	ptr = PyLong_FromLong(seconds);
	return ptr;
    }

PyObject*
getNSecs(globus_abstime_t * modTime)
    {
	PyObject * ptr;
	long seconds;
	seconds = modTime->tv_nsec;
	ptr = PyLong_FromLong(seconds);
	return ptr;
    }

%}

extern int
globus_module_activate(globus_module_descriptor_t *module_descriptor);

extern int
globus_module_deactivate(globus_module_descriptor_t *module_descriptor);

extern int
globus_module_deactivate_all(void);


